@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2017 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -101,6 +101,7 @@ void managesieve_parser_reset(struct managesieve_parser *parser)
 
 	if ( parser->str_stream != NULL )
 		i_stream_unref(&parser->str_stream);
+	parser->str_stream = NULL;
 }
 
 const char *managesieve_parser_get_error
@@ -111,7 +112,8 @@ const char *managesieve_parser_get_error
 }
 
 /* skip over everything parsed so far, plus the following whitespace */
-static int managesieve_parser_skip_to_next(struct managesieve_parser *parser,
+static bool
+managesieve_parser_skip_to_next(struct managesieve_parser *parser,
 				    const unsigned char **data,
 				    size_t *data_size)
 {
@@ -200,7 +202,8 @@ static void managesieve_parser_save_arg(struct managesieve_parser *parser,
 	parser->cur_type = ARG_PARSE_NONE;
 }
 
-static int is_valid_atom_char(struct managesieve_parser *parser, char chr)
+static bool
+is_valid_atom_char(struct managesieve_parser *parser, char chr)
 {
 	if (IS_ATOM_SPECIAL((unsigned char)chr)) {
 		parser->error = "Invalid characters in atom";
@@ -213,7 +216,8 @@ static int is_valid_atom_char(struct managesieve_parser *parser, char chr)
 	return TRUE;
 }
 
-static int managesieve_parser_read_atom(struct managesieve_parser *parser,
+static bool
+managesieve_parser_read_atom(struct managesieve_parser *parser,
 				 const unsigned char *data, size_t data_size)
 {
 	size_t i;
@@ -229,10 +233,11 @@ static int managesieve_parser_read_atom(struct managesieve_parser *parser,
 	}
 
 	parser->cur_pos = i;
-	return parser->cur_type == ARG_PARSE_NONE;
+	return ( parser->cur_type == ARG_PARSE_NONE );
 }
 
-static int managesieve_parser_read_string(struct managesieve_parser *parser,
+static bool
+managesieve_parser_read_string(struct managesieve_parser *parser,
 				   const unsigned char *data, size_t data_size)
 {
 	size_t i;
@@ -286,10 +291,11 @@ static int managesieve_parser_read_string(struct managesieve_parser *parser,
 	}
 
 	parser->cur_pos = i;
-	return parser->cur_type == ARG_PARSE_NONE;
+	return ( parser->cur_type == ARG_PARSE_NONE );
 }
 
-static int managesieve_parser_literal_end(struct managesieve_parser *parser)
+static bool
+managesieve_parser_literal_end(struct managesieve_parser *parser)
 {
 	if ((parser->flags & MANAGESIEVE_PARSE_FLAG_STRING_STREAM) == 0) {
 		if (parser->line_size >= parser->max_line_size ||
@@ -309,7 +315,8 @@ static int managesieve_parser_literal_end(struct managesieve_parser *parser)
 	return TRUE;
 }
 
-static int managesieve_parser_read_literal(struct managesieve_parser *parser,
+static bool
+managesieve_parser_read_literal(struct managesieve_parser *parser,
 				    const unsigned char *data,
 				    size_t data_size)
 {
@@ -353,7 +360,8 @@ static int managesieve_parser_read_literal(struct managesieve_parser *parser,
 	return FALSE;
 }
 
-static int managesieve_parser_read_literal_data(struct managesieve_parser *parser,
+static bool
+managesieve_parser_read_literal_data(struct managesieve_parser *parser,
 					 const unsigned char *data,
 					 size_t data_size)
 {
@@ -414,7 +422,8 @@ static int managesieve_parser_read_literal_data(struct managesieve_parser *parse
 
 /* Returns TRUE if argument was fully processed. Also returns TRUE if
    an argument inside a list was processed. */
-static int managesieve_parser_read_arg(struct managesieve_parser *parser)
+static bool
+managesieve_parser_read_arg(struct managesieve_parser *parser)
 {
 	const unsigned char *data;
 	size_t data_size;
@@ -620,7 +629,7 @@ static ssize_t quoted_string_istream_read(struct istream_private *stream)
 	struct quoted_string_istream *qsstream =
 		(struct quoted_string_istream *)stream;
 	const unsigned char *data;
-	size_t i, dest, size;
+	size_t i, dest, size, avail;
 	ssize_t ret = 0;
 	bool slash;
 
@@ -634,13 +643,15 @@ static ssize_t quoted_string_istream_read(struct istream_private *stream)
 	if (size == 0) {
 		ret = i_stream_read(stream->parent);
 		if (ret <= 0 && (ret != -2 || stream->skip == 0)) {
-			if ( stream->istream.eof && stream->istream.stream_errno == 0 ) {
-				stream->istream.eof = 0;
-				stream->istream.stream_errno = EIO;
-			} else {
-				stream->istream.stream_errno = stream->parent->stream_errno;
-				stream->istream.eof = stream->parent->eof;
+			if ( stream->parent->eof && stream->parent->stream_errno == 0 ) {
+				io_stream_set_error(&stream->iostream,
+					"Quoted string ends without closing quotes");
+				stream->istream.stream_errno = EINVAL;
+				return -1;
 			}
+
+			stream->istream.stream_errno = stream->parent->stream_errno;
+			stream->istream.eof = stream->parent->eof;
 			return ret;
 		}
 		data = i_stream_get_data(stream->parent, &size);
@@ -648,7 +659,7 @@ static ssize_t quoted_string_istream_read(struct istream_private *stream)
 	}
 
 	/* Allocate buffer space */
-	if (!i_stream_get_buffer_space(stream, size, NULL))
+	if (!i_stream_try_alloc(stream, size, &avail))
 		return -2;
 
 	/* Parse quoted string content */
@@ -671,9 +682,9 @@ static ssize_t quoted_string_istream_read(struct istream_private *stream)
 			slash = FALSE;
 		} else if ( slash ) {
 			if ( !IS_QUOTED_SPECIAL(data[i]) ) {
-				qsstream->parser->error =
-					"Escaped quoted-string character is not a QUOTED-SPECIAL.";
-				stream->istream.stream_errno = EIO;
+				io_stream_set_error(&stream->iostream,
+					"Escaped quoted-string character is not a QUOTED-SPECIAL");
+				stream->istream.stream_errno = EINVAL;
 				ret = -1;
 				break;
 			}
@@ -681,8 +692,9 @@ static ssize_t quoted_string_istream_read(struct istream_private *stream)
 		}
 
 		if ( (data[i] & 0x80) == 0 && ( data[i] == '\r' || data[i] == '\n' ) ) {
-			qsstream->parser->error = "String contains invalid character.";
-			stream->istream.stream_errno = EIO;
+			io_stream_set_error(&stream->iostream,
+				"Quoted string contains an invalid character");
+			stream->istream.stream_errno = EINVAL;
 			ret = -1;
 			break;
 		}
@@ -712,10 +724,16 @@ static ssize_t quoted_string_istream_read(struct istream_private *stream)
 	return ret;
 }
 
-static const struct stat *quoted_string_istream_stat
+static int quoted_string_istream_stat
 (struct istream_private *stream, bool exact)
 {
-	return i_stream_stat(stream->parent, exact);
+	const struct stat *st;
+
+	if (i_stream_stat(stream->parent, exact, &st) < 0)
+		return -1;
+
+	stream->statbuf = *st;
+	return 0;
 }
 
 static struct istream *quoted_string_istream_create
