@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #ifndef __EXT_BODY_COMMON_H
@@ -36,8 +36,8 @@ extern const struct sieve_operation_def body_operation;
  * Message body part extraction
  */
 
-struct sieve_stringlist *ext_body_get_part_list
+int ext_body_get_part_list
 	(const struct sieve_runtime_env *renv, enum tst_body_transform transform,
-		const char * const *content_types);
+		const char * const *content_types, struct sieve_stringlist **strlist_r);
 
 #endif /* __EXT_BODY_COMMON_H */
