@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2017 Pigeonhole authors, see the included COPYING file
  */
 
 /* Extension body
@@ -7,27 +7,11 @@
  * Authors: Stephan Bosch
  *          Original CMUSieve implementation by Timo Sirainen
  * Specification: RFC 5173
- * Implementation: almost full, but text body-transform implementation is simple
+ * Implementation: full
  * Status: testing
  *
  */
 
-/* FIXME:
- *
- * From RFC with respect to :text body transform:
- *
- * "Sophisticated implementations MAY strip mark-up from the text prior
- *  to matching, and MAY convert media types other than text to text
- *  prior to matching.
- *
- *  (For example, they may be able to convert proprietary text editor
- *  formats to text or apply optical character recognition algorithms to
- *  image data.)"
- *
- * We might want to do this in the future, i.e. we must evaluate whether this is
- * feasible.
- */
-
 #include "lib.h"
 #include "array.h"
 
@@ -53,12 +37,9 @@ static bool ext_body_validator_load
 (const struct sieve_extension *ext, struct sieve_validator *valdtr);
 
 const struct sieve_extension_def body_extension = {
-	"body",
-	NULL, NULL,
-	ext_body_validator_load,
-	NULL, NULL, NULL, NULL, NULL,
-	SIEVE_EXT_DEFINE_OPERATION(body_operation),
-	SIEVE_EXT_DEFINE_NO_OPERANDS
+	.name = "body",
+	.validator_load =	ext_body_validator_load,
+	SIEVE_EXT_DEFINE_OPERATION(body_operation)
 };
 
 static bool ext_body_validator_load
