@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2017 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -320,10 +320,11 @@ static int _ext_enotify_option_check
 	struct sieve_enotify_env nenv;
 	const char *option = sieve_ast_argument_strc(arg);
 	const char *opt_name = NULL, *opt_value = NULL;
-	bool result = TRUE, check = TRUE;
+	bool check = TRUE;
+	int result = 1;
 
 	/* Compose log structure */
-	memset(&nenv, 0, sizeof(nenv));
+	i_zero(&nenv);
 	nenv.svinst = optn_context->svinst;
 	nenv.method = method;
 	nenv.ehandler = sieve_prefix_ehandler_create
@@ -349,13 +350,15 @@ static int _ext_enotify_option_check
 		/* Literal string: full option parse */
 		if ( !ext_enotify_option_parse
 			(&nenv, option, FALSE, &opt_name, &opt_value) )
-			result = FALSE;
+			result = -1;
 	}
 
 	/* Call method's option check function */
-	if ( result && check && method->def != NULL &&
-		method->def->compile_check_option != NULL )
-		result = method->def->compile_check_option(&nenv, opt_name, opt_value);
+	if ( result > 0 && check && method->def != NULL &&
+		method->def->compile_check_option != NULL ) {
+		result = ( method->def->compile_check_option
+			(&nenv, opt_name, opt_value) ? 1 : -1 );
+	}
 
 	sieve_error_handler_unref(&nenv.ehandler);
 
@@ -399,7 +402,7 @@ bool ext_enotify_compile_check_arguments
 	if ( method->def == NULL ) return TRUE;
 
 	/* Compose log structure */
-	memset(&nenv, 0, sizeof(nenv));
+	i_zero(&nenv);
 	nenv.svinst = svinst;
 	nenv.method = method;
 
@@ -498,7 +501,7 @@ bool ext_enotify_runtime_method_validate
 	if ( method->def != NULL && method->def->runtime_check_uri != NULL ) {
 		struct sieve_enotify_env nenv;
 
-		memset(&nenv, 0, sizeof(nenv));
+		i_zero(&nenv);
 		nenv.svinst = renv->svinst;
 		nenv.method = method;
 		nenv.ehandler = sieve_prefix_ehandler_create
@@ -562,7 +565,7 @@ const char *ext_enotify_runtime_get_method_capability
 		method->def->runtime_get_method_capability != NULL ) {
 		struct sieve_enotify_env nenv;
 
-		memset(&nenv, 0, sizeof(nenv));
+		i_zero(&nenv);
 		nenv.svinst = renv->svinst;
 		nenv.method = method;
 		nenv.ehandler = sieve_prefix_ehandler_create
@@ -596,7 +599,7 @@ int ext_enotify_runtime_check_operands
 		struct sieve_enotify_env nenv;
 		int result = SIEVE_EXEC_OK;
 
-		memset(&nenv, 0, sizeof(nenv));
+		i_zero(&nenv);
 		nenv.svinst = renv->svinst;
 		nenv.method = method;
 		nenv.ehandler = sieve_prefix_ehandler_create
