@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 /* Notify method mailto
@@ -23,13 +23,16 @@
 #include "str.h"
 #include "ioloop.h"
 #include "str-sanitize.h"
+#include "ostream.h"
 #include "message-date.h"
 #include "mail-storage.h"
 
 #include "sieve-common.h"
 #include "sieve-address.h"
+#include "sieve-address-source.h"
 #include "sieve-message.h"
 #include "sieve-smtp.h"
+#include "sieve-settings.h"
 
 #include "sieve-ext-enotify.h"
 
@@ -45,10 +48,23 @@
 #define NTFY_MAILTO_MAX_HEADERS     16
 #define NTFY_MAILTO_MAX_SUBJECT     256
 
+/*
+ * Mailto notification configuration
+ */
+
+struct ntfy_mailto_config {
+	struct sieve_address_source envelope_from;
+};
+
 /*
  * Mailto notification method
  */
 
+static bool ntfy_mailto_load
+	(const struct sieve_enotify_method *nmth, void **context);
+static void ntfy_mailto_unload
+	(const struct sieve_enotify_method *nmth);
+
 static bool ntfy_mailto_compile_check_uri
 	(const struct sieve_enotify_env *nenv, const char *uri, const char *uri_body);
 static bool ntfy_mailto_compile_check_from
@@ -73,13 +89,14 @@ static void ntfy_mailto_action_print
 	(const struct sieve_enotify_print_env *penv,
 		const struct sieve_enotify_action *nact);
 
-static bool ntfy_mailto_action_execute
+static int ntfy_mailto_action_execute
 	(const struct sieve_enotify_exec_env *nenv,
 		const struct sieve_enotify_action *nact);
 
 const struct sieve_enotify_method_def mailto_notify = {
 	"mailto",
-	NULL, NULL,
+	ntfy_mailto_load,
+	ntfy_mailto_unload,
 	ntfy_mailto_compile_check_uri,
 	NULL,
 	ntfy_mailto_compile_check_from,
@@ -131,6 +148,42 @@ struct ntfy_mailto_context {
 	const char *from_normalized;
 };
 
+/*
+ * Method registration
+ */
+
+static bool ntfy_mailto_load
+(const struct sieve_enotify_method *nmth, void **context)
+{
+	struct sieve_instance *svinst = nmth->svinst;
+	struct ntfy_mailto_config *config;
+
+	if ( *context != NULL ) {
+		ntfy_mailto_unload(nmth);
+	}
+
+	config = i_new(struct ntfy_mailto_config, 1);
+
+	(void)sieve_address_source_parse_from_setting	(svinst,
+		default_pool, "sieve_notify_mailto_envelope_from",
+		&config->envelope_from);
+
+	*context = (void *) config;
+
+	return TRUE;
+}
+
+static void ntfy_mailto_unload
+(const struct sieve_enotify_method *nmth)
+{
+	struct ntfy_mailto_config *config =
+		(struct ntfy_mailto_config *) nmth->context;
+	char *address = (char *)config->envelope_from.address;
+
+	i_free(address);
+	i_free(config);
+}
+
 /*
  * Validation
  */
@@ -218,7 +271,7 @@ static bool ntfy_mailto_runtime_check_operands
 				mtctx->from_normalized = p_strdup(context_pool, normalized);
 		} T_END;
 
-		if ( !normalized ) return FALSE;
+		if ( normalized == NULL ) return FALSE;
 	}
 
 	if ( (parsed_uri=uri_mailto_parse
@@ -304,7 +357,8 @@ static void ntfy_mailto_action_print
 	/* Print main method parameters */
 
 	sieve_enotify_method_printf
-		(penv,   "    => importance   : %d\n", nact->importance);
+		(penv,   "    => importance   : %llu\n",
+			(unsigned long long)nact->importance);
 
 	if ( nact->message != NULL )
 		sieve_enotify_method_printf
@@ -373,55 +427,86 @@ static bool _contains_8bit(const char *msg)
 	return FALSE;
 }
 
-static bool ntfy_mailto_send
+static int ntfy_mailto_send
 (const struct sieve_enotify_exec_env *nenv,
-	const struct sieve_enotify_action *nact, const char *recipient)
+	const struct sieve_enotify_action *nact,
+	const char *owner_email)
 {
+	struct sieve_instance *svinst = nenv->svinst;
 	const struct sieve_message_data *msgdata = nenv->msgdata;
 	const struct sieve_script_env *senv = nenv->scriptenv;
 	struct ntfy_mailto_context *mtctx =
 		(struct ntfy_mailto_context *) nact->method_context;
+	struct ntfy_mailto_config *mth_config =
+		(struct ntfy_mailto_config *)nenv->method->context;
+	struct sieve_address_source env_from =
+		mth_config->envelope_from;
 	const char *from = NULL, *from_smtp = NULL;
 	const char *subject = mtctx->uri->subject;
 	const char *body = mtctx->uri->body;
-	string_t *to, *cc;
+	string_t *to, *cc, *all;
 	const struct uri_mailto_recipient *recipients;
-	void *smtp_handle;
-	unsigned int count, i;
-	FILE *f;
-	const char *outmsgid;
+	const struct uri_mailto_header_field *headers;
+	struct sieve_smtp_context *sctx;
+	struct ostream *output;
+	string_t *msg;
+	unsigned int count, i, hcount, h;
+	const char *outmsgid, *error;
+	int ret;
 
 	/* Get recipients */
 	recipients = array_get(&mtctx->uri->recipients, &count);
 	if ( count == 0  ) {
 		sieve_enotify_warning(nenv,
 			"notify mailto uri specifies no recipients; action has no effect");
-		return TRUE;
+		return 0;
 	}
 
 	/* Just to be sure */
 	if ( !sieve_smtp_available(senv) ) {
 		sieve_enotify_global_warning(nenv,
 			"notify mailto method has no means to send mail");
-		return TRUE;
+		return 0;
+	}
+
+	/* Determine which sender to use
+
+	   From RFC 5436, Section 2.3:
+
+		 The ":from" tag overrides the default sender of the notification
+		 message.  "Sender", here, refers to the value used in the [RFC5322]
+		 "From" header.  Implementations MAY also use this value in the
+		 [RFC5321] "MAIL FROM" command (the "envelope sender"), or they may
+		 prefer to establish a mailbox that receives bounces from notification
+		 messages.
+	 */
+	if ( (nenv->flags & SIEVE_EXECUTE_FLAG_NO_ENVELOPE) == 0 ) {
+		from_smtp = sieve_message_get_sender(nenv->msgctx);
+		if ( from_smtp == NULL ) {
+			/* "<>" */
+			memset(&env_from, 0, sizeof(env_from));
+			env_from.type = SIEVE_ADDRESS_SOURCE_EXPLICIT;
+		}
+	}
+	if ( (ret=sieve_address_source_get_address(&env_from, svinst,
+		senv, nenv->msgctx, nenv->flags, &from_smtp)) < 0 ) {
+		from_smtp = NULL;
+	} else if ( ret == 0 ) {
+		if ( mtctx->from_normalized != NULL )
+			from_smtp = mtctx->from_normalized;
+		else if ( svinst->user_email != NULL )
+			from_smtp = sieve_address_to_string(svinst->user_email);
+		else
+			from_smtp = senv->postmaster_address;
 	}
 
 	/* Determine message from address */
 	if ( nact->from == NULL ) {
-		from = t_strdup_printf("Postmaster <%s>", senv->postmaster_address);
+		from = t_strdup_printf("<%s>", from_smtp);
 	} else {
 		from = nact->from;
 	}
 
-	/* Determine SMTP from address */
-	if ( sieve_message_get_sender(nenv->msgctx) != NULL ) {
-		if ( mtctx->from_normalized == NULL ) {
-			from_smtp = senv->postmaster_address;
-		} else {
-			from_smtp = mtctx->from_normalized;
-		}
-	}
-
 	/* Determine subject */
 	if ( nact->message != NULL ) {
 		/* FIXME: handle UTF-8 */
@@ -441,6 +526,7 @@ static bool ntfy_mailto_send
 	/* Compose To and Cc headers */
 	to = NULL;
 	cc = NULL;
+	all = t_str_new(256);
 	for ( i = 0; i < count; i++ ) {
 		if ( recipients[i].carbon_copy ) {
 			if ( cc == NULL ) {
@@ -459,126 +545,166 @@ static bool ntfy_mailto_send
 				str_append(to, recipients[i].full);
 			}
 		}
+		if ( i < 3) {
+			if ( i > 0 )
+				str_append(all, ", ");
+			str_append_c(all, '<');
+			str_append(all, str_sanitize(recipients[i].normalized, 256));
+			str_append_c(all, '>');
+		} else if (i == 3) {
+			str_printfa(all, ", ... (%u total)", count);
+		}
 	}
 
+	msg = t_str_new(512);
+	outmsgid = sieve_message_get_new_id(svinst);
+
+	rfc2822_header_write(msg, "X-Sieve", SIEVE_IMPLEMENTATION);
+	rfc2822_header_write(msg, "Message-ID", outmsgid);
+	rfc2822_header_write(msg, "Date", message_date_create(ioloop_time));
+	rfc2822_header_utf8_printf(msg, "Subject", "%s", subject);
+
+	rfc2822_header_utf8_printf(msg, "From", "%s", from);
+
+	if ( to != NULL )
+		rfc2822_header_utf8_printf(msg, "To", "%s", str_c(to));
+
+	if ( cc != NULL )
+		rfc2822_header_utf8_printf(msg, "Cc", "%s", str_c(cc));
+
+	rfc2822_header_printf(msg, "Auto-Submitted",
+		"auto-notified; owner-email=\"%s\"", owner_email);
+	rfc2822_header_write(msg, "Precedence", "bulk");
+
+	/* Set importance */
+	switch ( nact->importance ) {
+	case 1:
+		rfc2822_header_write(msg, "X-Priority", "1 (Highest)");
+		rfc2822_header_write(msg, "Importance", "High");
+		break;
+	case 3:
+		rfc2822_header_write(msg, "X-Priority", "5 (Lowest)");
+		rfc2822_header_write(msg, "Importance", "Low");
+		break;
+	case 2:
+	default:
+		rfc2822_header_write(msg, "X-Priority", "3 (Normal)");
+		rfc2822_header_write(msg, "Importance", "Normal");
+		break;
+	}
+
+	/* Add custom headers */
+
+	headers = array_get(&mtctx->uri->headers, &hcount);
+	for ( h = 0; h < hcount; h++ ) {
+		const char *name = rfc2822_header_field_name_sanitize(headers[h].name);
+
+		rfc2822_header_write(msg, name, headers[h].body);
+	}
+
+	/* Generate message body */
+
+	rfc2822_header_write(msg, "MIME-Version", "1.0");
+	if ( body != NULL ) {
+		if (_contains_8bit(body)) {
+			rfc2822_header_write
+				(msg, "Content-Type", "text/plain; charset=utf-8");
+			rfc2822_header_write(msg, "Content-Transfer-Encoding", "8bit");
+		} else {
+			rfc2822_header_write
+				(msg, "Content-Type", "text/plain; charset=us-ascii");
+			rfc2822_header_write(msg, "Content-Transfer-Encoding", "7bit");
+		}
+		str_printfa(msg, "\r\n%s\r\n", body);
+
+	} else {
+		rfc2822_header_write
+			(msg, "Content-Type", "text/plain; charset=US-ASCII");
+		rfc2822_header_write(msg, "Content-Transfer-Encoding", "7bit");
+
+		str_append(msg, "\r\nNotification of new message.\r\n");
+	}
+
+	sctx = sieve_smtp_start(senv, from_smtp);
+
 	/* Send message to all recipients */
-	for ( i = 0; i < count; i++ ) {
-		const struct uri_mailto_header_field *headers;
-		unsigned int h, hcount;
+	for ( i = 0; i < count; i++ )
+		sieve_smtp_add_rcpt(sctx, recipients[i].normalized);
 
-		smtp_handle = sieve_smtp_open
-			(senv, recipients[i].normalized, from_smtp, &f);
-		outmsgid = sieve_message_get_new_id(nenv->svinst);
+	output = sieve_smtp_send(sctx);
+	o_stream_send(output, str_data(msg), str_len(msg));
 
-		rfc2822_header_field_write(f, "X-Sieve", SIEVE_IMPLEMENTATION);
-		rfc2822_header_field_write(f, "Message-ID", outmsgid);
-		rfc2822_header_field_write(f, "Date", message_date_create(ioloop_time));
-		rfc2822_header_field_utf8_printf(f, "Subject", "%s", subject);
-
-		rfc2822_header_field_utf8_printf(f, "From", "%s", from);
-
-		if ( to != NULL )
-			rfc2822_header_field_utf8_printf(f, "To", "%s", str_c(to));
-
-		if ( cc != NULL )
-			rfc2822_header_field_utf8_printf(f, "Cc", "%s", str_c(cc));
-
-		rfc2822_header_field_printf(f, "Auto-Submitted",
-			"auto-notified; owner-email=\"%s\"", recipient);
-		rfc2822_header_field_write(f, "Precedence", "bulk");
-
-		/* Set importance */
-		switch ( nact->importance ) {
-		case 1:
-			rfc2822_header_field_write(f, "X-Priority", "1 (Highest)");
-			rfc2822_header_field_write(f, "Importance", "High");
-			break;
-		case 3:
-			rfc2822_header_field_write(f, "X-Priority", "5 (Lowest)");
-			rfc2822_header_field_write(f, "Importance", "Low");
-			break;
-		case 2:
-		default:
-			rfc2822_header_field_write(f, "X-Priority", "3 (Normal)");
-			rfc2822_header_field_write(f, "Importance", "Normal");
-			break;
-		}
-
-		/* Add custom headers */
-
-		headers = array_get(&mtctx->uri->headers, &hcount);
-		for ( h = 0; h < hcount; h++ ) {
-			const char *name = rfc2822_header_field_name_sanitize(headers[h].name);
-
-			rfc2822_header_field_write(f, name, headers[h].body);
-		}
-
-		/* Generate message body */
-		if ( body != NULL ) {
-			if (_contains_8bit(body)) {
-				rfc2822_header_field_write(f, "MIME-Version", "1.0");
-				rfc2822_header_field_write
-					(f, "Content-Type", "text/plain; charset=UTF-8");
-				rfc2822_header_field_write(f, "Content-Transfer-Encoding", "8bit");
-			}
-
-			fprintf(f, "\r\n");
-			fprintf(f, "%s\r\n", body);
-
-		} else {
-			fprintf(f, "\r\n");
-			fprintf(f, "Notification of new message.\r\n");
-		}
-
-		if ( sieve_smtp_close(senv, smtp_handle) ) {
-			sieve_enotify_global_info(nenv,
-				"sent mail notification to <%s>",
-				str_sanitize(recipients[i].normalized, 80));
-		} else {
+	if ( (ret=sieve_smtp_finish(sctx, &error)) <= 0 ) {
+		if (ret < 0)  {
 			sieve_enotify_global_error(nenv,
-				"failed to send mail notification to <%s> "
-				"(refer to system log for more information)",
-				str_sanitize(recipients[i].normalized, 80));
+				"failed to send mail notification to %s: %s (temporary failure)",
+				str_c(all),	str_sanitize(error, 512));
+		} else {
+			sieve_enotify_global_log_error(nenv,
+				"failed to send mail notification to %s: %s (permanent failure)",
+				str_c(all),	str_sanitize(error, 512));
 		}
+	} else {
+		sieve_enotify_global_info(nenv,
+			"sent mail notification to %s", str_c(all));
 	}
 
-	return TRUE;
+	return 0;
 }
 
-static bool ntfy_mailto_action_execute
+static int ntfy_mailto_action_execute
 (const struct sieve_enotify_exec_env *nenv,
 	const struct sieve_enotify_action *nact)
 {
-	const char *const *headers;
-	const char *sender = sieve_message_get_sender(nenv->msgctx);
-	const char *recipient = sieve_message_get_final_recipient(nenv->msgctx);
+	struct sieve_instance *svinst = nenv->svinst;
+	const struct sieve_script_env *senv = nenv->scriptenv;
+	struct mail *mail = nenv->msgdata->mail;
+	const char *owner_email;
+	const char *const *hdsp;
+	int ret;
 
-	/* Is the recipient unset?
-	 */
-	if ( recipient == NULL ) {
-		sieve_enotify_global_warning(nenv,
-			"notify mailto action aborted: envelope recipient is <>");
-		return TRUE;
-	}
+	owner_email = sieve_address_to_string(svinst->user_email);
+	if ( owner_email == NULL &&
+		(nenv->flags & SIEVE_EXECUTE_FLAG_NO_ENVELOPE) == 0 )
+		owner_email = sieve_message_get_final_recipient(nenv->msgctx);
+	if ( owner_email == NULL )
+		owner_email = senv->postmaster_address;
+	i_assert( owner_email != NULL );
 
 	/* Is the message an automatic reply ? */
-	if ( mail_get_headers
-		(nenv->msgdata->mail, "auto-submitted", &headers) >= 0 ) {
-		const char *const *hdsp = headers;
-
-		/* Theoretically multiple headers could exist, so lets make sure */
-		while ( *hdsp != NULL ) {
-			if ( strcasecmp(*hdsp, "no") != 0 ) {
-				sieve_enotify_global_info(nenv,
-					"not sending notification for auto-submitted message from <%s>",
-					str_sanitize(sender, 128));
-					return TRUE;
-			}
-			hdsp++;
-		}
+	if ( mail_get_headers(mail, "auto-submitted", &hdsp) < 0 ) {
+		sieve_enotify_critical(nenv,
+			"mailto notification: "
+				"failed to read `auto-submitted' header field",
+			"mailto notification: "
+				"failed to read `auto-submitted' header field: %s",
+			mailbox_get_last_error(mail->box, NULL));
+		return -1;
 	}
 
-	return ntfy_mailto_send(nenv, nact, recipient);
+	/* Theoretically multiple headers could exist, so lets make sure */
+	while ( *hdsp != NULL ) {
+		if ( strcasecmp(*hdsp, "no") != 0 ) {
+			const char *from = NULL;
+
+			if ( (nenv->flags & SIEVE_EXECUTE_FLAG_NO_ENVELOPE) == 0 )
+				from = sieve_message_get_sender(nenv->msgctx);
+			from = (from == NULL ? "" :
+				t_strdup_printf(" from <%s>", str_sanitize(from, 256)));
+
+			sieve_enotify_global_info(nenv,
+				"not sending notification for auto-submitted message%s",
+				from);
+			return 0;
+		}
+		hdsp++;
+	}
+
+	T_BEGIN {
+		ret = ntfy_mailto_send(nenv, nact, owner_email);
+	} T_END;
+
+	return ret;
 }
 
 
