@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2015 Pigeonhole authors, see the included COPYING file
  */
 
 #ifndef __EXT_ENVIRONMENT_COMMON_H
@@ -47,11 +47,10 @@ bool ext_environment_init(const struct sieve_extension *ext, void **context);
 void ext_environment_deinit(const struct sieve_extension *ext);
 
 /*
- * Environment item retrieval
+ * Validator context
  */
 
-const char *ext_environment_item_get_value
-	(const struct sieve_extension *ext, const char *name,
-		const struct sieve_script_env *senv);
+void ext_environment_interpreter_init
+(const struct sieve_extension *this_ext, struct sieve_interpreter *interp);
 
 #endif /* __EXT_VARIABLES_COMMON_H */
