@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 /* Extension variables
@@ -6,8 +6,8 @@
  *
  * Authors: Stephan Bosch
  * Specification: RFC 5183
- * Implementation: basic
- * Status: experimental, not thoroughly tested
+ * Implementation: full
+ * Status: testing
  *
  */
 
@@ -30,22 +30,30 @@
 
 static bool ext_environment_validator_load
 	(const struct sieve_extension *ext, struct sieve_validator *valdtr);
+static bool ext_environment_interpreter_load
+(const struct sieve_extension *ext,
+	const struct sieve_runtime_env *renv, sieve_size_t *address);
 
 const struct sieve_extension_def environment_extension = {
-	"environment",
-	ext_environment_init,
-	ext_environment_deinit,
-	ext_environment_validator_load,
-	NULL, NULL, NULL, NULL, NULL,
-	SIEVE_EXT_DEFINE_OPERATION(tst_environment_operation),
-	SIEVE_EXT_DEFINE_NO_OPERANDS
+	.name = "environment",
+	.validator_load = ext_environment_validator_load,
+	.interpreter_load = ext_environment_interpreter_load,
+	SIEVE_EXT_DEFINE_OPERATION(tst_environment_operation)
 };
 
 static bool ext_environment_validator_load
 (const struct sieve_extension *ext, struct sieve_validator *valdtr)
 {
 	sieve_validator_register_command(valdtr, ext, &tst_environment);
+	return TRUE;
+}
 
+static bool ext_environment_interpreter_load
+(const struct sieve_extension *ext,
+	const struct sieve_runtime_env *renv,
+	sieve_size_t *address ATTR_UNUSED)
+{
+	ext_environment_interpreter_init(ext, renv->interp);
 	return TRUE;
 }
 
