@@ -1,9 +1,11 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
 #include "str.h"
 #include "str-sanitize.h"
+#include "mail-storage.h"
+#include "imap-arg.h"
 
 #include "sieve-common.h"
 #include "sieve-commands.h"
@@ -20,12 +22,6 @@
 
 #include "ext-imap4flags-common.h"
 
-/*
- * Forward declarations
- */
-
-static bool flag_is_valid(const char *flag);
-
 /*
  * Tagged arguments
  */
@@ -131,9 +127,9 @@ bool ext_imap4flags_command_validate
 		ext_imap4flags_iter_init(&fiter, sieve_ast_argument_str(arg));
 
 		while ( (flag=ext_imap4flags_iter_get_flag(&fiter)) != NULL ) {
-			if ( !flag_is_valid(flag) ) {
+			if ( !sieve_ext_imap4flags_flag_is_valid(flag) ) {
 				sieve_argument_validate_warning(valdtr, arg,
-                	"IMAP flag '%s' specified for the %s command is invalid "
+					"IMAP flag '%s' specified for the %s command is invalid "
 					"and will be ignored (only first invalid is reported)",
 					str_sanitize(flag, 64), sieve_command_identifier(cmd));
 				break;
@@ -150,24 +146,28 @@ bool ext_imap4flags_command_validate
 
 void ext_imap4flags_attach_flags_tag
 (struct sieve_validator *valdtr, const struct sieve_extension *ext,
-	const char *command, bool implicit)
+	const char *command)
 {
 	/* Register :flags tag with the command and we don't care whether it is
 	 * registered or even whether it will be registered at all. The validator
 	 * handles either situation gracefully
 	 */
 
-	if ( !implicit ) {
-		/* Tag specified by user */
-		sieve_validator_register_external_tag
-			(valdtr, command, ext, &tag_flags, SIEVE_OPT_SIDE_EFFECT);
-	}
-
-    /* Implicit tag if none is specified */
-	sieve_validator_register_persistent_tag
-		(valdtr, command, ext, &tag_flags_implicit);
+	/* Tag specified by user */
+	sieve_validator_register_external_tag
+		(valdtr, command, ext, &tag_flags, SIEVE_OPT_SIDE_EFFECT);
 }
 
+void sieve_ext_imap4flags_register_side_effect
+(struct sieve_validator *valdtr, const struct sieve_extension *flg_ext,
+	const char *command)
+{
+	/* Implicit tag if none is specified */
+	sieve_validator_register_persistent_tag
+		(valdtr, command, flg_ext, &tag_flags_implicit);
+}
+
+
 /*
  * Result context
  */
@@ -242,18 +242,20 @@ static string_t *_get_flags_string
  * Runtime initialization
  */
 
-static void ext_imap4flags_runtime_init
-(const struct sieve_extension *ext, const struct sieve_runtime_env *renv,
-	void *context ATTR_UNUSED)
+static int ext_imap4flags_runtime_init
+(const struct sieve_extension *ext,
+	const struct sieve_runtime_env *renv,
+	void *context ATTR_UNUSED, bool deferred ATTR_UNUSED)
 {
 	sieve_result_add_implicit_side_effect
 		(renv->result, NULL, TRUE, ext, &flags_side_effect, NULL);
+	return SIEVE_EXEC_OK;
 }
 
-const struct sieve_interpreter_extension imap4flags_interpreter_extension = {
-	&imap4flags_extension,
-	ext_imap4flags_runtime_init,
-	NULL,
+const struct sieve_interpreter_extension
+imap4flags_interpreter_extension = {
+	.ext_def = &imap4flags_extension,
+	.run = ext_imap4flags_runtime_init
 };
 
 /*
@@ -264,9 +266,12 @@ const struct sieve_interpreter_extension imap4flags_interpreter_extension = {
  * flags, making the internal or variable flag list indefinitely long
  */
 
-static bool flag_is_valid(const char *flag)
+bool sieve_ext_imap4flags_flag_is_valid(const char *flag)
 {
-	if (*flag == '\\') {
+	if ( *flag == '\0' )
+		return FALSE;
+
+	if ( *flag == '\\' ) {
 		/* System flag */
 		const char *atom = t_str_ucase(flag);
 
@@ -287,20 +292,10 @@ static bool flag_is_valid(const char *flag)
 		 * Syntax (IMAP4rev1, RFC 3501, Section 9. Formal Syntax) :
 		 *  flag-keyword    = atom
 		 *  atom            = 1*ATOM-CHAR
-		 *  ATOM-CHAR       = <any CHAR except atom-specials>
-		 *  atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
-		 *                    quoted-specials / resp-specials
-		 *  CTL             =  %x00-1F / %x7F
-		 *  list-wildcards  = "%" / "*"
-		 *  quoted-specials = DQUOTE / "\"
-		 *  resp-specials   = "]"
 		 */
-
 		p = flag;
 		while ( *p != '\0' ) {
-			if ( *p == '(' || *p == ')' || *p == '{' || *p == ' ' ||
-				*p <= 0x1F || *p == 0x7F || *p == '%' || *p ==  '*' ||
-				*p == '"' || *p == '\\' || *p == ']' )
+			if ( !IS_ATOM_CHAR(*p) )
 				return FALSE;
 			p++;
 		}
@@ -403,6 +398,13 @@ static void ext_imap4flags_iter_delete_last
 
 /* Flag operations */
 
+static string_t *ext_imap4flags_get_flag_variable
+(const struct sieve_runtime_env *renv,
+	const struct sieve_extension *flg_ext,
+	struct sieve_variable_storage *storage,
+	unsigned int var_index)
+	ATTR_NULL(2);
+
 static bool flags_list_flag_exists
 (string_t *flags_list, const char *flag)
 {
@@ -443,7 +445,8 @@ static void flags_list_add_flags
 	ext_imap4flags_iter_init(&flit, flags);
 
 	while ( (flg=ext_imap4flags_iter_get_flag(&flit)) != NULL ) {
-		if ( flag_is_valid(flg) && !flags_list_flag_exists(flags_list, flg) ) {
+		if ( sieve_ext_imap4flags_flag_is_valid(flg) &&
+			!flags_list_flag_exists(flags_list, flg) ) {
 			if ( str_len(flags_list) != 0 )
 				str_append_c(flags_list, ' ');
 			str_append(flags_list, flg);
@@ -478,7 +481,9 @@ static void flags_list_clear_flags
 }
 
 static string_t *ext_imap4flags_get_flag_variable
-(const struct sieve_runtime_env *renv, struct sieve_variable_storage *storage,
+(const struct sieve_runtime_env *renv,
+	const struct sieve_extension *flg_ext,
+	struct sieve_variable_storage *storage,
 	unsigned int var_index)
 {
 	string_t *flags;
@@ -497,18 +502,22 @@ static string_t *ext_imap4flags_get_flag_variable
 		if ( !sieve_variable_get_modifiable(storage, var_index, &flags) )
 			return NULL;
 	} else {
-		flags = _get_flags_string(renv->oprtn->ext, renv->result);
+		i_assert( sieve_extension_is(flg_ext, imap4flags_extension) );
+		flags = _get_flags_string(flg_ext, renv->result);
 	}
 
 	return flags;
 }
 
-int ext_imap4flags_set_flags
-(const struct sieve_runtime_env *renv, struct sieve_variable_storage *storage,
-	unsigned int var_index, struct sieve_stringlist *flags)
+int sieve_ext_imap4flags_set_flags
+(const struct sieve_runtime_env *renv,
+	const struct sieve_extension *flg_ext,
+	struct sieve_variable_storage *storage,
+	unsigned int var_index,
+	struct sieve_stringlist *flags)
 {
 	string_t *cur_flags = ext_imap4flags_get_flag_variable
-		(renv, storage, var_index);
+		(renv, flg_ext, storage, var_index);
 
 	if ( cur_flags != NULL ) {
 		string_t *flags_item;
@@ -530,12 +539,15 @@ int ext_imap4flags_set_flags
 	return SIEVE_EXEC_BIN_CORRUPT;
 }
 
-int ext_imap4flags_add_flags
-(const struct sieve_runtime_env *renv, struct sieve_variable_storage *storage,
-	unsigned int var_index, struct sieve_stringlist *flags)
+int sieve_ext_imap4flags_add_flags
+(const struct sieve_runtime_env *renv,
+	const struct sieve_extension *flg_ext,
+	struct sieve_variable_storage *storage,
+	unsigned int var_index,
+	struct sieve_stringlist *flags)
 {
 	string_t *cur_flags = ext_imap4flags_get_flag_variable
-		(renv, storage, var_index);
+		(renv, flg_ext, storage, var_index);
 
 	if ( cur_flags != NULL ) {
 		string_t *flags_item;
@@ -556,12 +568,15 @@ int ext_imap4flags_add_flags
 	return SIEVE_EXEC_BIN_CORRUPT;
 }
 
-int ext_imap4flags_remove_flags
-(const struct sieve_runtime_env *renv, struct sieve_variable_storage *storage,
-	unsigned int var_index, struct sieve_stringlist *flags)
+int sieve_ext_imap4flags_remove_flags
+(const struct sieve_runtime_env *renv,
+	const struct sieve_extension *flg_ext,
+	struct sieve_variable_storage *storage,
+	unsigned int var_index,
+	struct sieve_stringlist *flags)
 {
 	string_t *cur_flags = ext_imap4flags_get_flag_variable
-		(renv, storage, var_index);
+		(renv, flg_ext, storage, var_index);
 
 	if ( cur_flags != NULL ) {
 		string_t *flags_item;
@@ -689,12 +704,16 @@ static void ext_imap4flags_stringlist_reset
 
 /* Flag access */
 
-struct sieve_stringlist *ext_imap4flags_get_flags
-(const struct sieve_runtime_env *renv, struct sieve_stringlist *flags_list)
+struct sieve_stringlist *sieve_ext_imap4flags_get_flags
+(const struct sieve_runtime_env *renv,
+	const struct sieve_extension *flg_ext,
+	struct sieve_stringlist *flags_list)
 {
-	if ( flags_list == NULL )
+	if ( flags_list == NULL ) {
+		i_assert( sieve_extension_is(flg_ext, imap4flags_extension) );
 		return ext_imap4flags_stringlist_create_single
-			(renv, _get_flags_string(renv->oprtn->ext, renv->result), FALSE);
+			(renv, _get_flags_string(flg_ext, renv->result), FALSE);
+	}
 
 	return ext_imap4flags_stringlist_create(renv, flags_list, TRUE);
 }
