@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2014 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -43,7 +43,7 @@ static inline struct ext_include_generator_context *
 /* Interpreter context */
 
 struct ext_include_interpreter_global {
-	ARRAY_DEFINE(included_scripts, struct sieve_script *);
+	ARRAY(struct sieve_script *) included_scripts;
 
 	struct sieve_variable_scope_binary *var_scope;
 	struct sieve_variable_storage *var_storage;
@@ -454,8 +454,9 @@ struct sieve_variable_storage *ext_include_interpreter_get_global_variables
 
 int ext_include_generate_include
 (const struct sieve_codegen_env *cgenv, struct sieve_command *cmd,
-	enum ext_include_script_location location, struct sieve_script *script,
-	const struct ext_include_script_info **included_r, bool once)
+	enum ext_include_script_location location, 	enum ext_include_flags flags,
+	struct sieve_script *script,
+	const struct ext_include_script_info **included_r)
 {
 	const struct sieve_extension *this_ext = cmd->ext;
 	struct ext_include_context *ext_ctx =
@@ -470,7 +471,7 @@ int ext_include_generate_include
 		ext_include_get_generator_context(this_ext, gentr);
 	struct ext_include_generator_context *pctx;
 	struct sieve_error_handler *ehandler = sieve_generator_error_handler(gentr);
-	const struct ext_include_script_info *included;
+	struct ext_include_script_info *included;
 
 	*included_r = NULL;
 
@@ -489,7 +490,7 @@ int ext_include_generate_include
 	}
 
 	/* Check for circular include */
-	if ( !once ) {
+	if ( (flags & EXT_INCLUDE_FLAG_ONCE) == 0 ) {
 		pctx = ctx;
 		while ( pctx != NULL ) {
 			if ( sieve_script_equals(pctx->script, script) ) {
@@ -515,12 +516,19 @@ int ext_include_generate_include
 	binctx = ext_include_binary_init(this_ext, sbin, cgenv->ast);
 
 	/* Is the script already compiled into the current binary? */
-	if ( !ext_include_binary_script_is_included(binctx, script, &included) )
-	{
-		struct sieve_binary_block *inc_block;
+	included = ext_include_binary_script_get_include_info(binctx, script);
+	if ( included != NULL ) {
+		/* Yes, only update flags */
+		if ( (flags & EXT_INCLUDE_FLAG_OPTIONAL) == 0 )
+			included->flags &= ~EXT_INCLUDE_FLAG_OPTIONAL;
+		if ( (flags & EXT_INCLUDE_FLAG_ONCE) == 0 )
+			included->flags &= ~EXT_INCLUDE_FLAG_ONCE;
+	} else 	{
 		const char *script_name = sieve_script_name(script);
 		enum sieve_compile_flags cpflags = cgenv->flags;
 
+		/* No, include new script */
+
 		/* Check whether include limit is exceeded */
 		if ( ext_include_binary_script_get_count(binctx) >=
 			ext_ctx->max_includes ) {
@@ -530,57 +538,69 @@ int ext_include_generate_include
 	 		return -1;
 		}
 
-		/* No, allocate a new block in the binary and mark the script as included.
+		/* Allocate a new block in the binary and mark the script as included.
 		 */
-		inc_block = sieve_binary_block_create(sbin);
-		included = ext_include_binary_script_include
-			(binctx, script, location, inc_block);
+		if ( !sieve_script_is_open(script) ) {
+			/* Just making an empty entry to mark a missing script */
+			i_assert((flags & EXT_INCLUDE_FLAG_MISSING_AT_UPLOAD) != 0 ||
+				(flags & EXT_INCLUDE_FLAG_OPTIONAL) != 0);
+			included = ext_include_binary_script_include
+				(binctx, location, flags, script, NULL);
+			result = 0;
 
-		/* Parse */
-		if ( (ast = sieve_parse(script, ehandler, NULL)) == NULL ) {
-	 		sieve_command_generate_error(gentr, cmd,
-	 			"failed to parse included script '%s'", str_sanitize(script_name, 80));
-	 		return -1;
-		}
+		}	else {
+			struct sieve_binary_block *inc_block = sieve_binary_block_create(sbin);
 
-		/* Included scripts inherit global variable scope */
-		(void)ext_include_create_ast_context(this_ext, ast, cmd->ast_node->ast);
+			/* Real include */
+			included = ext_include_binary_script_include
+				(binctx, location, flags, script, inc_block);
 
-		if ( location == EXT_INCLUDE_LOCATION_GLOBAL )
+			/* Parse */
+			if ( (ast = sieve_parse(script, ehandler, NULL)) == NULL ) {
+		 		sieve_command_generate_error(gentr, cmd,
+		 			"failed to parse included script '%s'", str_sanitize(script_name, 80));
+		 		return -1;
+			}
+
+			/* Included scripts inherit global variable scope */
+			(void)ext_include_create_ast_context(this_ext, ast, cmd->ast_node->ast);
+
+			if ( location == EXT_INCLUDE_LOCATION_GLOBAL )
 				cpflags &= ~SIEVE_RUNTIME_FLAG_NOGLOBAL;
-		else
+			else
 				cpflags |= SIEVE_RUNTIME_FLAG_NOGLOBAL;
 
-		/* Validate */
-		if ( !sieve_validate(ast, ehandler, cpflags, NULL) ) {
-			sieve_command_generate_error(gentr, cmd,
-				"failed to validate included script '%s'",
-				str_sanitize(script_name, 80));
-	 		sieve_ast_unref(&ast);
-	 		return -1;
-	 	}
+			/* Validate */
+			if ( !sieve_validate(ast, ehandler, cpflags, NULL) ) {
+				sieve_command_generate_error(gentr, cmd,
+					"failed to validate included script '%s'",
+					str_sanitize(script_name, 80));
+		 		sieve_ast_unref(&ast);
+		 		return -1;
+		 	}
 
-		/* Generate
-		 *
-		 * FIXME: It might not be a good idea to recurse code generation for
-		 * included scripts.
-		 */
-	 	subgentr = sieve_generator_create(ast, ehandler, cpflags);
-		ext_include_initialize_generator_context(cmd->ext, subgentr, ctx, script);
+			/* Generate
+			 *
+			 * FIXME: It might not be a good idea to recurse code generation for
+			 * included scripts.
+			 */
+		 	subgentr = sieve_generator_create(ast, ehandler, cpflags);
+			ext_include_initialize_generator_context(cmd->ext, subgentr, ctx, script);
+	
+			if ( sieve_generator_run(subgentr, &inc_block) == NULL ) {
+				sieve_command_generate_error(gentr, cmd,
+					"failed to generate code for included script '%s'",
+					str_sanitize(script_name, 80));
+		 		result = -1;
+			}
 
-		if ( sieve_generator_run(subgentr, &inc_block) == NULL ) {
-			sieve_command_generate_error(gentr, cmd,
-				"failed to generate code for included script '%s'",
-				str_sanitize(script_name, 80));
-	 		result = -1;
+			sieve_generator_free(&subgentr);
+
+			/* Cleanup */
+			sieve_ast_unref(&ast);
 		}
-
-		sieve_generator_free(&subgentr);
-
-		/* Cleanup */
-		sieve_ast_unref(&ast);
 	}
-
+	
 	if ( result > 0 ) *included_r = included;
 
 	return result;
@@ -628,7 +648,8 @@ static bool ext_include_runtime_include_mark
 }
 
 int ext_include_execute_include
-(const struct sieve_runtime_env *renv, unsigned int include_id, bool once)
+(const struct sieve_runtime_env *renv, unsigned int include_id,
+	enum ext_include_flags flags)
 {
 	const struct sieve_extension *this_ext = renv->oprtn->ext;
 	int result = SIEVE_EXEC_OK;
@@ -636,6 +657,7 @@ int ext_include_execute_include
 	const struct ext_include_script_info *included;
 	struct ext_include_binary_context *binctx =
 		ext_include_binary_get_context(this_ext, renv->sbin);
+	bool once = ( (flags & EXT_INCLUDE_FLAG_ONCE) != 0 );
 	unsigned int block_id;
 
 	/* Check for invalid include id (== corrupt binary) */
