@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2014 Pigeonhole authors, see the included COPYING file
  */
 
 /* Extension mailbox
@@ -6,7 +6,7 @@
  *
  * Authors: Stephan Bosch
  * Specification: RFC 5490
- * Implementation: almost full; acl support is missing for mailboxexists
+ * Implementation: full
  * Status: testing
  *
  */
@@ -32,10 +32,8 @@ static bool ext_mailbox_validator_load
 (const struct sieve_extension *ext, struct sieve_validator *valdtr);
 
 const struct sieve_extension_def mailbox_extension = {
-	"mailbox",
-	NULL, NULL,
-	ext_mailbox_validator_load,
-	NULL, NULL, NULL, NULL, NULL,
+	.name = "mailbox",
+	.validator_load = ext_mailbox_validator_load,
 	SIEVE_EXT_DEFINE_OPERATION(mailboxexists_operation),
 	SIEVE_EXT_DEFINE_OPERAND(mailbox_create_operand)
 };
