@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2014 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -6,6 +6,7 @@
 #include "str.h"
 #include "ioloop.h"
 #include "str-sanitize.h"
+#include "ostream.h"
 #include "message-date.h"
 #include "mail-storage.h"
 
@@ -145,22 +146,17 @@ static int act_notify_check_duplicate
 static void act_notify_print
 	(const struct sieve_action *action, const struct sieve_result_print_env *rpenv,
 		bool *keep);
-static bool act_notify_commit
+static int act_notify_commit
 	(const struct sieve_action *action,	const struct sieve_action_exec_env *aenv,
 		void *tr_context, bool *keep);
 
 /* Action object */
 
 const struct sieve_action_def act_notify_old = {
-	"notify",
-	0,
-	NULL,
-	act_notify_check_duplicate,
-	NULL,
-	act_notify_print,
-	NULL, NULL,
-	act_notify_commit,
-	NULL
+	.name = "notify",
+	.check_duplicate = act_notify_check_duplicate,
+	.print = act_notify_print,
+	.commit = act_notify_commit
 };
 
 /*
@@ -487,8 +483,9 @@ static int cmd_notify_operation_execute
 		/* Process message */
 
 		out_message = t_str_new(1024);
-		ext_notify_construct_message
-			(renv, (message == NULL ? NULL : str_c(message)), out_message);
+		if ( (ret=ext_notify_construct_message
+			(renv, (message == NULL ? NULL : str_c(message)), out_message)) <= 0 )
+			return ret;
 		act->message = p_strdup(pool, str_c(out_message));
 
 		/* Normalize and verify all :options addresses */
@@ -688,10 +685,12 @@ static bool act_notify_send
 {
 	const struct sieve_script_env *senv = aenv->scriptenv;
 	const struct ext_notify_recipient *recipients;
-	void *smtp_handle;
+	struct sieve_smtp_context *sctx;
 	unsigned int count, i;
-	FILE *f;
-	const char *outmsgid;
+	struct ostream *output;
+	string_t *msg, *to, *all;
+	const char *outmsgid, *error;
+	int ret;
 
 	/* Get recipients */
 	recipients = array_get(&act->recipients, &count);
@@ -702,108 +701,140 @@ static bool act_notify_send
 	}
 
 	/* Just to be sure */
-	if ( senv->smtp_open == NULL || senv->smtp_close == NULL ) {
+	if ( !sieve_smtp_available(senv) ) {
 		sieve_result_global_warning(aenv,
 			"notify action has no means to send mail");
 		return TRUE;
 	}
 
-	/* Send message to all recipients */
+	/* Compose common headers */
+	msg = t_str_new(512);
+	rfc2822_header_write(msg, "X-Sieve", SIEVE_IMPLEMENTATION);
+	rfc2822_header_write(msg, "Date", message_date_create(ioloop_time));
+
+	/* Set importance */
+	switch ( act->importance ) {
+	case 1:
+		rfc2822_header_write(msg, "X-Priority", "1 (Highest)");
+		rfc2822_header_write(msg, "Importance", "High");
+		break;
+	case 3:
+		rfc2822_header_write(msg, "X-Priority", "5 (Lowest)");
+		rfc2822_header_write(msg, "Importance", "Low");
+		break;
+	case 2:
+	default:
+		rfc2822_header_write(msg, "X-Priority", "3 (Normal)");
+		rfc2822_header_write(msg, "Importance", "Normal");
+		break;
+	}
+
+	rfc2822_header_printf(msg, "From",
+		"Postmaster <%s>", senv->postmaster_address);
+
+	rfc2822_header_write(msg, "Subject", "[SIEVE] New mail notification");
+
+	rfc2822_header_write(msg, "Auto-Submitted", "auto-generated (notify)");
+	rfc2822_header_write(msg, "Precedence", "bulk");
+
+	rfc2822_header_write(msg, "MIME-Version", "1.0");
+	if (contains_8bit(act->message)) {
+		rfc2822_header_write(msg,
+			"Content-Type", "text/plain; charset=utf-8");
+		rfc2822_header_write(msg, "Content-Transfer-Encoding", "8bit");
+	} else {
+		rfc2822_header_write(msg,
+			"Content-Type", "text/plain; charset=us-ascii");
+		rfc2822_header_write(msg, "Content-Transfer-Encoding", "7bit");
+	}
+
+	outmsgid = sieve_message_get_new_id(aenv->svinst);
+	rfc2822_header_write(msg, "Message-ID", outmsgid);
+
+	if ( sieve_message_get_sender(aenv->msgctx) != NULL )
+		sctx = sieve_smtp_start(senv, senv->postmaster_address);
+	else
+		sctx = sieve_smtp_start(senv, NULL);
+
+	/* Add all recipients (and compose To header field) */
+	to = t_str_new(128);
+	all = t_str_new(256);
 	for ( i = 0; i < count; i++ ) {
-
-		if ( sieve_message_get_sender(aenv->msgctx) != NULL )
-			smtp_handle = sieve_smtp_open
-				(senv, recipients[i].normalized, senv->postmaster_address, &f);
-		else
-			smtp_handle = sieve_smtp_open
-				(senv, recipients[i].normalized, NULL, &f);
-
-		outmsgid = sieve_message_get_new_id(aenv->svinst);
-
-		rfc2822_header_field_write(f, "X-Sieve", SIEVE_IMPLEMENTATION);
-		rfc2822_header_field_write(f, "Message-ID", outmsgid);
-		rfc2822_header_field_write(f, "Date", message_date_create(ioloop_time));
-
-		/* Set importance */
-		switch ( act->importance ) {
-		case 1:
-			rfc2822_header_field_write(f, "X-Priority", "1 (Highest)");
-			rfc2822_header_field_write(f, "Importance", "High");
-			break;
-		case 3:
-			rfc2822_header_field_write(f, "X-Priority", "5 (Lowest)");
-			rfc2822_header_field_write(f, "Importance", "Low");
-			break;
-		case 2:
-		default:
-			rfc2822_header_field_write(f, "X-Priority", "3 (Normal)");
-			rfc2822_header_field_write(f, "Importance", "Normal");
-			break;
+		sieve_smtp_add_rcpt(sctx, recipients[i].normalized);
+		if ( i > 0 )
+			str_append(to, ", ");
+		str_append(to, recipients[i].full);
+		if ( i < 3) {
+			if ( i > 0 )
+				str_append(all, ", ");
+			str_append_c(all, '<');
+			str_append(all, str_sanitize(recipients[i].normalized, 256));
+			str_append_c(all, '>');
+		} else if (i == 3) {
+			str_printfa(all, ", ... (%u total)", count);
 		}
+	}
 
-		rfc2822_header_field_printf(f, "From", "%s",
-			t_strdup_printf("Postmaster <%s>", senv->postmaster_address));
+	rfc2822_header_write(msg, "To", str_c(to));
 
-		rfc2822_header_field_printf(f, "To", "%s", recipients[i].full);
+	/* Generate message body */
+	str_printfa(msg, "\r\n%s\r\n", act->message);
 
-		rfc2822_header_field_write(f, "Subject", "[SIEVE] New mail notification");
+	output = sieve_smtp_send(sctx);
+	o_stream_send(output, str_data(msg), str_len(msg));
 
-		rfc2822_header_field_write(f, "Auto-Submitted", "auto-generated (notify)");
-		rfc2822_header_field_write(f, "Precedence", "bulk");
-
-		if (contains_8bit(act->message)) {
-			rfc2822_header_field_write(f, "MIME-Version", "1.0");
-			rfc2822_header_field_write(f,
-				"Content-Type", "text/plain; charset=UTF-8");
-			rfc2822_header_field_write(f, "Content-Transfer-Encoding", "8bit");
-		}
-
-		/* Generate message body */
-		fprintf(f, "\r\n");
-		fprintf(f, "%s\r\n", act->message);
-
-		if ( sieve_smtp_close(senv, smtp_handle) ) {
-			sieve_result_global_log(aenv,
-				"sent mail notification to <%s>",
-				str_sanitize(recipients[i].normalized, 80));
-		} else {
+	if ( (ret=sieve_smtp_finish(sctx, &error)) <= 0 ) {
+		if (ret < 0) {
 			sieve_result_global_error(aenv,
-				"failed to send mail notification to <%s> "
-				"(refer to system log for more information)",
-				str_sanitize(recipients[i].normalized, 80));
+				"failed to send mail notification to %s: %s (temporary failure)",
+				str_c(all),	str_sanitize(error, 512));
+		} else {
+			sieve_result_global_log_error(aenv,
+				"failed to send mail notification to %s: %s (permanent failure)",
+				str_c(all),	str_sanitize(error, 512));
 		}
+	} else {
+		sieve_result_global_log(aenv,
+			"sent mail notification to %s", str_c(all));
 	}
 
 	return TRUE;
 }
 
-static bool act_notify_commit
+static int act_notify_commit
 (const struct sieve_action *action, const struct sieve_action_exec_env *aenv,
 	void *tr_context ATTR_UNUSED, bool *keep ATTR_UNUSED)
 {
+	struct mail *mail = aenv->msgdata->mail;
 	const struct ext_notify_action *act =
 		(const struct ext_notify_action *) action->context;
 	const struct sieve_message_data *msgdata = aenv->msgdata;
-	const char *const *headers;
+	const char *const *hdsp;
+	bool result;
 
 	/* Is the message an automatic reply ? */
-	if ( mail_get_headers
-		(msgdata->mail, "auto-submitted", &headers) >= 0 ) {
-		const char *const *hdsp = headers;
-
-		/* Theoretically multiple headers could exist, so lets make sure */
-		while ( *hdsp != NULL ) {
-			if ( strcasecmp(*hdsp, "no") != 0 ) {
-				sieve_result_global_log(aenv,
-					"not sending notification for auto-submitted message from <%s>",
-					str_sanitize(msgdata->return_path, 128));
-					return TRUE;
-			}
-			hdsp++;
-		}
+	if ( mail_get_headers(mail, "auto-submitted", &hdsp) < 0 ) {
+		return sieve_result_mail_error(aenv, mail,
+			"notify action: "
+			"failed to read `auto-submitted' header field");
 	}
 
-	return act_notify_send(aenv, act);
+	/* Theoretically multiple headers could exist, so lets make sure */
+	while ( *hdsp != NULL ) {
+		if ( strcasecmp(*hdsp, "no") != 0 ) {
+			sieve_result_global_log(aenv,
+				"not sending notification for auto-submitted message from <%s>",
+				str_sanitize(msgdata->return_path, 128));
+			return SIEVE_EXEC_OK;
+		}
+		hdsp++;
+	}
+
+	T_BEGIN {
+		result = act_notify_send(aenv, act);
+	} T_END;
+
+	return ( result ? SIEVE_EXEC_OK : SIEVE_EXEC_FAILURE );
 }
 
 
