@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2017 Pigeonhole authors, see the included COPYING file
  */
 
 #ifndef __EXT_NOTIFY_COMMON_H
@@ -62,7 +62,7 @@ struct ext_notify_action {
  * Message construct
  */
 
-void ext_notify_construct_message
+int ext_notify_construct_message
 	(const struct sieve_runtime_env *renv, const char *msg_format,
 		string_t *out_msg);
 
