@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #include "sieve-common.h"
@@ -11,13 +11,12 @@
  */
 
 static const struct sieve_extension_objects ext_match_types =
-    SIEVE_EXT_DEFINE_MATCH_TYPE(regex_match_type);
+	SIEVE_EXT_DEFINE_MATCH_TYPE(regex_match_type);
 
 const struct sieve_operand_def regex_match_type_operand = {
-    "regex match",
-    &regex_extension,
-    0,
-    &sieve_match_type_operand_class,
-    &ext_match_types
+	.name = "regex match",
+	.ext_def = &regex_extension,
+	.class = &sieve_match_type_operand_class,
+	.interface = &ext_match_types
 };
 
