@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #ifndef __EXT_SPAMVIRUSTEST_COMMON_H
@@ -24,9 +24,9 @@ void ext_spamvirustest_unload(const struct sieve_extension *ext);
 extern const struct sieve_command_def spamtest_test;
 extern const struct sieve_command_def virustest_test;
 
-const char *ext_spamvirustest_get_value
+int ext_spamvirustest_get_value
 (const struct sieve_runtime_env *renv, const struct sieve_extension *ext,
-	 bool percent);
+	 bool percent, const char **value_r);
 
 /*
  * Operations
