@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -7,9 +7,11 @@
 #include "md5.h"
 #include "hostpid.h"
 #include "str-sanitize.h"
+#include "ostream.h"
 #include "message-address.h"
 #include "message-date.h"
 #include "ioloop.h"
+#include "mail-storage.h"
 
 #include "rfc2822.h"
 
@@ -189,22 +191,19 @@ int act_vacation_check_conflict
 static void act_vacation_print
 	(const struct sieve_action *action,
 		const struct sieve_result_print_env *rpenv, bool *keep);
-static bool act_vacation_commit
+static int act_vacation_commit
 	(const struct sieve_action *action,	const struct sieve_action_exec_env *aenv,
 		void *tr_context, bool *keep);
 
 /* Action object */
 
 const struct sieve_action_def act_vacation = {
-	"vacation",
-	SIEVE_ACTFLAG_SENDS_RESPONSE,
-	NULL,
-	act_vacation_check_duplicate,
-	act_vacation_check_conflict,
-	act_vacation_print,
-	NULL, NULL,
-	act_vacation_commit,
-	NULL
+	.name = "vacation",
+	.flags = SIEVE_ACTFLAG_SENDS_RESPONSE,
+	.check_duplicate = act_vacation_check_duplicate,
+	.check_conflict = act_vacation_check_conflict,
+	.print = act_vacation_print,
+	.commit = act_vacation_commit
 };
 
 /* Action context information */
@@ -693,7 +692,7 @@ static int ext_vacation_operation_execute
 
 	/* Normalize all addresses */
 	if ( addresses != NULL ) {
-		ARRAY_DEFINE(norm_addresses, const char *);
+		ARRAY(const char *) norm_addresses;
 		string_t *raw_address;
 		int ret;
 
@@ -894,17 +893,17 @@ static bool _contains_8bit(const char *text)
 	return FALSE;
 }
 
-static bool act_vacation_send
+static int act_vacation_send
 (const struct sieve_action_exec_env *aenv, struct act_vacation_context *ctx,
-	const char *reply_to, const char *reply_from)
+ 	const char *reply_to, const char *reply_from, const char *smtp_from)
 {
 	const struct sieve_message_data *msgdata = aenv->msgdata;
 	const struct sieve_script_env *senv = aenv->scriptenv;
-	void *smtp_handle;
-	FILE *f;
- 	const char *outmsgid;
+	struct sieve_smtp_context *sctx;
+	struct ostream *output;
+	string_t *msg;
  	const char *const *headers;
-	const char *subject;
+	const char *outmsgid, *subject, *error;
 	int ret;
 
 	/* Check smpt functions just to be sure */
@@ -912,14 +911,19 @@ static bool act_vacation_send
 	if ( !sieve_smtp_available(senv) ) {
 		sieve_result_global_warning(aenv,
 			"vacation action has no means to send mail");
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
 	/* Make sure we have a subject for our reply */
 
 	if ( ctx->subject == NULL || *(ctx->subject) == '\0' ) {
 		if ( mail_get_headers_utf8
-			(msgdata->mail, "subject", &headers) >= 0 && headers[0] != NULL ) {
+			(msgdata->mail, "subject", &headers) < 0 ) {
+			return sieve_result_mail_error(aenv, msgdata->mail,
+				"vacation action: "
+				"failed to read header field `subject'");
+		}
+		if ( headers[0] != NULL ) {
 			subject = t_strconcat("Auto: ", headers[0], NULL);
 		}	else {
 			subject = "Automated reply";
@@ -932,72 +936,85 @@ static bool act_vacation_send
 
 	/* Open smtp session */
 
-	smtp_handle = sieve_smtp_open(senv, reply_to, NULL, &f);
+	sctx = sieve_smtp_start_single(senv, reply_to, smtp_from, &output);
+
 	outmsgid = sieve_message_get_new_id(aenv->svinst);
 
 	/* Produce a proper reply */
 
-	rfc2822_header_field_write(f, "X-Sieve", SIEVE_IMPLEMENTATION);
-	rfc2822_header_field_write(f, "Message-ID", outmsgid);
-	rfc2822_header_field_write(f, "Date", message_date_create(ioloop_time));
+	msg = t_str_new(512);
+	rfc2822_header_write(msg, "X-Sieve", SIEVE_IMPLEMENTATION);
+	rfc2822_header_write(msg, "Message-ID", outmsgid);
+	rfc2822_header_write(msg, "Date", message_date_create(ioloop_time));
 
 	if ( ctx->from != NULL && *(ctx->from) != '\0' )
-		rfc2822_header_field_utf8_printf(f, "From", "%s", ctx->from);
+		rfc2822_header_utf8_printf(msg, "From", "%s", ctx->from);
 	else if ( reply_from != NULL )
-		rfc2822_header_field_printf(f, "From", "<%s>", reply_from);
+		rfc2822_header_printf(msg, "From", "<%s>", reply_from);
 	else
-		rfc2822_header_field_printf(f, "From", "Postmaster <%s>", senv->postmaster_address);
+		rfc2822_header_printf(msg, "From", "Postmaster <%s>", senv->postmaster_address);
 
-	/* FIXME: If From header of message has same address, we should use that in
-	 * stead properly include the phrase part.
+	/* FIXME: If From header of message has same address, we should use that
+	 * instead to properly include the phrase part.
 	 */
-	rfc2822_header_field_printf(f, "To", "<%s>", reply_to);
+	rfc2822_header_printf(msg, "To", "<%s>", reply_to);
 
 	if ( _contains_8bit(subject) )
-		rfc2822_header_field_utf8_printf(f, "Subject", "%s", subject);
+		rfc2822_header_utf8_printf(msg, "Subject", "%s", subject);
 	else
-		rfc2822_header_field_printf(f, "Subject", "%s", subject);
+		rfc2822_header_printf(msg, "Subject", "%s", subject);
 
 	/* Compose proper in-reply-to and references headers */
 
-	ret = mail_get_headers
-		(aenv->msgdata->mail, "references", &headers);
+	if ( mail_get_headers
+		(msgdata->mail, "references", &headers) ) {
+		return sieve_result_mail_error(aenv, msgdata->mail,
+			"vacation action: "
+			"failed to read header field `references'");
+	}
 
 	if ( msgdata->id != NULL ) {
-		rfc2822_header_field_write(f, "In-Reply-To", msgdata->id);
+		rfc2822_header_write(msg, "In-Reply-To", msgdata->id);
 
-		if ( ret >= 0 && headers[0] != NULL )
-			rfc2822_header_field_write
-				(f, "References", t_strconcat(headers[0], " ", msgdata->id, NULL));
+		if ( headers[0] != NULL )
+			rfc2822_header_write
+				(msg, "References", t_strconcat(headers[0], " ", msgdata->id, NULL));
 		else
-			rfc2822_header_field_write(f, "References", msgdata->id);
-	} else if ( ret >= 0 && headers[0] != NULL ) {
-		rfc2822_header_field_write(f, "References", headers[0]);
+			rfc2822_header_write(msg, "References", msgdata->id);
+	} else if ( headers[0] != NULL ) {
+		rfc2822_header_write(msg, "References", headers[0]);
 	}
 
-	rfc2822_header_field_write(f, "Auto-Submitted", "auto-replied (vacation)");
-	rfc2822_header_field_write(f, "Precedence", "bulk");
+	rfc2822_header_write(msg, "Auto-Submitted", "auto-replied (vacation)");
+	rfc2822_header_write(msg, "Precedence", "bulk");
 
-	rfc2822_header_field_write(f, "MIME-Version", "1.0");
+	rfc2822_header_write(msg, "MIME-Version", "1.0");
 
 	if ( !ctx->mime ) {
-		rfc2822_header_field_write(f, "Content-Type", "text/plain; charset=utf-8");
-		rfc2822_header_field_write(f, "Content-Transfer-Encoding", "8bit");
-		fprintf(f, "\r\n");
+		rfc2822_header_write(msg, "Content-Type", "text/plain; charset=utf-8");
+		rfc2822_header_write(msg, "Content-Transfer-Encoding", "8bit");
+		str_append(msg, "\r\n");
 	}
 
-	fprintf(f, "%s\r\n", ctx->reason);
+	str_printfa(msg, "%s\r\n", ctx->reason);
+  o_stream_send(output, str_data(msg), str_len(msg));
 
 	/* Close smtp session */
-	if ( !sieve_smtp_close(senv, smtp_handle) ) {
-		sieve_result_global_error(aenv,
-			"failed to send vacation response to <%s> "
-			"(refer to server log for more information)",
-			str_sanitize(reply_to, 128));
-		return FALSE;
+	if ( (ret=sieve_smtp_finish(sctx, &error)) <= 0 ) {
+		if ( ret < 0 ) {
+			sieve_result_global_error(aenv,
+				"failed to send vacation response to <%s>: %s (temporary error)",
+				str_sanitize(reply_to, 256), str_sanitize(error, 512));
+		} else {
+			sieve_result_global_log_error(aenv,
+				"failed to send vacation response to <%s>: %s (permanent error)",
+				str_sanitize(reply_to, 256), str_sanitize(error, 512));
+		}
+		/* This error will be ignored in the end */
+		return SIEVE_EXEC_FAILURE;
 	}
 
-	return TRUE;
+	return SIEVE_EXEC_OK;
 }
 
 static void act_vacation_hash
@@ -1014,14 +1031,13 @@ static void act_vacation_hash
 	md5_final(&ctx, hash_r);
 }
 
-static bool act_vacation_commit
+static int act_vacation_commit
 (const struct sieve_action *action, const struct sieve_action_exec_env *aenv,
 	void *tr_context ATTR_UNUSED, bool *keep ATTR_UNUSED)
 {
 	const struct sieve_extension *ext = action->ext;
 	const struct ext_vacation_config *config =
 		(const struct ext_vacation_config *) ext->context;
-	const struct sieve_message_data *msgdata = aenv->msgdata;
 	const struct sieve_script_env *senv = aenv->scriptenv;
 	struct act_vacation_context *ctx =
 		(struct act_vacation_context *) action->context;
@@ -1029,23 +1045,23 @@ static bool act_vacation_commit
 	struct mail *mail = sieve_message_get_mail(aenv->msgctx);
 	const char *sender = sieve_message_get_sender(aenv->msgctx);
 	const char *recipient = sieve_message_get_final_recipient(aenv->msgctx);
-	const char *const *hdsp;
-	const char *const *headers;
-	const char *reply_from = NULL;
+	const char *const *hdsp, *const *headers;
+	const char *reply_from = NULL, *orig_recipient = NULL, *smtp_from = NULL;
+	int ret;
 
 	/* Is the recipient unset?
 	 */
 	if ( recipient == NULL ) {
 		sieve_result_global_warning
 			(aenv, "vacation action aborted: envelope recipient is <>");
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
 	/* Is the return path unset ?
 	 */
 	if ( sender == NULL ) {
 		sieve_result_global_log(aenv, "discarded vacation reply to <>");
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
 	/* Are we perhaps trying to respond to ourselves ?
@@ -1054,7 +1070,7 @@ static bool act_vacation_commit
 		sieve_result_global_log(aenv,
 			"discarded vacation reply to own address <%s>",
 			str_sanitize(sender, 128));
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
 	/* Are we perhaps trying to respond to one of our alternative :addresses?
@@ -1068,7 +1084,7 @@ static bool act_vacation_commit
 					"discarded vacation reply to own address <%s> "
 					"(as specified using :addresses argument)",
 					str_sanitize(sender, 128));
-				return TRUE;
+				return SIEVE_EXEC_OK;
 			}
 			alt_address++;
 		}
@@ -1083,55 +1099,66 @@ static bool act_vacation_commit
 			sieve_result_global_log(aenv,
 				"discarded duplicate vacation response to <%s>",
 				str_sanitize(sender, 128));
-			return TRUE;
+			return SIEVE_EXEC_OK;
 		}
 	}
 
 	/* Are we trying to respond to a mailing list ? */
 	hdsp = _list_headers;
 	while ( *hdsp != NULL ) {
-		if ( mail_get_headers
-			(mail, *hdsp, &headers) >= 0 && headers[0] != NULL ) {
+		if ( mail_get_headers(mail, *hdsp, &headers) < 0 ) {
+			return sieve_result_mail_error(aenv, mail,
+				"vacation action: "
+				"failed to read header field `%s'", *hdsp);
+		}
+
+		if ( headers[0] != NULL ) {
 			/* Yes, bail out */
 			sieve_result_global_log(aenv,
 				"discarding vacation response to mailinglist recipient <%s>",
 				str_sanitize(sender, 128));
-			return TRUE;
+			return SIEVE_EXEC_OK;
 		}
 		hdsp++;
 	}
 
 	/* Is the message that we are replying to an automatic reply ? */
 	if ( mail_get_headers
-		(msgdata->mail, "auto-submitted", &headers) >= 0 ) {
-		/* Theoretically multiple headers could exist, so lets make sure */
-		hdsp = headers;
-		while ( *hdsp != NULL ) {
-			if ( strcasecmp(*hdsp, "no") != 0 ) {
-				sieve_result_global_log(aenv,
-					"discarding vacation response to auto-submitted message from <%s>",
- 					str_sanitize(sender, 128));
-					return TRUE;
-			}
-			hdsp++;
+		(mail, "auto-submitted", &headers) < 0 ) {
+		return sieve_result_mail_error(aenv, mail,
+			"vacation action: "
+			"failed to read header field `auto-submitted'");
+	}
+	/* Theoretically multiple headers could exist, so lets make sure */
+	hdsp = headers;
+	while ( *hdsp != NULL ) {
+		if ( strcasecmp(*hdsp, "no") != 0 ) {
+			sieve_result_global_log(aenv,
+				"discarding vacation response to auto-submitted message from <%s>",
+				str_sanitize(sender, 128));
+				return SIEVE_EXEC_OK;
 		}
+		hdsp++;
 	}
 
 	/* Check for the (non-standard) precedence header */
 	if ( mail_get_headers
-		(mail, "precedence", &headers) >= 0 ) {
-		/* Theoretically multiple headers could exist, so lets make sure */
-		hdsp = headers;
-		while ( *hdsp != NULL ) {
-			if ( strcasecmp(*hdsp, "junk") == 0 || strcasecmp(*hdsp, "bulk") == 0 ||
-				strcasecmp(*hdsp, "list") == 0 ) {
-				sieve_result_global_log(aenv,
-					"discarding vacation response to precedence=%s message from <%s>",
-					*hdsp, str_sanitize(sender, 128));
-					return TRUE;
-			}
-			hdsp++;
+		(mail, "precedence", &headers) < 0 ) {
+		return sieve_result_mail_error(aenv, mail,
+			"vacation action: "
+			"failed to read header field `precedence'");
+	}
+	/* Theoretically multiple headers could exist, so lets make sure */
+	hdsp = headers;
+	while ( *hdsp != NULL ) {
+		if ( strcasecmp(*hdsp, "junk") == 0 || strcasecmp(*hdsp, "bulk") == 0 ||
+			strcasecmp(*hdsp, "list") == 0 ) {
+			sieve_result_global_log(aenv,
+				"discarding vacation response to precedence=%s message from <%s>",
+				*hdsp, str_sanitize(sender, 128));
+				return SIEVE_EXEC_OK;
 		}
+		hdsp++;
 	}
 
 	/* Do not reply to system addresses */
@@ -1139,55 +1166,72 @@ static bool act_vacation_commit
 		sieve_result_global_log(aenv,
 			"not sending vacation response to system address <%s>",
 			str_sanitize(sender, 128));
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
+	/* Fetch original recipient if necessary */
+	if ( config->use_original_recipient  )
+		orig_recipient = sieve_message_get_orig_recipient(aenv->msgctx);
+
 	/* Is the original message directly addressed to the user or the addresses
 	 * specified using the :addresses tag?
 	 */
-	if ( !config->dont_check_recipient ) {
-		const char *orig_recipient = NULL;
-
-		if ( config->use_original_recipient  )
-			orig_recipient = sieve_message_get_orig_recipient(aenv->msgctx);
-
-		hdsp = _my_address_headers;
-		while ( *hdsp != NULL ) {
-			if ( mail_get_headers
-				(mail, *hdsp, &headers) >= 0 && headers[0] != NULL ) {
-
-				if ( _contains_my_address(headers, recipient) ) {
-					reply_from = recipient;
-					break;
-				}
-
-				if ( orig_recipient != NULL && _contains_my_address(headers, orig_recipient) ) {
-					reply_from = orig_recipient;
-					break;
-				}
-
-				if ( ctx->addresses != NULL ) {
-					bool found = FALSE;
-					const char * const *my_address = ctx->addresses;
-
-					while ( !found && *my_address != NULL ) {
-						if ( (found=_contains_my_address(headers, *my_address)) )
-							reply_from = *my_address;
-						my_address++;
-					}
-
-					if ( found ) break;
-				}
-			}
-			hdsp++;
+	hdsp = _my_address_headers;
+	while ( *hdsp != NULL ) {
+		if ( mail_get_headers(mail, *hdsp, &headers) < 0 ) {
+			return sieve_result_mail_error(aenv, mail,
+				"vacation action: "
+				"failed to read header field `%s'", *hdsp);
 		}
+		if ( headers[0] != NULL ) {
 
+			/* Final recipient directly listed in headers? */
+			if ( _contains_my_address(headers, recipient) ) {
+				reply_from = recipient;
+				smtp_from = recipient;
+				break;
+			}
 
-		/* My address not found in the headers; we got an implicit delivery */
-		if ( *hdsp == NULL ) {
+			/* Original recipient directly listed in headers? */
+			if ( orig_recipient != NULL &&
+				_contains_my_address(headers, orig_recipient) ) {
+				reply_from = orig_recipient;
+				smtp_from = orig_recipient;
+				break;
+			}
+
+			/* User-provided :addresses listed in headers? */
+			if ( ctx->addresses != NULL ) {
+				bool found = FALSE;
+				const char * const *my_address = ctx->addresses;
+
+				while ( !found && *my_address != NULL ) {
+					if ( (found=_contains_my_address(headers, *my_address)) ) {
+						reply_from = *my_address;
+						/* Avoid letting user determine SMTP sender directly */
+						smtp_from =
+							( orig_recipient == NULL ? recipient : orig_recipient );
+					}
+					my_address++;
+				}
+
+				if ( found ) break;
+			}
+		}
+		hdsp++;
+	}
+
+	/* My address not found in the headers; we got an implicit delivery */
+	if ( *hdsp == NULL ) {
+		if ( config->dont_check_recipient ) {
+			/* Send reply from envelope recipient address */
+			reply_from = recipient;
+			smtp_from = recipient;
+
+		} else {
 			const char *original_recipient = "";
 
-			/* No, bail out */
+			/* Bail out */
 
 			if ( config->use_original_recipient ) {
 				original_recipient = t_strdup_printf("original-recipient=<%s>, ",
@@ -1200,15 +1244,20 @@ static bool act_vacation_commit
 				"no known (envelope) recipient address found in message headers "
 				"(recipient=<%s>, %sand%s additional `:addresses' are specified)",
 				str_sanitize(recipient, 128), original_recipient,
-				(ctx->addresses == NULL ? " no" : ""));
+				(ctx->addresses == NULL || *ctx->addresses == NULL ? " no" : ""));
 
-			return TRUE;
+			return SIEVE_EXEC_OK;
 		}
 	}
 
 	/* Send the message */
 
-	if ( act_vacation_send(aenv, ctx, sender, reply_from) ) {
+	T_BEGIN {
+		ret = act_vacation_send(aenv, ctx, sender, reply_from,
+			(config->send_from_recipient ? smtp_from : NULL));
+	} T_END;
+
+	if ( ret == SIEVE_EXEC_OK ) {
 		sieve_number_t seconds;
 
 		sieve_result_global_log(aenv, "sent vacation response to <%s>",
@@ -1226,11 +1275,13 @@ static bool act_vacation_commit
 			sieve_action_duplicate_mark
 				(senv, dupl_hash, sizeof(dupl_hash), ioloop_time + seconds);
 		}
-
-		return TRUE;
 	}
 
-	return TRUE;
+	if ( ret == SIEVE_EXEC_TEMP_FAILURE )
+		return SIEVE_EXEC_TEMP_FAILURE;
+
+	/* Ignore all other errors */
+	return SIEVE_EXEC_OK;
 }
 
 
