@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2017 Pigeonhole authors, see the included COPYING file
  */
 
 #ifndef __EXT_DEBUG_COMMON_H
@@ -8,7 +8,7 @@
  * Extensions
  */
 
-extern const struct sieve_extension_def debug_extension;
+extern const struct sieve_extension_def vnd_debug_extension;
 
 /*
  * Commands
