@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -10,6 +10,7 @@
 #include "unichar.h"
 #include "istream.h"
 #include "istream-header-filter.h"
+#include "ostream.h"
 #include "mail-deliver.h"
 #include "mail-storage.h"
 #include "message-date.h"
@@ -36,29 +37,6 @@
 const struct sieve_operand_class sieve_side_effect_operand_class =
 	{ "SIDE-EFFECT" };
 
-bool sieve_opr_side_effect_read
-(const struct sieve_runtime_env *renv, sieve_size_t *address,
-	struct sieve_side_effect *seffect)
-{
-	const struct sieve_side_effect_def *sdef;
-
-	seffect->context = NULL;
-
-	if ( !sieve_opr_object_read
-		(renv, &sieve_side_effect_operand_class, address, &seffect->object) )
-		return FALSE;
-
-	sdef = seffect->def =
-		(const struct sieve_side_effect_def *) seffect->object.def;
-
-	if ( sdef->read_context != NULL &&
-		!sdef->read_context(seffect, renv, address, &seffect->context) ) {
-		return FALSE;
-	}
-
-	return TRUE;
-}
-
 bool sieve_opr_side_effect_dump
 (const struct sieve_dumptime_env *denv, sieve_size_t *address)
 {
@@ -83,6 +61,30 @@ bool sieve_opr_side_effect_dump
 	return TRUE;
 }
 
+int sieve_opr_side_effect_read
+(const struct sieve_runtime_env *renv, sieve_size_t *address,
+	struct sieve_side_effect *seffect)
+{
+	const struct sieve_side_effect_def *sdef;
+	int ret;
+
+	seffect->context = NULL;
+
+	if ( !sieve_opr_object_read
+		(renv, &sieve_side_effect_operand_class, address, &seffect->object) )
+		return SIEVE_EXEC_BIN_CORRUPT;
+
+	sdef = seffect->def =
+		(const struct sieve_side_effect_def *) seffect->object.def;
+
+	if ( sdef->read_context != NULL && (ret=sdef->read_context
+		(seffect, renv, address, &seffect->context)) <= 0 ) {
+		return ret;
+	}
+
+	return SIEVE_EXEC_OK;
+}
+
 /*
  * Optional operands
  */
@@ -189,13 +191,13 @@ static void act_store_print
 	(const struct sieve_action *action,
 		const struct sieve_result_print_env *rpenv, bool *keep);
 
-static bool act_store_start
+static int act_store_start
 	(const struct sieve_action *action,
 		const struct sieve_action_exec_env *aenv, void **tr_context);
-static bool act_store_execute
+static int act_store_execute
 	(const struct sieve_action *action,
 		const struct sieve_action_exec_env *aenv, void *tr_context);
-static bool act_store_commit
+static int act_store_commit
 	(const struct sieve_action *action,
 		const struct sieve_action_exec_env *aenv, void *tr_context, bool *keep);
 static void act_store_rollback
@@ -205,16 +207,17 @@ static void act_store_rollback
 /* Action object */
 
 const struct sieve_action_def act_store = {
-	"store",
-	SIEVE_ACTFLAG_TRIES_DELIVER,
-	act_store_equals,
-	act_store_check_duplicate,
-	NULL,
-	act_store_print,
-	act_store_start,
-	act_store_execute,
-	act_store_commit,
-	act_store_rollback,
+	.name = "store",
+	.flags =
+		SIEVE_ACTFLAG_TRIES_DELIVER | 
+		SIEVE_ACTFLAG_MAIL_STORAGE,
+	.equals = act_store_equals,
+	.check_duplicate = act_store_check_duplicate,
+	.print = act_store_print,
+	.start = act_store_start,
+	.execute = act_store_execute,
+	.commit = act_store_commit,
+	.rollback = act_store_rollback,
 };
 
 /* API */
@@ -282,16 +285,6 @@ void sieve_act_store_add_flags
 	trans->flags_altered = TRUE;
 }
 
-void sieve_act_store_get_storage_error
-(const struct sieve_action_exec_env *aenv, struct act_store_transaction *trans)
-{
-	pool_t pool = sieve_result_pool(aenv->result);
-
-	trans->error = p_strdup(pool,
-		mail_storage_get_last_error(mailbox_get_storage(trans->box),
-		&trans->error_code));
-}
-
 /* Equality */
 
 static bool act_store_equals
@@ -352,6 +345,17 @@ static void act_store_print
 
 /* Action implementation */
 
+void sieve_act_store_get_storage_error
+(const struct sieve_action_exec_env *aenv,
+	struct act_store_transaction *trans)
+{
+	pool_t pool = sieve_result_pool(aenv->result);
+
+	trans->error = p_strdup(pool,
+		mail_storage_get_last_error(mailbox_get_storage(trans->box),
+		&trans->error_code));
+}
+
 static bool act_store_mailbox_open
 (const struct sieve_action_exec_env *aenv, const char *mailbox,
 	struct mailbox **box_r, enum mail_error *error_code_r, const char **error_r)
@@ -383,7 +387,7 @@ static bool act_store_mailbox_open
 	return TRUE;
 }
 
-static bool act_store_start
+static int act_store_start
 (const struct sieve_action *action,
 	const struct sieve_action_exec_env *aenv, void **tr_context)
 {
@@ -408,7 +412,8 @@ static bool act_store_start
 	 * to NULL. This implementation will then skip actually storing the message.
 	 */
 	if ( senv->user != NULL ) {
-		if ( !act_store_mailbox_open(aenv, ctx->mailbox, &box, &error_code, &error) ) {
+		if ( !act_store_mailbox_open
+			(aenv, ctx->mailbox, &box, &error_code, &error) ) {
 			open_failed = TRUE;
 		}
 	} else {
@@ -433,8 +438,17 @@ static bool act_store_start
 
 	*tr_context = (void *)trans;
 
-	return ( trans->error_code == MAIL_ERROR_NONE ||
-		trans->error_code == MAIL_ERROR_NOTFOUND );
+	switch ( trans->error_code ) {
+	case MAIL_ERROR_NONE:
+	case MAIL_ERROR_NOTFOUND:
+		return SIEVE_EXEC_OK;
+	case MAIL_ERROR_TEMP:
+		return SIEVE_EXEC_TEMP_FAILURE;
+	default:
+		break;
+	}
+	
+	return SIEVE_EXEC_FAILURE;
 }
 
 static struct mail_keywords *act_store_keywords_create
@@ -450,7 +464,6 @@ static struct mail_keywords *act_store_keywords_create
 		(void)array_append_space(keywords);
 		kwds = array_idx(keywords, 0);
 
-		/* FIXME: Do we need to clear duplicates? */
 		if ( mailbox_keywords_create(box, kwds, &box_keywords) < 0) {
 			sieve_result_error(aenv, "invalid keywords set for stored message");
 			return NULL;
@@ -460,7 +473,7 @@ static struct mail_keywords *act_store_keywords_create
 	return box_keywords;
 }
 
-static bool act_store_execute
+static int act_store_execute
 (const struct sieve_action *action,
 	const struct sieve_action_exec_env *aenv, void *tr_context)
 {
@@ -468,27 +481,47 @@ static bool act_store_execute
 		(struct act_store_transaction *) tr_context;
 	struct mail *mail =	( action->mail != NULL ?
 		action->mail : aenv->msgdata->mail );
-	struct mail *real_mail = mail_get_real_mail(mail);
 	struct mail_save_context *save_ctx;
 	struct mail_keywords *keywords = NULL;
-	bool result = TRUE;
+	bool backends_equal = FALSE;
+	int status = SIEVE_EXEC_OK;
 
 	/* Verify transaction */
-	if ( trans == NULL ) return FALSE;
+	if ( trans == NULL ) return SIEVE_EXEC_FAILURE;
 
 	/* Check whether we need to do anything */
-	if ( trans->disabled ) return TRUE;
+	if ( trans->disabled ) return SIEVE_EXEC_OK;
 
 	/* Exit early if mailbox is not available */
-	if ( trans->box == NULL || trans->error_code != MAIL_ERROR_NONE )
-		return FALSE;
+	if ( trans->box == NULL )
+		return SIEVE_EXEC_FAILURE;
+
+	/* Exit early if transaction already failed */
+ 	switch ( trans->error_code ) {
+	case MAIL_ERROR_NONE:
+		break;
+	case MAIL_ERROR_TEMP:
+		return SIEVE_EXEC_TEMP_FAILURE;
+	default:
+		return SIEVE_EXEC_FAILURE;
+	}
+
 
 	/* If the message originates from the target mailbox, only update the flags
 	 * and keywords (if not read-only)
 	 */
-	if ( mailbox_backends_equal(trans->box, mail->box) ||
-		(real_mail != mail && mailbox_backends_equal(trans->box, real_mail->box)) )
-		{
+	if ( mailbox_backends_equal(trans->box, mail->box) ) {
+		backends_equal = TRUE;
+	} else {
+		struct mail *real_mail;
+
+		if ( mail_get_backend_mail(mail, &real_mail) < 0 )
+			return SIEVE_EXEC_FAILURE;
+		if ( real_mail != mail &&
+			mailbox_backends_equal(trans->box, real_mail->box) )
+			backends_equal = TRUE;
+	}
+	if (backends_equal) {
 		trans->redundant = TRUE;
 
 		if ( trans->flags_altered && !mailbox_is_readonly(mail->box) ) {
@@ -503,7 +536,7 @@ static bool act_store_execute
 			mail_update_flags(mail, MODIFY_REPLACE, trans->flags);
 		}
 
-		return TRUE;
+		return SIEVE_EXEC_OK;
 
 	/* If the message is modified, only store it in the source mailbox when it is
 	 * not opened read-only. Mail structs of modified messages have their own
@@ -515,7 +548,7 @@ static bool act_store_execute
 		&& ( mailbox_backends_equal(trans->box, aenv->msgdata->mail->box) ) ) {
 
 		trans->redundant = TRUE;
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
 	/* Mark attempt to store in default mailbox */
@@ -548,9 +581,10 @@ static bool act_store_execute
 		mailbox_save_copy_flags(save_ctx, mail);
 	}
 
-	if ( mailbox_copy(&save_ctx, mail) < 0 ) {
+	if ( mailbox_save_using_mail(&save_ctx, mail) < 0 ) {
 		sieve_act_store_get_storage_error(aenv, trans);
- 		result = FALSE;
+		status = ( trans->error_code == MAIL_ERROR_TEMP ?
+			SIEVE_EXEC_TEMP_FAILURE : SIEVE_EXEC_FAILURE );
 	}
 
 	/* Deallocate keywords */
@@ -558,7 +592,7 @@ static bool act_store_execute
  		mailbox_keywords_unref(&keywords);
  	}
 
-	return result;
+	return status;
 }
 
 static void act_store_log_status
@@ -596,20 +630,26 @@ static void act_store_log_status
 		const char *errstr;
 		enum mail_error error_code;
 
-		if ( trans->error != NULL ) {
-			errstr = trans->error;
-			error_code = trans->error_code;
-		} else {
-			errstr = mail_storage_get_last_error
-				(mailbox_get_storage(trans->box), &error_code);
-		}
+		if ( trans->error == NULL )
+			sieve_act_store_get_storage_error(aenv, trans);
 
-		if ( error_code != MAIL_ERROR_NOTFOUND && error_code != MAIL_ERROR_PARAMS )
-			{
-			sieve_result_global_error(aenv, "failed to store into mailbox %s: %s",
+		errstr = trans->error;
+		error_code = trans->error_code;
+
+		if ( error_code == MAIL_ERROR_NOQUOTA ) {
+			/* Never log quota problems as error in global log */
+			sieve_result_global_log_error(aenv,
+				"failed to store into mailbox %s: %s",
+				mailbox_name, errstr);
+		} else if ( error_code == MAIL_ERROR_NOTFOUND ||
+			error_code == MAIL_ERROR_PARAMS ||
+			error_code == MAIL_ERROR_PERM ) {
+			sieve_result_error(aenv,
+				"failed to store into mailbox %s: %s",
 				mailbox_name, errstr);
 		} else {
-			sieve_result_error(aenv, "failed to store into mailbox %s: %s",
+			sieve_result_global_error(aenv,
+				"failed to store into mailbox %s: %s",
 				mailbox_name, errstr);
 		}
 
@@ -625,7 +665,7 @@ static void act_store_log_status
 	}
 }
 
-static bool act_store_commit
+static int act_store_commit
 (const struct sieve_action *action ATTR_UNUSED,
 	const struct sieve_action_exec_env *aenv, void *tr_context, bool *keep)
 {
@@ -634,7 +674,7 @@ static bool act_store_commit
 	bool status = TRUE;
 
 	/* Verify transaction */
-	if ( trans == NULL ) return FALSE;
+	if ( trans == NULL ) return SIEVE_EXEC_FAILURE;
 
 	/* Check whether we need to do anything */
 	if ( trans->disabled ) {
@@ -642,14 +682,14 @@ static bool act_store_commit
 		*keep = FALSE;
 		if ( trans->box != NULL )
 			mailbox_free(&trans->box);
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	} else if ( trans->redundant ) {
 		act_store_log_status(trans, aenv, FALSE, status);
 		aenv->exec_status->keep_original = TRUE;
 		aenv->exec_status->message_saved = TRUE;
 		if ( trans->box != NULL )
 			mailbox_free(&trans->box);
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
 	/* Mark attempt to use storage. Can only get here when all previous actions
@@ -667,6 +707,8 @@ static bool act_store_commit
 	/* Note the fact that the message was stored at least once */
 	if ( status )
 		aenv->exec_status->message_saved = TRUE;
+	else
+		aenv->exec_status->store_failed = TRUE;
 
 	/* Log our status */
 	act_store_log_status(trans, aenv, FALSE, status);
@@ -678,7 +720,11 @@ static bool act_store_commit
 	if ( trans->box != NULL )
 		mailbox_free(&trans->box);
 
-	return status;
+	if (status)
+		return SIEVE_EXEC_OK;
+
+	return ( trans->error_code == MAIL_ERROR_TEMP ?
+			SIEVE_EXEC_TEMP_FAILURE : SIEVE_EXEC_FAILURE );
 }
 
 static void act_store_rollback
@@ -688,6 +734,13 @@ static void act_store_rollback
 	struct act_store_transaction *trans =
 		(struct act_store_transaction *) tr_context;
 
+	i_assert( trans->box != NULL );
+
+	if (!success) {
+		aenv->exec_status->last_storage = mailbox_get_storage(trans->box);
+		aenv->exec_status->store_failed = TRUE;
+	}
+
 	/* Log status */
 	act_store_log_status(trans, aenv, TRUE, success);
 
@@ -700,8 +753,7 @@ static void act_store_rollback
 		mailbox_transaction_rollback(&trans->mail_trans);
 
 	/* Close the mailbox */
-	if ( trans->box != NULL )
-		mailbox_free(&trans->box);
+	mailbox_free(&trans->box);
 }
 
 /*
@@ -735,9 +787,18 @@ void sieve_action_duplicate_mark
 	senv->duplicate_mark(senv, id, id_size, time);
 }
 
+void sieve_action_duplicate_flush
+(const struct sieve_script_env *senv)
+{
+	if ( senv->duplicate_flush == NULL )
+		return;
+	senv->duplicate_flush(senv);
+}
+
+
 /* Rejecting the mail */
 
-static bool sieve_action_do_reject_mail
+static int sieve_action_do_reject_mail
 (const struct sieve_action_exec_env *aenv, const char *sender,
 	const char *recipient, const char *reason)
 {
@@ -745,107 +806,118 @@ static bool sieve_action_do_reject_mail
 	const struct sieve_script_env *senv = aenv->scriptenv;
 	const struct sieve_message_data *msgdata = aenv->msgdata;
 	struct istream *input;
-	void *smtp_handle;
-	struct message_size hdr_size;
-	FILE *f;
-	const char *new_msgid, *boundary;
-	const unsigned char *data;
-	const char *header;
-	size_t size;
+	struct ostream *output;
+	struct sieve_smtp_context *sctx;
+	const char *new_msgid, *boundary, *error;
+	const char *orig_recipient =
+		sieve_message_get_orig_recipient(aenv->msgctx);
+  string_t *hdr;
 	int ret;
 
+	sctx = sieve_smtp_start_single(senv, sender, NULL, &output);
+
 	/* Just to be sure */
-	if ( !sieve_smtp_available(senv) ) {
+	if ( sctx == NULL ) {
 		sieve_result_global_warning
 			(aenv, "reject action has no means to send mail");
-		return TRUE;
+		return SIEVE_EXEC_OK;
 	}
 
-	smtp_handle = sieve_smtp_open(senv, sender, NULL, &f);
-
 	new_msgid = sieve_message_get_new_id(svinst);
 	boundary = t_strdup_printf("%s/%s", my_pid, svinst->hostname);
 
-	rfc2822_header_field_write(f, "X-Sieve", SIEVE_IMPLEMENTATION);
-	rfc2822_header_field_write(f, "Message-ID", new_msgid);
-	rfc2822_header_field_write(f, "Date", message_date_create(ioloop_time));
-	rfc2822_header_field_printf(f, "From", "Mail Delivery Subsystem <%s>",
+  hdr = t_str_new(512);
+	rfc2822_header_write(hdr, "X-Sieve", SIEVE_IMPLEMENTATION);
+	rfc2822_header_write(hdr, "Message-ID", new_msgid);
+	rfc2822_header_write(hdr, "Date", message_date_create(ioloop_time));
+	rfc2822_header_printf(hdr, "From", "Mail Delivery Subsystem <%s>",
 		senv->postmaster_address);
-	rfc2822_header_field_printf(f, "To", "<%s>", sender);
-	rfc2822_header_field_write(f, "Subject", "Automatically rejected mail");
-	rfc2822_header_field_write(f, "Auto-Submitted", "auto-replied (rejected)");
-	rfc2822_header_field_write(f, "Precedence", "bulk");
+	rfc2822_header_printf(hdr, "To", "<%s>", sender);
+	rfc2822_header_write(hdr, "Subject", "Automatically rejected mail");
+	rfc2822_header_write(hdr, "Auto-Submitted", "auto-replied (rejected)");
+	rfc2822_header_write(hdr, "Precedence", "bulk");
 
-	rfc2822_header_field_write(f, "MIME-Version", "1.0");
-	rfc2822_header_field_printf(f, "Content-Type",
-		"multipart/report; report-type=disposition-notification;\n"
+	rfc2822_header_write(hdr, "MIME-Version", "1.0");
+	rfc2822_header_printf(hdr, "Content-Type",
+		"multipart/report; report-type=disposition-notification;\r\n"
 		"boundary=\"%s\"", boundary);
 
-	fprintf(f, "\r\nThis is a MIME-encapsulated message\r\n\r\n");
+	str_append(hdr, "\r\nThis is a MIME-encapsulated message\r\n\r\n");
 
 	/* Human readable status report */
-	fprintf(f, "--%s\r\n", boundary);
-	fprintf(f, "Content-Type: text/plain; charset=utf-8\r\n");
-	fprintf(f, "Content-Disposition: inline\r\n");
-	fprintf(f, "Content-Transfer-Encoding: 8bit\r\n\r\n");
+	str_printfa(hdr, "--%s\r\n", boundary);
+	rfc2822_header_write(hdr, "Content-Type", "text/plain; charset=utf-8");
+	rfc2822_header_write(hdr, "Content-Disposition", "inline");
+	rfc2822_header_write(hdr, "Content-Transfer-Encoding", "8bit");
 
-	fprintf(f, "Your message to <%s> was automatically rejected:\r\n"
+	str_printfa(hdr, "\r\nYour message to <%s> was automatically rejected:\r\n"
 		"%s\r\n", recipient, reason);
 
 	/* MDN status report */
-	fprintf(f, "--%s\r\n"
-		"Content-Type: message/disposition-notification\r\n\r\n", boundary);
-	fprintf(f, "Reporting-UA: %s; Dovecot Mail Delivery Agent\r\n",
-		svinst->hostname);
-	if (mail_get_first_header(msgdata->mail, "Original-Recipient", &header) > 0)
-		fprintf(f, "Original-Recipient: rfc822; %s\r\n", header);
-	fprintf(f, "Final-Recipient: rfc822; %s\r\n", recipient);
+	str_printfa(hdr, "--%s\r\n", boundary);
+	rfc2822_header_write(hdr, "Content-Type", "message/disposition-notification");
+	str_append(hdr, "\r\n");
+	rfc2822_header_write(hdr, "Reporting-UA: %s; Dovecot Mail Delivery Agent",
+		svinst->hostname);	
+	if ( orig_recipient != NULL ) {
+		rfc2822_header_printf
+			(hdr, "Original-Recipient", "rfc822; %s", orig_recipient);
+	}
+	rfc2822_header_printf(hdr, "Final-Recipient", "rfc822; %s", recipient);
 
 	if ( msgdata->id != NULL )
-		fprintf(f, "Original-Message-ID: %s\r\n", msgdata->id);
-	fprintf(f, "Disposition: "
-		"automatic-action/MDN-sent-automatically; deleted\r\n");
-	fprintf(f, "\r\n");
+		rfc2822_header_write(hdr, "Original-Message-ID", msgdata->id);
+	rfc2822_header_write(hdr, "Disposition",
+		"automatic-action/MDN-sent-automatically; deleted");
+	str_append(hdr, "\r\n");
 
 	/* original message's headers */
-	fprintf(f, "--%s\r\nContent-Type: message/rfc822\r\n\r\n", boundary);
+	str_printfa(hdr, "--%s\r\n", boundary);
+	rfc2822_header_write(hdr, "Content-Type", "message/rfc822");
+	str_append(hdr, "\r\n");
+	o_stream_send(output, str_data(hdr), str_len(hdr));
 
-	if (mail_get_stream(msgdata->mail, &hdr_size, NULL, &input) == 0) {
-		/* Note: If you add more headers, they need to be sorted.
-		 * We'll drop Content-Type because we're not including the message
-		 * body, and having a multipart Content-Type may confuse some
-		 * MIME parsers when they don't see the message boundaries.
-		 */
-		static const char *const exclude_headers[] = {
-			"Content-Type"
-		};
+	if (mail_get_hdr_stream(msgdata->mail, NULL, &input) == 0) {
+    /* Note: If you add more headers, they need to be sorted.
+       We'll drop Content-Type because we're not including the message
+       body, and having a multipart Content-Type may confuse some
+       MIME parsers when they don't see the message boundaries. */
+    static const char *const exclude_headers[] = {
+	    "Content-Type"
+    };
 
-		input = i_stream_create_header_filter(input,
-			HEADER_FILTER_EXCLUDE | HEADER_FILTER_NO_CR | HEADER_FILTER_HIDE_BODY,
-			exclude_headers, N_ELEMENTS(exclude_headers),
-			null_header_filter_callback, NULL);
+    input = i_stream_create_header_filter(input,
+    		HEADER_FILTER_EXCLUDE | HEADER_FILTER_NO_CR |
+		HEADER_FILTER_HIDE_BODY, exclude_headers,
+		N_ELEMENTS(exclude_headers),
+		*null_header_filter_callback, (void *)NULL);
 
-		while ((ret = i_stream_read_data(input, &data, &size, 0)) > 0) {
-			if (fwrite(data, size, 1, f) == 0)
-				break;
-			i_stream_skip(input, size);
+    ret = o_stream_send_istream(output, input);
+    i_stream_unref(&input);
+
+    i_assert(ret != 0);
+  }
+
+  str_truncate(hdr, 0);
+  str_printfa(hdr, "\r\n\r\n--%s--\r\n", boundary);
+  o_stream_send(output, str_data(hdr), str_len(hdr));
+
+	if ( (ret=sieve_smtp_finish(sctx, &error)) <= 0 ) {
+		if ( ret < 0 ) {
+			sieve_result_global_error(aenv,
+				"failed to send rejection message to <%s>: %s "
+				"(temporary failure)",
+				str_sanitize(sender, 256), str_sanitize(error, 512));
+		} else {
+			sieve_result_global_log_error(aenv,
+				"failed to send rejection message to <%s>: %s "
+				"(permanent failure)",
+				str_sanitize(sender, 256), str_sanitize(error, 512));
 		}
-		i_stream_unref(&input);
-
-		i_assert(ret != 0);
+		return SIEVE_EXEC_FAILURE;
 	}
 
-	fprintf(f, "\r\n\r\n--%s--\r\n", boundary);
-
-	if ( !sieve_smtp_close(senv, smtp_handle) ) {
-		sieve_result_global_error(aenv,
-			"failed to send rejection message to <%s> "
-			"(refer to server log for more information)",
-			str_sanitize(sender, 80));
-		return FALSE;
-	}
-
-	return TRUE;
+	return SIEVE_EXEC_OK;
 }
 
 bool sieve_action_reject_mail
@@ -853,14 +925,31 @@ bool sieve_action_reject_mail
 	const char *sender, const char *recipient, const char *reason)
 {
 	const struct sieve_script_env *senv = aenv->scriptenv;
+	bool result;
 
-	if ( senv->reject_mail != NULL ) {
-		return ( senv->reject_mail(senv, recipient, reason) >= 0 );
+	T_BEGIN {
+		if ( senv->reject_mail != NULL ) {
+			result =
+				( senv->reject_mail(senv, recipient, reason) >= 0 );
+		} else {
+			result = sieve_action_do_reject_mail(aenv, sender, recipient, reason);
+		}
+	} T_END;
+
+	return result;
+}
+
+/*
+ * Mailbox
+ */
+
+bool sieve_mailbox_check_name(const char *mailbox, const char **error_r)
+{
+	if ( !uni_utf8_str_is_valid(mailbox) ) {
+		*error_r = "mailbox is utf-8";
+		return FALSE;
 	}
-
-	return sieve_action_do_reject_mail(aenv, sender, recipient, reason);
+	return TRUE;
 }
 
 
-
-
