@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2013 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2017 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -198,23 +198,12 @@ static bool _sieve_binary_save
 (struct sieve_binary *sbin, struct ostream *stream)
 {
 	struct sieve_binary_header header;
-	struct sieve_binary_extension_reg *const *regs;
 	struct sieve_binary_block *ext_block;
 	unsigned int ext_count, blk_count, i;
 	uoff_t block_index;
 
 	blk_count = sieve_binary_block_count(sbin);
 
-	/* Signal all extensions to finish generating their blocks */
-
-	regs = array_get(&sbin->extensions, &ext_count);
-	for ( i = 0; i < ext_count; i++ ) {
-		const struct sieve_binary_extension *binext = regs[i]->binext;
-
-		if ( binext != NULL && binext->binary_save != NULL )
-			binext->binary_save(regs[i]->extension, sbin, regs[i]->context);
-	}
-
 	/* Create header */
 
 	header.magic = SIEVE_BINARY_MAGIC;
@@ -233,11 +222,11 @@ static bool _sieve_binary_save
 		sizeof(struct sieve_binary_block_index) * blk_count, &block_index) )
 		return FALSE;
 
-	/* Create block containing all used extensions
-	 *   FIXME: Per-extension this should also store binary version numbers.
-	 */
+	/* Create block containing all used extensions */
+
 	ext_block = sieve_binary_block_get(sbin, SBIN_SYSBLOCK_EXTENSIONS);
 	i_assert( ext_block != NULL );
+	sieve_binary_block_clear(ext_block);
 
 	ext_count = array_count(&sbin->linked_extensions);
 	sieve_binary_emit_unsigned(ext_block, ext_count);
@@ -248,6 +237,8 @@ static bool _sieve_binary_save
 
 		sieve_binary_emit_cstring
 			(ext_block, sieve_extension_name((*ext)->extension));
+		sieve_binary_emit_unsigned
+			(ext_block, sieve_extension_version((*ext)->extension));
 		sieve_binary_emit_unsigned(ext_block, (*ext)->block_id);
 	}
 
@@ -275,6 +266,8 @@ int sieve_binary_save
 	int result, fd;
 	string_t *temp_path;
 	struct ostream *stream;
+	struct sieve_binary_extension_reg *const *regs;
+	unsigned int ext_count, i;
 
 	if ( error_r != NULL )
 		*error_r = SIEVE_ERROR_NONE;
@@ -299,24 +292,36 @@ int sieve_binary_save
 				"binary save: failed to create temporary file: %s",
 				eacces_error_get_creating("open", str_c(temp_path)));
 			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_NO_PERM;
+				*error_r = SIEVE_ERROR_NO_PERMISSION;
 		} else {
 			sieve_sys_error(sbin->svinst,
 				"binary save: failed to create temporary file: open(%s) failed: %m",
 				str_c(temp_path));
 			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_TEMP_FAIL;
+				*error_r = SIEVE_ERROR_TEMP_FAILURE;
 		}
 		return -1;
 	}
 
+	/* Signal all extensions that we're about to save the binary */
+	regs = array_get(&sbin->extensions, &ext_count);
+	for ( i = 0; i < ext_count; i++ ) {
+		const struct sieve_binary_extension *binext = regs[i]->binext;
+
+		if ( binext != NULL && binext->binary_pre_save != NULL &&
+			!binext->binary_pre_save
+				(regs[i]->extension, sbin, regs[i]->context, error_r)) {
+			return -1;
+		}
+	}
+
 	/* Save binary */
 	result = 1;
 	stream = o_stream_create_fd(fd, 0, FALSE);
 	if ( !_sieve_binary_save(sbin, stream) ) {
 		result = -1;
 		if ( error_r != NULL )
-			*error_r = SIEVE_ERROR_TEMP_FAIL;
+			*error_r = SIEVE_ERROR_TEMP_FAILURE;
 	}
 	o_stream_destroy(&stream);
 
@@ -328,17 +333,17 @@ int sieve_binary_save
 	}
 
 	/* Replace any original binary atomically */
-	if ( result && (rename(str_c(temp_path), path) < 0) ) {
+	if ( result > 0 && (rename(str_c(temp_path), path) < 0) ) {
 		if ( errno == EACCES ) {
 			sieve_sys_error(sbin->svinst, "binary save: failed to save binary: %s",
 				eacces_error_get_creating("rename", path));
 			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_NO_PERM;
+				*error_r = SIEVE_ERROR_NO_PERMISSION;
 		} else {
 			sieve_sys_error(sbin->svinst, "binary save: failed to save binary: "
 				"rename(%s, %s) failed: %m", str_c(temp_path), path);
 			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_TEMP_FAIL;
+				*error_r = SIEVE_ERROR_TEMP_FAILURE;
 		}
 		result = -1;
 	}
@@ -351,9 +356,28 @@ int sieve_binary_save
 				str_c(temp_path));
 		}
 	} else {
-		if ( sbin->path == NULL ) {
+		if ( sbin->path == NULL )
 			sbin->path = p_strdup(sbin->pool, path);
+
+		/* Signal all extensions that we successfully saved the binary */
+		regs = array_get(&sbin->extensions, &ext_count);
+		for ( i = 0; i < ext_count; i++ ) {
+			const struct sieve_binary_extension *binext = regs[i]->binext;
+
+			if ( binext != NULL && binext->binary_post_save != NULL &&
+				!binext->binary_post_save
+					(regs[i]->extension, sbin, regs[i]->context, error_r)) {
+				result = -1;
+				break;
+			}
 		}
+
+		if ( result < 0 && unlink(path) < 0 && errno != ENOENT ) {
+			sieve_sys_error(sbin->svinst,
+				"binary save: failed to clean up after error: "
+				"unlink(%s) failed: %m", path);
+		}
+		sbin->path = NULL;
 	}
 
 	return result;
@@ -384,13 +408,13 @@ bool sieve_binary_file_open
 			sieve_sys_error(svinst, "binary open: failed to open: %s",
 				eacces_error_get("open", path));
 			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_NO_PERM;
+				*error_r = SIEVE_ERROR_NO_PERMISSION;
 			break;
 		default:
 			sieve_sys_error(svinst, "binary open: failed to open: "
 				"open(%s) failed: %m", path);
 			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_TEMP_FAIL;
+				*error_r = SIEVE_ERROR_TEMP_FAILURE;
 			break;
 		}
 		return FALSE;
@@ -719,38 +743,48 @@ static bool _read_block_index_record
 	return TRUE;
 }
 
-static bool _read_extensions(struct sieve_binary_block *sblock)
+static int _read_extensions(struct sieve_binary_block *sblock)
 {
 	struct sieve_binary *sbin = sblock->sbin;
 	sieve_size_t offset = 0;
 	unsigned int i, count;
-	bool result = TRUE;
+	int result = 1;
 
 	if ( !sieve_binary_read_unsigned(sblock, &offset, &count) )
-		return FALSE;
+		return -1;
 
-	for ( i = 0; result && i < count; i++ ) {
+	for ( i = 0; result > 0 && i < count; i++ ) {
 		T_BEGIN {
 			string_t *extension;
 			const struct sieve_extension *ext;
+			unsigned int version;
 
 			if ( sieve_binary_read_string(sblock, &offset, &extension) ) {
 				ext = sieve_extension_get_by_name(sbin->svinst, str_c(extension));
 
 				if ( ext == NULL ) {
 					sieve_sys_error(sbin->svinst,
-						"binary open: binary %s requires unknown extension '%s'",
+						"binary open: binary %s requires unknown extension `%s'",
 						sbin->path, str_sanitize(str_c(extension), 128));
-					result = FALSE;
+					result = 0;
 				} else {
 					struct sieve_binary_extension_reg *ereg = NULL;
 
 					(void) sieve_binary_extension_register(sbin, ext, &ereg);
-					if ( !sieve_binary_read_unsigned(sblock, &offset, &ereg->block_id) )
-						result = FALSE;
+					if ( !sieve_binary_read_unsigned(sblock, &offset, &version) ||
+						!sieve_binary_read_unsigned(sblock, &offset, &ereg->block_id) ) {
+						result = -1;
+					} else if ( !sieve_extension_version_is(ext, version) ) {
+						sieve_sys_debug(sbin->svinst,
+							"binary open: binary %s was compiled with different version "
+							"of the `%s' extension (compiled v%d, expected v%d;"
+							"automatically fixed when re-compiled)", sbin->path,
+							sieve_extension_name(ext), version, sieve_extension_version(ext));
+						result = 0;
+					}
 				}
 			}	else
-				result = FALSE;
+				result = -1;
 		} T_END;
 	}
 
@@ -764,6 +798,7 @@ static bool _sieve_binary_open(struct sieve_binary *sbin)
 	const struct sieve_binary_header *header;
 	struct sieve_binary_block *ext_block;
 	unsigned int i, blk_count;
+	int ret;
 
 	/* Verify header */
 
@@ -838,10 +873,12 @@ static bool _sieve_binary_open(struct sieve_binary *sbin)
 		ext_block = sieve_binary_block_get(sbin, SBIN_SYSBLOCK_EXTENSIONS);
 		if ( ext_block == NULL ) {
 			result = FALSE;
-		} else if ( !_read_extensions(ext_block) ) {
-			sieve_sys_error(sbin->svinst,
-				"binary open: binary %s is corrupt: failed to load extension block",
-				sbin->path);
+		} else if ( (ret=_read_extensions(ext_block)) <= 0 ) {
+			if ( ret < 0 ) {
+				sieve_sys_error(sbin->svinst,
+					"binary open: binary %s is corrupt: failed to load extension block",
+					sbin->path);
+			}
 			result = FALSE;
 		}
 	} T_END;
