@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2013 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2014 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -77,20 +77,29 @@ void sieve_direct_verror
 	unsigned int flags, const char *location, const char *fmt, va_list args)
 {
 	if ( (flags & SIEVE_ERROR_FLAG_GLOBAL) != 0 &&
-		(ehandler == NULL || ehandler->parent == NULL) &&
-		svinst->system_ehandler != ehandler &&
-		svinst->system_ehandler->verror != NULL ) {
-		va_list args_copy;
+		(ehandler == NULL || ehandler->parent == NULL)) {
+		i_assert(svinst->system_ehandler != NULL);
+		if (svinst->system_ehandler != ehandler ||
+			(flags & SIEVE_ERROR_FLAG_GLOBAL_MAX_INFO) != 0) {
+			va_list args_copy;
 
-		VA_COPY(args_copy, args);
+			VA_COPY(args_copy, args);
 
-		if ( (flags & SIEVE_ERROR_FLAG_GLOBAL_MAX_INFO) != 0 ) {
-			svinst->system_ehandler->vinfo
-				(svinst->system_ehandler, 0, location, fmt, args_copy);
-		} else {
-			svinst->system_ehandler->verror
-				(svinst->system_ehandler, 0, location, fmt, args_copy);
+			if ( (flags & SIEVE_ERROR_FLAG_GLOBAL_MAX_INFO) != 0 ) {
+				if (svinst->system_ehandler->vinfo != NULL ) {
+					svinst->system_ehandler->vinfo
+						(svinst->system_ehandler, 0, location, fmt, args_copy);
+				}
+			} else {
+				if ( svinst->system_ehandler->verror != NULL ) {
+					svinst->system_ehandler->verror
+						(svinst->system_ehandler, 0, location, fmt, args_copy);
+				}
+			}
+			va_end(args_copy);
 		}
+		if (svinst->system_ehandler == ehandler)
+			return;
 	}
 
 	if ( ehandler == NULL )
@@ -110,20 +119,29 @@ void sieve_direct_vwarning
 	unsigned int flags, const char *location, const char *fmt, va_list args)
 {
 	if ( (flags & SIEVE_ERROR_FLAG_GLOBAL) != 0 &&
-		(ehandler == NULL || ehandler->parent == NULL) &&
-		svinst->system_ehandler != ehandler &&
-		svinst->system_ehandler->vwarning != NULL ) {
-		va_list args_copy;
+		(ehandler == NULL || ehandler->parent == NULL)) {
+		i_assert(svinst->system_ehandler != NULL);
+		if (svinst->system_ehandler != ehandler ||
+			(flags & SIEVE_ERROR_FLAG_GLOBAL_MAX_INFO) != 0) {
+			va_list args_copy;
 
-		VA_COPY(args_copy, args);
+			VA_COPY(args_copy, args);
 
-		if ( (flags & SIEVE_ERROR_FLAG_GLOBAL_MAX_INFO) != 0 ) {
-			svinst->system_ehandler->vinfo
-				(svinst->system_ehandler, 0, location, fmt, args_copy);
-		} else {
-			svinst->system_ehandler->vwarning
-				(svinst->system_ehandler, 0, location, fmt, args_copy);
+			if ( (flags & SIEVE_ERROR_FLAG_GLOBAL_MAX_INFO) != 0 ) {
+				if (svinst->system_ehandler->vinfo != NULL ) {
+					svinst->system_ehandler->vinfo
+						(svinst->system_ehandler, 0, location, fmt, args_copy);
+				}
+			} else {
+				if ( svinst->system_ehandler->vwarning != NULL ) {
+					svinst->system_ehandler->vwarning
+						(svinst->system_ehandler, 0, location, fmt, args_copy);
+				}
+			}
+			va_end(args_copy);
 		}
+		if (svinst->system_ehandler == ehandler)
+			return;
 	}
 
 	if ( ehandler == NULL )
@@ -142,14 +160,16 @@ void sieve_direct_vinfo
 {
 	if ( (flags & SIEVE_ERROR_FLAG_GLOBAL) != 0 &&
 		(ehandler == NULL || ehandler->parent == NULL) &&
-		svinst->system_ehandler != ehandler &&
-		svinst->system_ehandler->vinfo != NULL ) {
-		va_list args_copy;
+		svinst->system_ehandler != ehandler) {
+		i_assert(svinst->system_ehandler != NULL);
+		if (svinst->system_ehandler->vinfo != NULL ) {
+			va_list args_copy;
 
-		VA_COPY(args_copy, args);
-
-		svinst->system_ehandler->vinfo
-			(svinst->system_ehandler, 0, location, fmt, args_copy);
+			VA_COPY(args_copy, args);
+			svinst->system_ehandler->vinfo
+				(svinst->system_ehandler, 0, location, fmt, args_copy);
+			va_end(args_copy);
+		}
 	}
 
 	if ( ehandler == NULL )
@@ -167,14 +187,16 @@ void sieve_direct_vdebug
 {
 	if ( (flags & SIEVE_ERROR_FLAG_GLOBAL) != 0 &&
 		(ehandler == NULL || ehandler->parent == NULL) &&
-		svinst->system_ehandler != ehandler &&
-		svinst->system_ehandler->vdebug != NULL ) {
-		va_list args_copy;
+		svinst->system_ehandler != ehandler) {
+		i_assert(svinst->system_ehandler != NULL);
+		if (svinst->system_ehandler->vdebug != NULL ) {
+			va_list args_copy;
 
-		VA_COPY(args_copy, args);
-
-		svinst->system_ehandler->vdebug
-			(svinst->system_ehandler, 0, location, fmt, args_copy);
+			VA_COPY(args_copy, args);
+			svinst->system_ehandler->vdebug
+				(svinst->system_ehandler, 0, location, fmt, args_copy);
+			va_end(args_copy);
+		}
 	}
 
 	if ( ehandler == NULL )
@@ -446,9 +468,6 @@ void sieve_vcritical
 (struct sieve_instance *svinst, struct sieve_error_handler *ehandler,
 	const char *location, const char *user_prefix, const char *fmt, va_list args)
 {
-	char str[256];
-	struct tm *tm;
-
 	if ( location == NULL || *location == '\0' ) {
 		sieve_direct_verror
 			(svinst, svinst->system_ehandler, 0, NULL, fmt, args);
@@ -457,19 +476,7 @@ void sieve_vcritical
 			(svinst, svinst->system_ehandler, 0, location, fmt, args);
 	}
 
-	if ( ehandler == NULL || ehandler == svinst->system_ehandler ) return;
-
-	tm = localtime(&ioloop_time);
-
-	if ( user_prefix == NULL || *user_prefix == '\0' ) {
-		sieve_direct_error(svinst, ehandler, 0, location, "%s",
-			( strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
-				str : CRITICAL_MSG ));
-	} else {
-		sieve_direct_error(svinst, ehandler, 0, location, "%s: %s", user_prefix,
-			( strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
-				str : CRITICAL_MSG ));
-	}
+	sieve_internal_error(ehandler, location, user_prefix);
 }
 
 void sieve_error
@@ -534,6 +541,32 @@ void sieve_critical
 	va_end(args);
 }
 
+void sieve_internal_error
+(struct sieve_error_handler *ehandler, const char *location,
+	const char *user_prefix)
+{
+	char str[256];
+	struct tm *tm;
+
+	if ( ehandler == NULL ||
+		ehandler == ehandler->svinst->system_ehandler )
+		return;
+
+	tm = localtime(&ioloop_time);
+
+	if ( user_prefix == NULL || *user_prefix == '\0' ) {
+		sieve_direct_error(ehandler->svinst, ehandler, 0,
+			location, "%s",
+			( strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
+				str : CRITICAL_MSG ));
+	} else {
+		sieve_direct_error(ehandler->svinst, ehandler, 0,
+			location, "%s: %s", user_prefix,
+			( strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
+				str : CRITICAL_MSG ));
+	}
+}
+
 /*
  * Error statistics
  */
@@ -999,14 +1032,22 @@ static void sieve_logfile_start(struct sieve_logfile_ehandler *ehandler)
 
 			/* Rotate logfile */
 			rotated = t_strconcat(ehandler->logfile, ".0", NULL);
-			if ( rename(ehandler->logfile, rotated) < 0 ) {
-				sieve_sys_error(svinst,
-					"failed to rotate logfile: rename(%s, %s) failed: %m",
-					ehandler->logfile, rotated);
+			if ( rename(ehandler->logfile, rotated) < 0 && errno != ENOENT ) {
+				if ( errno == EACCES ) {
+					sieve_sys_error(svinst,
+						"failed to rotate logfile: %s",
+						eacces_error_get_creating("rename",
+							t_strconcat(ehandler->logfile, ", ", rotated, NULL)));
+				} else {
+					sieve_sys_error(svinst,
+						"failed to rotate logfile: rename(%s, %s) failed: %m",
+						ehandler->logfile, rotated);
+				}
 			}
 
 			/* Open clean logfile (overwrites existing if rename() failed earlier) */
-			fd = open(ehandler->logfile, O_CREAT | O_WRONLY | O_TRUNC, 0600);
+			fd = open(ehandler->logfile,
+				O_CREAT | O_APPEND | O_WRONLY | O_TRUNC, 0600);
 			if (fd == -1) {
 				if ( errno == EACCES ) {
 					sieve_sys_error(svinst,
@@ -1247,7 +1288,7 @@ struct sieve_varexpand_ehandler {
 	struct sieve_error_handler handler;
 
 	const char *format;
-	ARRAY_DEFINE(table, struct var_expand_table);
+	ARRAY(struct var_expand_table) table;
 };
 
 static const char *ATTR_FORMAT(3, 0) _expand_message
