@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -31,9 +31,9 @@ static struct sieve_extension *_sieve_extension_register
  */
 
 struct sieve_extension_registry {
-	ARRAY_DEFINE(extensions, struct sieve_extension *);
-	struct hash_table *extension_index;
-	struct hash_table *capabilities_index;
+	ARRAY(struct sieve_extension *) extensions;
+	HASH_TABLE(const char *, struct sieve_extension *) extension_index;
+	HASH_TABLE(const char *, struct sieve_capability_registration *) capabilities_index;
 
 	/* Core language 'extensions' */
 	const struct sieve_extension *comparator_extension;
@@ -41,7 +41,7 @@ struct sieve_extension_registry {
 	const struct sieve_extension *address_part_extension;
 
 	/* Preloaded extensions */
-	ARRAY_DEFINE(preloaded_extensions, const struct sieve_extension *);
+	ARRAY(const struct sieve_extension *) preloaded_extensions;
 };
 
 /*
@@ -58,35 +58,38 @@ extern const struct sieve_extension_def address_part_extension;
 
 /* FIXME: This is stupid. Define a comparator-* extension and be done with it */
 
-static const struct sieve_extension_def comparator_i_octet_extension = {
-	"comparator-i;octet",
-	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-	SIEVE_EXT_DEFINE_NO_OPERATIONS,
-	SIEVE_EXT_DEFINE_NO_OPERANDS
+const struct sieve_extension_def comparator_i_octet_extension = {
+	.name = "comparator-i;octet",
 };
 
-static const struct sieve_extension_def comparator_i_ascii_casemap_extension = {
-	"comparator-i;ascii-casemap",
-	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-	SIEVE_EXT_DEFINE_NO_OPERATIONS,
-	SIEVE_EXT_DEFINE_NO_OPERANDS
+const struct sieve_extension_def comparator_i_ascii_casemap_extension = {
+	.name = "comparator-i;ascii-casemap",
 };
 
 /*
- * Core extensions
+ * List of native extensions
  */
 
+/* Dummy extensions */
+
+extern const struct sieve_extension_def comparator_i_octet_extension;
+extern const struct sieve_extension_def comparator_i_ascii_casemap_extension;
+
+const struct sieve_extension_def *sieve_dummy_extensions[] = {
+	&comparator_i_octet_extension, &comparator_i_ascii_casemap_extension
+};
+
+const unsigned int sieve_dummy_extensions_count =
+	N_ELEMENTS(sieve_dummy_extensions);
+
+/* Core */
+
 extern const struct sieve_extension_def fileinto_extension;
 extern const struct sieve_extension_def reject_extension;
 extern const struct sieve_extension_def envelope_extension;
 extern const struct sieve_extension_def encoded_character_extension;
 
-/*
- * Native 'plugin' extensions
- */
-
 extern const struct sieve_extension_def vacation_extension;
-extern const struct sieve_extension_def vacation_seconds_extension;
 extern const struct sieve_extension_def subaddress_extension;
 extern const struct sieve_extension_def comparator_i_ascii_numeric_extension;
 extern const struct sieve_extension_def relational_extension;
@@ -100,29 +103,13 @@ extern const struct sieve_extension_def enotify_extension;
 extern const struct sieve_extension_def environment_extension;
 extern const struct sieve_extension_def mailbox_extension;
 extern const struct sieve_extension_def date_extension;
-extern const struct sieve_extension_def spamtest_extension;
-extern const struct sieve_extension_def spamtestplus_extension;
-extern const struct sieve_extension_def virustest_extension;
+extern const struct sieve_extension_def index_extension;
 extern const struct sieve_extension_def ihave_extension;
-extern const struct sieve_extension_def editheader_extension;
-
-/* vnd.dovecot. */
-extern const struct sieve_extension_def debug_extension;
 extern const struct sieve_extension_def duplicate_extension;
-
-/*
- * List of native extensions
- */
-
-const struct sieve_extension_def *sieve_dummy_extensions[] = {
-	/* Dummy extensions */
-	&comparator_i_octet_extension, &comparator_i_ascii_casemap_extension
-};
-
-const unsigned int sieve_dummy_extensions_count =
-	N_ELEMENTS(sieve_dummy_extensions);
-
-/* Core */
+extern const struct sieve_extension_def mime_extension;
+extern const struct sieve_extension_def foreverypart_extension;
+extern const struct sieve_extension_def mboxmetadata_extension;
+extern const struct sieve_extension_def servermetadata_extension;
 
 const struct sieve_extension_def *sieve_core_extensions[] = {
 	/* Core extensions */
@@ -135,7 +122,8 @@ const struct sieve_extension_def *sieve_core_extensions[] = {
 	&relational_extension, &regex_extension, &imap4flags_extension,
 	&copy_extension, &include_extension, &body_extension,
 	&variables_extension, &enotify_extension, &environment_extension,
-	&mailbox_extension, &date_extension, &ihave_extension,
+	&mailbox_extension, &date_extension, &index_extension, &ihave_extension,
+	&duplicate_extension, &mime_extension, &foreverypart_extension
 };
 
 const unsigned int sieve_core_extensions_count =
@@ -146,12 +134,22 @@ const unsigned int sieve_core_extensions_count =
  *   necessary to make these useful.
  */
 
+extern const struct sieve_extension_def vacation_seconds_extension;
+extern const struct sieve_extension_def spamtest_extension;
+extern const struct sieve_extension_def spamtestplus_extension;
+extern const struct sieve_extension_def virustest_extension;
+extern const struct sieve_extension_def editheader_extension;
+
+extern const struct sieve_extension_def vnd_debug_extension;
+extern const struct sieve_extension_def vnd_environment_extension;
+
 const struct sieve_extension_def *sieve_extra_extensions[] = {
 	&vacation_seconds_extension, &spamtest_extension, &spamtestplus_extension,
 	&virustest_extension, &editheader_extension,
+	&mboxmetadata_extension, &servermetadata_extension,
 
 	/* vnd.dovecot. */
-	&debug_extension, &duplicate_extension
+	&vnd_debug_extension, &vnd_environment_extension
 };
 
 const unsigned int sieve_extra_extensions_count =
@@ -163,10 +161,12 @@ const unsigned int sieve_extra_extensions_count =
 
 extern const struct sieve_extension_def imapflags_extension;
 extern const struct sieve_extension_def notify_extension;
+extern const struct sieve_extension_def vnd_duplicate_extension;
 
 const struct sieve_extension_def *sieve_deprecated_extensions[] = {
 	&imapflags_extension,
-	&notify_extension
+	&notify_extension,
+	&vnd_duplicate_extension
 };
 
 const unsigned int sieve_deprecated_extensions_count =
@@ -327,8 +327,8 @@ static void sieve_extension_registry_init(struct sieve_instance *svinst)
 	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
 
 	p_array_init(&ext_reg->extensions, svinst->pool, 50);
-	ext_reg->extension_index = hash_table_create
-		(default_pool, default_pool, 0, str_hash, (hash_cmp_callback_t *)strcmp);
+	hash_table_create
+		(&ext_reg->extension_index, default_pool, 0, str_hash, strcmp);
 }
 
 static void sieve_extension_registry_deinit(struct sieve_instance *svinst)
@@ -337,7 +337,7 @@ static void sieve_extension_registry_deinit(struct sieve_instance *svinst)
 	struct sieve_extension * const *exts;
     unsigned int i, ext_count;
 
-	if ( ext_reg->extension_index == NULL ) return;
+	if ( !hash_table_is_created(ext_reg->extension_index) ) return;
 
     exts = array_get_modifiable(&ext_reg->extensions, &ext_count);
 	for ( i = 0; i < ext_count; i++ ) {
@@ -363,35 +363,64 @@ bool sieve_extension_reload(const struct sieve_extension *ext)
 	return FALSE;
 }
 
+static struct sieve_extension *sieve_extension_lookup
+(struct sieve_instance *svinst, const char *name)
+{
+	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
+
+	return 	hash_table_lookup(ext_reg->extension_index, name);
+}
+
+static void sieve_extension_insert
+(struct sieve_instance *svinst, const char *name,
+	struct sieve_extension *ext)
+{
+	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
+
+	hash_table_insert(ext_reg->extension_index, name, ext);
+} 
+
+static struct sieve_extension *sieve_extension_alloc
+(struct sieve_instance *svinst,
+	const struct sieve_extension_def *extdef)
+{
+	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
+	struct sieve_extension *ext, **extr;
+	int ext_id;
+
+	ext_id = (int)array_count(&ext_reg->extensions);
+
+	/* Add extension to the registry */
+	extr = array_append_space(&ext_reg->extensions);
+	*extr = ext = p_new(svinst->pool, struct sieve_extension, 1);
+	ext->id = ext_id;
+	ext->def = extdef;
+	ext->svinst = svinst;
+	return ext;
+}
+
 static struct sieve_extension *_sieve_extension_register
 (struct sieve_instance *svinst, const struct sieve_extension_def *extdef,
 	bool load, bool required)
 {
-	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
-	struct sieve_extension *ext = (struct sieve_extension *)
-		hash_table_lookup(ext_reg->extension_index, extdef->name);
+	struct sieve_extension *ext;
+
+	ext = sieve_extension_lookup(svinst, extdef->name);
 
 	/* Register extension if it is not registered already */
 	if ( ext == NULL ) {
-		struct sieve_extension **extr;
+		ext = sieve_extension_alloc(svinst, extdef);
+		sieve_extension_insert(svinst, extdef->name, ext);
 
-		int ext_id = (int)array_count(&ext_reg->extensions);
+	} else if ( ext->overridden ) {
+		/* Create a dummy */
+		ext = sieve_extension_alloc(svinst, extdef);
 
-		/* Add extension to the registry */
-
-		extr = array_append_space(&ext_reg->extensions);
-		*extr = ext = p_new(svinst->pool, struct sieve_extension, 1);
-		ext->id = ext_id;
-		ext->def = extdef;
-		ext->svinst = svinst;
-
-		hash_table_insert
-			(ext_reg->extension_index, (void *) extdef->name, (void *) ext);
-
-	/* Re-register it if it were previously unregistered
-	 * (not going to happen)
-	 */
-	} else if ( ext->def == NULL ) {
+	} else {
+		/* Re-register it if it were previously unregistered
+		 * (not going to happen)
+		 */
+		i_assert( ext->def == NULL || ext->def == extdef );
 		ext->def = extdef;
 	}
 
@@ -429,11 +458,11 @@ void sieve_extension_unregister(const struct sieve_extension *ext)
 	if ( ext_id >= 0 && ext_id < (int) array_count(&ext_reg->extensions) ) {
 		mod_ext = array_idx(&ext_reg->extensions, ext_id);
 
-	sieve_extension_capabilities_unregister(*mod_ext);
-	_sieve_extension_unload(*mod_ext);
-	(*mod_ext)->loaded = FALSE;
-	(*mod_ext)->enabled = FALSE;
-	(*mod_ext)->def = NULL;
+		sieve_extension_capabilities_unregister(*mod_ext);
+		_sieve_extension_unload(*mod_ext);
+		(*mod_ext)->loaded = FALSE;
+		(*mod_ext)->enabled = FALSE;
+		(*mod_ext)->def = NULL;
 	}
 }
 
@@ -444,6 +473,25 @@ const struct sieve_extension *sieve_extension_require
 	return _sieve_extension_register(svinst, extdef, load, TRUE);
 }
 
+void sieve_extension_override
+(struct sieve_instance *svinst, const char *name,
+	const struct sieve_extension *ext)
+{
+	struct sieve_extension_registry *ext_reg = ext->svinst->ext_reg;
+	struct sieve_extension * const *mod_ext;
+	struct sieve_extension *old_ext;
+
+	i_assert( ext->id >= 0 &&
+		ext->id < (int) array_count(&ext_reg->extensions) );
+	mod_ext = array_idx(&ext_reg->extensions, ext->id);
+
+	old_ext = sieve_extension_lookup(svinst, name);
+	i_assert( old_ext == NULL || !old_ext->overridden );
+
+	sieve_extension_insert(svinst, name, *mod_ext);
+	(*mod_ext)->overridden = TRUE;
+}
+
 int sieve_extensions_get_count(struct sieve_instance *svinst)
 {
 	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
@@ -470,7 +518,6 @@ const struct sieve_extension *sieve_extension_get_by_id
 const struct sieve_extension *sieve_extension_get_by_name
 (struct sieve_instance *svinst, const char *name)
 {
-	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
 	const struct sieve_extension *ext;
 
 	if ( *name == '@' )
@@ -479,9 +526,7 @@ const struct sieve_extension *sieve_extension_get_by_name
 	if ( strlen(name) > 128 )
 		return NULL;
 
-	ext = (const struct sieve_extension *)
-		hash_table_lookup(ext_reg->extension_index, name);
-
+	ext = sieve_extension_lookup(svinst, name);
 	if ( ext == NULL || ext->def == NULL || (!ext->enabled && !ext->required))
 		return NULL;
 
@@ -566,8 +611,8 @@ void sieve_extensions_set_string
 (struct sieve_instance *svinst, const char *ext_string, bool global)
 {
 	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
-	ARRAY_DEFINE(enabled_extensions, const struct sieve_extension *);
-	ARRAY_DEFINE(disabled_extensions, const struct sieve_extension *);
+	ARRAY(const struct sieve_extension *) enabled_extensions;
+	ARRAY(const struct sieve_extension *) disabled_extensions;
 	const struct sieve_extension *const *ext_enabled;
 	const struct sieve_extension *const *ext_disabled;
 	struct sieve_extension **exts;
@@ -611,8 +656,7 @@ void sieve_extensions_set_string
 				if ( *name == '@' )
 					ext = NULL;
 				else
-					ext = (const struct sieve_extension *)
-						hash_table_lookup(ext_reg->extension_index, name);
+					ext = hash_table_lookup(ext_reg->extension_index, name);
 
 				if ( ext == NULL || ext->def == NULL ) {
 					sieve_sys_warning(svinst,
@@ -703,7 +747,7 @@ const struct sieve_extension *sieve_get_address_part_extension
 
 void sieve_enable_debug_extension(struct sieve_instance *svinst)
 {
-	(void) sieve_extension_register(svinst, &debug_extension, TRUE);
+	(void) sieve_extension_register(svinst, &vnd_debug_extension, TRUE);
 }
 
 /*
@@ -719,15 +763,15 @@ void sieve_capability_registry_init(struct sieve_instance *svinst)
 {
 	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
 
-	ext_reg->capabilities_index = hash_table_create
-		(default_pool, default_pool, 0, str_hash, (hash_cmp_callback_t *)strcmp);
+	hash_table_create
+		(&ext_reg->capabilities_index, default_pool, 0, str_hash, strcmp);
 }
 
 void sieve_capability_registry_deinit(struct sieve_instance *svinst)
 {
 	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
 
-	if ( ext_reg->capabilities_index == NULL ) return;
+	if ( !hash_table_is_created(ext_reg->capabilities_index) ) return;
 
 	hash_table_destroy(&svinst->ext_reg->capabilities_index);
 }
@@ -744,8 +788,7 @@ void sieve_extension_capabilities_register
 	reg->ext = ext;
 	reg->capabilities = cap;
 
-	hash_table_insert
-		(ext_reg->capabilities_index, (void *) cap->name, (void *) reg);
+	hash_table_insert(ext_reg->capabilities_index, cap->name, reg);
 }
 
 void sieve_extension_capabilities_unregister
@@ -753,15 +796,13 @@ void sieve_extension_capabilities_unregister
 {
 	struct sieve_extension_registry *ext_reg = ext->svinst->ext_reg;
 	struct hash_iterate_context *hictx;
-	void *key = NULL, *value = NULL;
+	const char *name;
+	struct sieve_capability_registration *reg;
 
 	hictx = hash_table_iterate_init(ext_reg->capabilities_index);
-	while ( hash_table_iterate(hictx, &key, &value) ) {
-		struct sieve_capability_registration *reg =
-			(struct sieve_capability_registration *) value;
-
+	while ( hash_table_iterate(hictx, ext_reg->capabilities_index, &name, &reg) ) {
 		if ( reg->ext == ext )
-			hash_table_remove(ext_reg->capabilities_index, key);
+			hash_table_remove(ext_reg->capabilities_index, name);
 	}
 	hash_table_iterate_deinit(&hictx);
 }
@@ -771,8 +812,7 @@ const char *sieve_extension_capabilities_get_string
 {
 	struct sieve_extension_registry *ext_reg = svinst->ext_reg;
 	const struct sieve_capability_registration *cap_reg =
-		(const struct sieve_capability_registration *)
-			hash_table_lookup(ext_reg->capabilities_index, cap_name);
+		hash_table_lookup(ext_reg->capabilities_index, cap_name);
 	const struct sieve_extension_capabilities *cap;
 
 	if ( cap_reg == NULL || cap_reg->capabilities == NULL )
