@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2015 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -9,6 +9,7 @@
 #include "mail-storage.h"
 
 #include "sieve-common.h"
+#include "sieve-limits.h"
 #include "sieve-script.h"
 #include "sieve-error.h"
 #include "sieve-extensions.h"
@@ -37,15 +38,28 @@ struct sieve_interpreter_extension_reg {
 	void *context;
 };
 
+/*
+ * Code loop
+ */
+
+struct sieve_interpreter_loop {
+	unsigned int level;
+	sieve_size_t begin, end;
+	const struct sieve_extension_def *ext_def;
+	pool_t pool;
+	void *context;
+};
+
 /*
  * Interpreter
  */
 
 struct sieve_interpreter {
 	pool_t pool;
+	struct sieve_interpreter *parent;
 
 	/* Runtime data for extensions */
-	ARRAY_DEFINE(extensions, struct sieve_interpreter_extension_reg);
+	ARRAY(struct sieve_interpreter_extension_reg) extensions;
 
 	sieve_size_t reset_vector;
 
@@ -55,6 +69,11 @@ struct sieve_interpreter {
 	bool interrupted;         /* Interpreter interrupt requested */
 	bool test_result;         /* Result of previous test command */
 
+	/* Loop stack */
+	ARRAY(struct sieve_interpreter_loop) loop_stack;
+	sieve_size_t loop_limit;
+	unsigned int parent_loop_level;
+
 	/* Runtime environment */
 	struct sieve_runtime_env runenv;
 	struct sieve_runtime_trace trace;
@@ -68,10 +87,15 @@ struct sieve_interpreter {
 };
 
 static struct sieve_interpreter *_sieve_interpreter_create
-(struct sieve_binary *sbin, struct sieve_binary_block *sblock,
-	struct sieve_script *script, const struct sieve_message_data *msgdata,
-	const struct sieve_script_env *senv, struct sieve_error_handler *ehandler,
+(struct sieve_binary *sbin,
+	struct sieve_binary_block *sblock,
+	struct sieve_script *script,
+	struct sieve_interpreter *parent,
+	const struct sieve_message_data *msgdata,
+	const struct sieve_script_env *senv,
+	struct sieve_error_handler *ehandler,
 	enum sieve_runtime_flags flags)
+	ATTR_NULL(3, 4)
 {
 	unsigned int i, ext_count;
 	struct sieve_interpreter *interp;
@@ -84,6 +108,7 @@ static struct sieve_interpreter *_sieve_interpreter_create
 
 	pool = pool_alloconly_create("sieve_interpreter", 4096);
 	interp = p_new(pool, struct sieve_interpreter, 1);
+	interp->parent = parent;
 	interp->pool = pool;
 
 	interp->runenv.ehandler = ehandler;
@@ -126,6 +151,12 @@ static struct sieve_interpreter *_sieve_interpreter_create
 
 	p_array_init(&interp->extensions, pool, sieve_extensions_get_count(svinst));
 
+	interp->parent_loop_level = 0;
+	if ( parent != NULL && array_is_created(&parent->loop_stack) ) {
+		interp->parent_loop_level = parent->parent_loop_level +
+			array_count(&parent->loop_stack);
+	}
+
 	/* Pre-load core language features implemented as 'extensions' */
 	ext_preloaded = sieve_extensions_get_preloaded(svinst, &ext_count);
 	for ( i = 0; i < ext_count; i++ ) {
@@ -194,8 +225,11 @@ static struct sieve_interpreter *_sieve_interpreter_create
 }
 
 struct sieve_interpreter *sieve_interpreter_create
-(struct sieve_binary *sbin, const struct sieve_message_data *msgdata,
-	const struct sieve_script_env *senv, struct sieve_error_handler *ehandler,
+(struct sieve_binary *sbin,
+	struct sieve_interpreter *parent,
+	const struct sieve_message_data *msgdata,
+	const struct sieve_script_env *senv,
+	struct sieve_error_handler *ehandler,
 	enum sieve_runtime_flags flags)
 {
 	struct sieve_binary_block *sblock;
@@ -204,42 +238,56 @@ struct sieve_interpreter *sieve_interpreter_create
 		== NULL )
 		return NULL;
 
- 	return _sieve_interpreter_create
-		(sbin, sblock, NULL, msgdata, senv, ehandler, flags);
+ 	return _sieve_interpreter_create(sbin, sblock, NULL,
+		parent, msgdata, senv, ehandler, flags);
 }
 
 struct sieve_interpreter *sieve_interpreter_create_for_block
-(struct sieve_binary_block *sblock, struct sieve_script *script,
-	const struct sieve_message_data *msgdata, const struct sieve_script_env *senv,
-	struct sieve_error_handler *ehandler, enum sieve_runtime_flags flags)
+(struct sieve_binary_block *sblock,
+	struct sieve_script *script,
+	struct sieve_interpreter *parent,
+	const struct sieve_message_data *msgdata,
+	const struct sieve_script_env *senv,
+	struct sieve_error_handler *ehandler,
+	enum sieve_runtime_flags flags)
 {
 	if ( sblock == NULL ) return NULL;
 
  	return _sieve_interpreter_create
-		(sieve_binary_block_get_binary(sblock), sblock, script, msgdata, senv,
-			ehandler, flags);
+		(sieve_binary_block_get_binary(sblock), sblock, script,
+			parent, msgdata, senv, ehandler, flags);
 }
 
-void sieve_interpreter_free(struct sieve_interpreter **interp)
+void sieve_interpreter_free(struct sieve_interpreter **_interp)
 {
+	struct sieve_interpreter *interp = *_interp;
+	struct sieve_runtime_env *renv = &interp->runenv;
 	const struct sieve_interpreter_extension_reg *eregs;
-	unsigned int ext_count, i;
+	struct sieve_interpreter_loop *loops;
+	unsigned int count, i;
 
-	sieve_runtime_trace_end(&((*interp)->runenv));
-
-	/* Signal registered extensions that the interpreter is being destroyed */
-	eregs = array_get(&(*interp)->extensions, &ext_count);
-	for ( i = 0; i < ext_count; i++ ) {
-		if ( eregs[i].intext != NULL && eregs[i].intext->free != NULL )
-			eregs[i].intext->free(eregs[i].ext, *interp, eregs[i].context);
+	if ( array_is_created(&interp->loop_stack) ) {
+		loops = array_get_modifiable(&interp->loop_stack, &count);
+		for ( i = 0; i < count; i++ )
+			pool_unref(&loops[i].pool);
 	}
 
-	sieve_binary_debug_reader_deinit(&(*interp)->dreader);
-	sieve_binary_unref(&(*interp)->runenv.sbin);
-	sieve_error_handler_unref(&(*interp)->runenv.ehandler);
+	interp->trace.indent = 0;
+	sieve_runtime_trace_end(renv);
 
-	pool_unref(&((*interp)->pool));
-	*interp = NULL;
+	/* Signal registered extensions that the interpreter is being destroyed */
+	eregs = array_get(&interp->extensions, &count);
+	for ( i = 0; i < count; i++ ) {
+		if ( eregs[i].intext != NULL && eregs[i].intext->free != NULL )
+			eregs[i].intext->free(eregs[i].ext, interp, eregs[i].context);
+	}
+
+	sieve_binary_debug_reader_deinit(&interp->dreader);
+	sieve_binary_unref(&renv->sbin);
+	sieve_error_handler_unref(&renv->ehandler);
+
+	pool_unref(&interp->pool);
+	*_interp = NULL;
 }
 
 /*
@@ -251,6 +299,12 @@ pool_t sieve_interpreter_pool(struct sieve_interpreter *interp)
 	return interp->pool;
 }
 
+struct sieve_interpreter *
+sieve_interpreter_get_parent(struct sieve_interpreter *interp)
+{
+	return interp->parent;
+}
+
 struct sieve_script *sieve_interpreter_script
 (struct sieve_interpreter *interp)
 {
@@ -349,6 +403,24 @@ void sieve_runtime_critical
 	va_end(args);
 }
 
+int sieve_runtime_mail_error
+(const struct sieve_runtime_env *renv, struct mail *mail,
+	const char *fmt, ...)
+{
+	const char *error_msg, *user_prefix;
+	va_list args;
+
+	error_msg = mailbox_get_last_error(mail->box, NULL);
+
+	va_start(args, fmt);
+	user_prefix = t_strdup_vprintf(fmt, args);
+	sieve_runtime_critical(renv, NULL, user_prefix,
+		"%s: %s", user_prefix, error_msg);
+	va_end(args);
+
+	return 	SIEVE_EXEC_TEMP_FAILURE;
+}
+
 /*
  * Source location
  */
@@ -434,6 +506,251 @@ void *sieve_interpreter_extension_get_context
 	return reg->context;
 }
 
+/*
+ * Loop handling
+ */
+
+int sieve_interpreter_loop_start
+(struct sieve_interpreter *interp, sieve_size_t loop_end,
+	const struct sieve_extension_def *ext_def,
+	struct sieve_interpreter_loop **loop_r)
+{
+	const struct sieve_runtime_env *renv = &interp->runenv;
+	struct sieve_interpreter_loop *loop;
+
+	i_assert( loop_end > interp->runenv.pc );
+
+	/* Check supplied end offset */
+	if ( loop_end > sieve_binary_block_get_size(renv->sblock) ) {
+		sieve_runtime_trace_error(renv,
+			"loop end offset out of range");
+		return SIEVE_EXEC_BIN_CORRUPT;
+	}
+
+	/* Trace */
+	if ( sieve_runtime_trace_active(renv, SIEVE_TRLVL_COMMANDS) ) {
+		unsigned int line =
+			sieve_runtime_get_source_location(renv, loop_end);
+
+		if ( sieve_runtime_trace_hasflag(renv, SIEVE_TRFLG_ADDRESSES) ) {
+			sieve_runtime_trace(renv, 0, "loop ends at line %d [%08llx]",
+				line, (long long unsigned int) loop_end);
+		} else {
+			sieve_runtime_trace(renv, 0, "loop ends at line %d", line);
+		}
+	}
+
+	/* Check loop nesting limit */
+	if ( !array_is_created(&interp->loop_stack) )
+		p_array_init(&interp->loop_stack, interp->pool, 8);
+	if ( (interp->parent_loop_level + array_count(&interp->loop_stack))
+		>= SIEVE_MAX_LOOP_DEPTH ) {
+		/* Should normally be caught at compile time */
+		sieve_runtime_error(renv, NULL,
+			"new program loop exceeds "
+			"the nesting limit (<= %u levels)",
+			SIEVE_MAX_LOOP_DEPTH);
+		return SIEVE_EXEC_FAILURE;
+	}
+
+	/* Create new loop */
+	loop = array_append_space(&interp->loop_stack);
+	loop->level = array_count(&interp->loop_stack)-1;
+	loop->ext_def = ext_def;
+	loop->begin = interp->runenv.pc;
+	loop->end = loop_end;
+	loop->pool =  pool_alloconly_create("sieve_interpreter", 128);
+
+	/* Set new loop limit */
+	interp->loop_limit = loop_end;
+	
+	*loop_r = loop;
+	return SIEVE_EXEC_OK;
+}
+
+struct sieve_interpreter_loop *sieve_interpreter_loop_get
+(struct sieve_interpreter *interp, sieve_size_t loop_end,
+	const struct sieve_extension_def *ext_def)
+{
+	struct sieve_interpreter_loop *loops;
+	unsigned int count, i;
+
+	if ( !array_is_created(&interp->loop_stack) )
+		return NULL;
+
+	loops = array_get_modifiable(&interp->loop_stack, &count);
+	for ( i = count; i > 0; i-- ) {
+		/* We're really making sure our loop matches */
+		if ( loops[i-1].end == loop_end &&
+			loops[i-1].ext_def == ext_def )
+			return &loops[i-1];
+	}
+	return NULL;
+}
+
+int sieve_interpreter_loop_next(struct sieve_interpreter *interp,
+	struct sieve_interpreter_loop *loop,
+	sieve_size_t loop_begin)
+{
+	const struct sieve_runtime_env *renv = &interp->runenv;
+	struct sieve_interpreter_loop *loops;
+	unsigned int count;
+
+	/* Trace */
+	if ( sieve_runtime_trace_active(renv, SIEVE_TRLVL_COMMANDS) ) {
+		unsigned int line =
+			sieve_runtime_get_source_location(renv, loop_begin);
+
+		if ( sieve_runtime_trace_hasflag(renv, SIEVE_TRFLG_ADDRESSES) ) {
+			sieve_runtime_trace(renv, 0, "looping back to line %d [%08llx]",
+				line, (long long unsigned int) loop_begin);
+		} else {
+			sieve_runtime_trace(renv, 0, "looping back to line %d", line);
+		}
+	}
+
+	/* Check the code for corruption */
+	if ( loop->begin != loop_begin ) {
+		sieve_runtime_trace_error(renv,
+			"loop begin offset invalid");
+		return SIEVE_EXEC_BIN_CORRUPT;
+	}
+
+	/* Check invariants */
+	i_assert( array_is_created(&interp->loop_stack) );
+	loops = array_get_modifiable(&interp->loop_stack, &count);
+	i_assert( &loops[count-1] == loop );
+
+	/* Return to beginning */
+	interp->runenv.pc = loop_begin;
+	return SIEVE_EXEC_OK;
+}
+
+int sieve_interpreter_loop_break(struct sieve_interpreter *interp,
+	struct sieve_interpreter_loop *loop)
+{
+	const struct sieve_runtime_env *renv = &interp->runenv;
+	struct sieve_interpreter_loop *loops;
+	sieve_size_t loop_end = loop->end;
+	unsigned int count, i;
+
+	/* Find the loop */
+	i_assert( array_is_created(&interp->loop_stack) );
+	loops = array_get_modifiable(&interp->loop_stack, &count);
+	i_assert( count > 0 );
+
+	i = count;
+	do {
+		pool_unref(&loops[i-1].pool);
+		i--;
+	} while ( i > 0 && &loops[i] != loop );
+	i_assert( &loops[i] == loop );
+
+	/* Set new loop limit */
+	if ( i > 0 )
+		interp->loop_limit = loops[i].end;
+	else
+		interp->loop_limit = 0;
+
+	/* Delete it and all deeper loops */
+	array_delete(&interp->loop_stack, i, count - i);
+
+	/* Trace */
+	if ( sieve_runtime_trace_active(renv, SIEVE_TRLVL_COMMANDS) ) {
+		unsigned int jmp_line =
+			sieve_runtime_get_source_location(renv, loop_end);
+
+		if ( sieve_runtime_trace_hasflag(renv, SIEVE_TRFLG_ADDRESSES) ) {
+			sieve_runtime_trace(renv, 0, "exiting loops at line %d [%08llx]",
+				jmp_line, (long long unsigned int) loop_end);
+		} else {
+			sieve_runtime_trace(renv, 0, "exiting loops at line %d", jmp_line);
+		}
+	}
+
+	/* Exit loop */
+	interp->runenv.pc = loop->end;
+	return SIEVE_EXEC_OK;
+}
+
+static int
+sieve_interpreter_loop_break_out(struct sieve_interpreter *interp,
+	sieve_size_t target)
+{
+	struct sieve_interpreter_loop *loops;
+	unsigned int count, i;
+
+	if ( !array_is_created(&interp->loop_stack) )
+		return SIEVE_EXEC_OK;
+
+	loops = array_get_modifiable(&interp->loop_stack, &count);
+	for ( i = count; i > 0; i-- ) {
+		/* We're really making sure our loop matches */
+		if ( loops[i-1].end > target )
+			break;
+	}
+	if ( i == count )
+		return SIEVE_EXEC_OK;
+
+	return sieve_interpreter_loop_break(interp, &loops[i]);
+}
+
+struct sieve_interpreter_loop *sieve_interpreter_loop_get_local
+(struct sieve_interpreter *interp,
+	struct sieve_interpreter_loop *loop,
+	const struct sieve_extension_def *ext_def)
+{
+	struct sieve_interpreter_loop *loops;
+	unsigned int count, i;
+
+	if ( !array_is_created(&interp->loop_stack) )
+		return NULL;
+
+	loops = array_get_modifiable(&interp->loop_stack, &count);
+	i_assert(loop == NULL || loop->level < count);
+
+	for ( i = (loop == NULL ? count : loop->level); i > 0; i-- ) {
+		if ( ext_def == NULL || loops[i-1].ext_def == ext_def )
+			return &loops[i-1];
+	}
+	return NULL;
+}
+
+struct sieve_interpreter_loop *sieve_interpreter_loop_get_global
+(struct sieve_interpreter *interp,
+	struct sieve_interpreter_loop *loop,
+	const struct sieve_extension_def *ext_def)
+{
+	struct sieve_interpreter_loop *result;
+
+	while (interp != NULL) {
+		result = sieve_interpreter_loop_get_local
+			(interp, loop, ext_def);
+		if (result != NULL)
+			return result;
+		interp = interp->parent;
+	}
+	return NULL;
+}
+
+pool_t sieve_interpreter_loop_get_pool
+(struct sieve_interpreter_loop *loop)
+{
+	return loop->pool;
+}
+
+void *sieve_interpreter_loop_get_context
+(struct sieve_interpreter_loop *loop)
+{
+	return loop->context;
+}
+
+void sieve_interpreter_loop_set_context
+(struct sieve_interpreter_loop *loop, void *context)
+{
+	loop->context = context;
+}
+
 /*
  * Program flow
  */
@@ -457,12 +774,14 @@ sieve_size_t sieve_interpreter_program_counter(struct sieve_interpreter *interp)
 }
 
 int sieve_interpreter_program_jump
-(struct sieve_interpreter *interp, bool jump)
+(struct sieve_interpreter *interp, bool jump, bool break_loops)
 {
 	const struct sieve_runtime_env *renv = &interp->runenv;
 	sieve_size_t *address = &(interp->runenv.pc);
-	sieve_size_t jmp_start = *address;
+	sieve_size_t jmp_start = *address, jmp_target;
+	sieve_size_t loop_limit = ( break_loops ? 0 : interp->loop_limit );
 	sieve_offset_t jmp_offset;
+	int ret;
 
 	if ( !sieve_binary_read_offset(renv->sblock, address, &jmp_offset) )
 	{
@@ -470,25 +789,28 @@ int sieve_interpreter_program_jump
 		return SIEVE_EXEC_BIN_CORRUPT;
 	}
 
-	if ( jmp_start + jmp_offset <= sieve_binary_block_get_size(renv->sblock) &&
+	jmp_target = jmp_start + jmp_offset;
+	if ( jmp_target <= sieve_binary_block_get_size(renv->sblock) &&
+		(loop_limit == 0 || jmp_target < loop_limit) &&
 		jmp_start + jmp_offset > 0 )
 	{
 		if ( jump ) {
-			sieve_size_t jmp_addr = jmp_start + jmp_offset;
-
 			if ( sieve_runtime_trace_active(renv, SIEVE_TRLVL_COMMANDS) ) {
 				unsigned int jmp_line =
-					sieve_runtime_get_source_location(renv, jmp_addr);
+					sieve_runtime_get_source_location(renv, jmp_target);
 
 				if ( sieve_runtime_trace_hasflag(renv, SIEVE_TRFLG_ADDRESSES) ) {
 					sieve_runtime_trace(renv, 0, "jumping to line %d [%08llx]",
-						jmp_line, (long long unsigned int) jmp_addr);
+						jmp_line, (long long unsigned int) jmp_target);
 				} else {
 					sieve_runtime_trace(renv, 0, "jumping to line %d", jmp_line);
 				}
 			}
 
-			*address = jmp_addr;
+			if ( break_loops && 
+				(ret=sieve_interpreter_loop_break_out(interp, jmp_target)) <= 0 )
+				return ret;
+			*address = jmp_target;
 		} else {
 			sieve_runtime_trace(renv, 0, "not jumping");
 		}
@@ -496,7 +818,13 @@ int sieve_interpreter_program_jump
 		return SIEVE_EXEC_OK;
 	}
 
-	sieve_runtime_trace_error(renv, "jump offset out of range");
+	if ( interp->loop_limit != 0 ) {
+		sieve_runtime_trace_error(renv,
+			"jump offset crosses loop boundary");
+	} else {
+		sieve_runtime_trace_error(renv,
+			"jump offset out of range");
+	}
 	return SIEVE_EXEC_BIN_CORRUPT;
 }
 
@@ -558,24 +886,31 @@ static int sieve_interpreter_operation_execute
 int sieve_interpreter_continue
 (struct sieve_interpreter *interp, bool *interrupted)
 {
+	const struct sieve_runtime_env *renv = &interp->runenv;
 	sieve_size_t *address = &(interp->runenv.pc);
 	int ret = SIEVE_EXEC_OK;
 
-	sieve_result_ref(interp->runenv.result);
+	sieve_result_ref(renv->result);
 	interp->interrupted = FALSE;
 
 	if ( interrupted != NULL )
 		*interrupted = FALSE;
 
 	while ( ret == SIEVE_EXEC_OK && !interp->interrupted &&
-		*address < sieve_binary_block_get_size(interp->runenv.sblock) ) {
+		*address < sieve_binary_block_get_size(renv->sblock) ) {
+		if ( interp->loop_limit != 0 && *address > interp->loop_limit ) {
+			sieve_runtime_trace_error(renv,
+				"program crossed loop boundary");
+			ret = SIEVE_EXEC_BIN_CORRUPT;
+			break;
+		}
 
 		ret = sieve_interpreter_operation_execute(interp);
+	}
 
-		if ( ret != SIEVE_EXEC_OK ) {
-			sieve_runtime_trace(&interp->runenv, SIEVE_TRLVL_NONE,
-				"[[EXECUTION ABORTED]]");
-		}
+	if ( ret != SIEVE_EXEC_OK ) {
+		sieve_runtime_trace(&interp->runenv, SIEVE_TRLVL_NONE,
+			"[[EXECUTION ABORTED]]");
 	}
 
 	if ( interrupted != NULL )
