@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2015 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -18,7 +18,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <stdlib.h>
 #include <unistd.h>
 #include <ctype.h>
 
@@ -64,20 +63,18 @@ const struct sieve_lexer *sieve_lexer_create
 (struct sieve_script *script, struct sieve_error_handler *ehandler,
 	enum sieve_error *error_r)
 {
-	pool_t pool;
 	struct sieve_lexical_scanner *scanner;
 	struct sieve_instance *svinst = sieve_script_svinst(script);
 	struct istream *stream;
 	const struct stat *st;
 
 	/* Open script as stream */
-	stream = sieve_script_open(script, error_r);
-	if ( stream == NULL )
+	if ( sieve_script_get_stream(script, &stream, error_r) < 0 )
 		return NULL;
 
 	/* Check script size */
-	st = i_stream_stat(stream, TRUE);
-	if ( st != NULL && st->st_size > 0 && svinst->max_script_size > 0 &&
+	if ( i_stream_stat(stream, TRUE, &st) >= 0 && st->st_size > 0 &&
+		svinst->max_script_size > 0 &&
 		(uoff_t)st->st_size > svinst->max_script_size ) {
 		sieve_error(ehandler, sieve_script_name(script),
 			"sieve script is too large (max %"PRIuSIZE_T" bytes)",
@@ -87,9 +84,7 @@ const struct sieve_lexer *sieve_lexer_create
 		return NULL;
 	}
 
-	pool = pool_alloconly_create("sieve_lexer_scanner", 1024);
-	scanner = p_new(pool, struct sieve_lexical_scanner, 1);
-	scanner->pool = pool;
+	scanner = i_new(struct sieve_lexical_scanner, 1);
 	scanner->lexer.scanner = scanner;
 
 	scanner->ehandler = ehandler;
@@ -106,7 +101,7 @@ const struct sieve_lexer *sieve_lexer_create
 	scanner->buffer_pos = 0;
 
 	scanner->lexer.token_type = STT_NONE;
-	scanner->lexer.token_str_value = str_new(pool, 256);
+	scanner->lexer.token_str_value = str_new(default_pool, 256);
 	scanner->lexer.token_int_value = 0;
 	scanner->lexer.token_line = 1;
 
@@ -115,20 +110,18 @@ const struct sieve_lexer *sieve_lexer_create
 	return &scanner->lexer;
 }
 
-void sieve_lexer_free(const struct sieve_lexer **lexer)
+void sieve_lexer_free(const struct sieve_lexer **_lexer)
 {
-	struct sieve_lexical_scanner *scanner = (*lexer)->scanner;
+	const struct sieve_lexer *lexer = *_lexer;
+	struct sieve_lexical_scanner *scanner = lexer->scanner;
 
 	i_stream_unref(&scanner->input);
-
-	sieve_script_close(scanner->script);
 	sieve_script_unref(&scanner->script);
-
 	sieve_error_handler_unref(&scanner->ehandler);
+	str_free(&scanner->lexer.token_str_value);
 
-	pool_unref(&scanner->pool);
-
-	*lexer = NULL;
+	i_free(scanner);
+	*_lexer = NULL;
 }
 
 /*
@@ -279,17 +272,61 @@ static inline const char *_char_sanitize(int ch)
 	return t_strdup_printf("0x%02x", ch);
 }
 
+static bool
+sieve_lexer_scan_hash_comment(struct sieve_lexical_scanner *scanner)
+{
+	struct sieve_lexer *lexer = &scanner->lexer;
+
+	while ( sieve_lexer_curchar(scanner) != '\n' ) {
+		switch( sieve_lexer_curchar(scanner) ) {
+		case -1:
+			if ( !scanner->input->eof ) {
+				lexer->token_type = STT_ERROR;
+				return FALSE;
+			}
+			sieve_lexer_warning(lexer,
+				"no newline (CRLF) at end of hash comment at end of file");
+			lexer->token_type = STT_WHITESPACE;
+			return TRUE;
+		case '\0':
+			sieve_lexer_error
+				(lexer, "encountered NUL character in hash comment");
+			lexer->token_type = STT_ERROR;
+			return FALSE;
+		default:
+			break;
+		}
+
+		/* Stray CR is ignored */
+
+		sieve_lexer_shift(scanner);
+	}
+
+	sieve_lexer_shift(scanner);
+
+	lexer->token_type = STT_WHITESPACE;
+	return TRUE;
+}
+
 /* sieve_lexer_scan_raw_token:
  *   Scans valid tokens and whitespace
  */
-static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
+static bool
+sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 {
 	struct sieve_lexer *lexer = &scanner->lexer;
 	string_t *str;
+	int ret;
 
 	/* Read first character */
 	if ( lexer->token_type == STT_NONE ) {
-		i_stream_read(scanner->input);
+		if ( (ret=i_stream_read(scanner->input)) < 0 ) {
+			i_assert( ret != -2 );
+			if ( !scanner->input->eof ) {
+				lexer->token_type = STT_ERROR;
+				return FALSE;
+			}
+		}
 		sieve_lexer_shift(scanner);
 	}
 
@@ -302,32 +339,7 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 	// hash-comment = ( "#" *CHAR-NOT-CRLF CRLF )
 	case '#':
 		sieve_lexer_shift(scanner);
-
-		while ( sieve_lexer_curchar(scanner) != '\n' ) {
-			switch( sieve_lexer_curchar(scanner) ) {
-			case -1:
-				sieve_lexer_warning
-					(lexer, "no newline (CRLF) at end of hash comment at end of file");
-				lexer->token_type = STT_WHITESPACE;
-				return TRUE;
-			case '\0':
-				sieve_lexer_error
-					(lexer, "encountered NUL character in hash comment");
-				lexer->token_type = STT_ERROR;
-				return FALSE;
-			default:
-				break;
-			}
-
-			/* Stray CR is ignored */
-
-			sieve_lexer_shift(scanner);
-		}
-
-		sieve_lexer_shift(scanner);
-
-		lexer->token_type = STT_WHITESPACE;
-		return TRUE;
+		return sieve_lexer_scan_hash_comment(scanner);
 
 	// bracket-comment = "/*" *(CHAR-NOT-STAR / ("*" CHAR-NOT-SLASH)) "*/"
 	//        ;; No */ allowed inside a comment.
@@ -343,9 +355,11 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 			while ( TRUE ) {
 				switch ( sieve_lexer_curchar(scanner) ) {
 				case -1:
-					sieve_lexer_error(lexer,
-						"end of file before end of bracket comment ('/* ... */') "
-						"started at line %d", lexer->token_line);
+					if ( scanner->input->eof ) {
+						sieve_lexer_error(lexer,
+							"end of file before end of bracket comment ('/* ... */') "
+							"started at line %d", lexer->token_line);
+					}
 					lexer->token_type = STT_ERROR;
 					return FALSE;
 				case '*':
@@ -409,17 +423,18 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 		str = lexer->token_str_value;
 
 		while ( sieve_lexer_curchar(scanner) != '"' ) {
-			if ( sieve_lexer_curchar(scanner) == '\\' ) {
+			if ( sieve_lexer_curchar(scanner) == '\\' )
 				sieve_lexer_shift(scanner);
-			}
 
 			switch ( sieve_lexer_curchar(scanner) ) {
 
 			/* End of file */
 			case -1:
-				sieve_lexer_error(lexer,
-					"end of file before end of quoted string "
-					"started at line %d", lexer->token_line);
+				if ( scanner->input->eof ) {
+					sieve_lexer_error(lexer,
+						"end of file before end of quoted string "
+						"started at line %d", lexer->token_line);
+				}
 				lexer->token_type = STT_ERROR;
 				return FALSE;
 
@@ -512,7 +527,11 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 
 	/* EOF */
 	case -1:
-	  lexer->token_type = STT_EOF;
+		if ( !scanner->input->eof ) {
+			lexer->token_type = STT_ERROR;
+			return FALSE;
+		}
+		lexer->token_type = STT_EOF;
 		return TRUE;
 
 	default:
@@ -629,24 +648,30 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
  					sieve_lexer_shift(scanner);
 
 				/* Discard hash comment or handle single CRLF */
-				switch ( sieve_lexer_curchar(scanner) ) {
-				case '#':
-					while ( sieve_lexer_curchar(scanner) != '\n' )
-						sieve_lexer_shift(scanner);
-					break;
-				case '\r':
+				if ( sieve_lexer_curchar(scanner) == '\r' )
 					sieve_lexer_shift(scanner);
-					break;
-				}
-
-				/* Terminating LF required */
  				switch ( sieve_lexer_curchar(scanner) ) {
+				case '#':
+					if ( !sieve_lexer_scan_hash_comment(scanner) )
+						return FALSE;
+					if ( scanner->input->eof ) {
+						sieve_lexer_error(lexer,
+							"end of file before end of multi-line string");
+						lexer->token_type = STT_ERROR;
+						return FALSE;
+					} else if ( scanner->input->stream_errno != 0 ) {
+						lexer->token_type = STT_ERROR;
+						return FALSE;
+					}
+					break;
 				case '\n':
 					sieve_lexer_shift(scanner);
 					break;
 				case -1:
-					sieve_lexer_error(lexer,
-						"end of file before end of multi-line string");
+					if ( scanner->input->eof ) {
+						sieve_lexer_error(lexer,
+							"end of file before end of multi-line string");
+					}
 					lexer->token_type = STT_ERROR;
 					return FALSE;
 				default:
@@ -694,9 +719,12 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 							return TRUE;
 						} else if ( cr_shifted ) {
 							/* Seen CR, but no LF */
-							sieve_lexer_error(lexer,
-								"found stray carriage-return (CR) character "
-								"in multi-line string started at line %d", lexer->token_line);
+							if ( sieve_lexer_curchar(scanner) != -1 ||
+								!scanner->input->eof ) {
+								sieve_lexer_error(lexer,
+									"found stray carriage-return (CR) character "
+									"in multi-line string started at line %d", lexer->token_line);
+							}
 							lexer->token_type = STT_ERROR;
 							return FALSE;
 						}
@@ -714,8 +742,10 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 
 						switch ( sieve_lexer_curchar(scanner) ) {
 						case -1:
-							sieve_lexer_error(lexer,
-								"end of file before end of multi-line string");
+							if ( scanner->input->eof ) {
+								sieve_lexer_error(lexer,
+									"end of file before end of multi-line string");
+							}
  							lexer->token_type = STT_ERROR;
  							return FALSE;
 						case '\0':
@@ -733,15 +763,17 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 					}
 
 					/* If exited loop due to CR, skip it */
-					if ( sieve_lexer_curchar(scanner) == '\r' ) {
+					if ( sieve_lexer_curchar(scanner) == '\r' )
 						sieve_lexer_shift(scanner);
-					}
 
 					/* Now we must see an LF */
 					if ( sieve_lexer_curchar(scanner) != '\n' ) {
-						sieve_lexer_error(lexer,
-							"found stray carriage-return (CR) character "
-							"in multi-line string started at line %d", lexer->token_line);
+						if ( sieve_lexer_curchar(scanner) != -1 ||
+							!scanner->input->eof ) {
+							sieve_lexer_error(lexer,
+								"found stray carriage-return (CR) character "
+								"in multi-line string started at line %d", lexer->token_line);
+						}
  						lexer->token_type = STT_ERROR;
  						return FALSE;
 					}
@@ -780,14 +812,22 @@ static bool sieve_lexer_scan_raw_token(struct sieve_lexical_scanner *scanner)
 	}
 }
 
-bool sieve_lexer_skip_token(const struct sieve_lexer *lexer)
+void sieve_lexer_skip_token(const struct sieve_lexer *lexer)
 {
 	/* Scan token while skipping whitespace */
 	do {
-		if ( !sieve_lexer_scan_raw_token(lexer->scanner) )
-			return FALSE;
-	} while ( lexer->token_type == STT_WHITESPACE );
+		struct sieve_lexical_scanner *scanner = lexer->scanner;
 
-	return TRUE;
+		if ( !sieve_lexer_scan_raw_token(scanner) ) {
+			if ( !scanner->input->eof && scanner->input->stream_errno != 0 ) {
+				sieve_critical(scanner->svinst, scanner->ehandler,
+					sieve_error_script_location(scanner->script, scanner->current_line),
+					"error reading script",
+					"error reading script during lexical analysis: %s",
+					i_stream_get_error(scanner->input));
+			}
+			return;
+		}
+	} while ( lexer->token_type == STT_WHITESPACE );
 }
 
