@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -10,6 +10,7 @@
 #include "str-sanitize.h"
 #include "var-expand.h"
 #include "message-address.h"
+#include "mail-storage.h"
 #include "mail-deliver.h"
 
 #include "sieve-common.h"
@@ -24,12 +25,6 @@
 
 #include <stdio.h>
 
-/*
- * Defaults
- */
-
-#define DEFAULT_ACTION_LOG_FORMAT "msgid=%m: %$"
-
 /*
  * Types
  */
@@ -76,9 +71,7 @@ struct sieve_result {
 	struct sieve_instance *svinst;
 
 	/* Context data for extensions */
-	ARRAY_DEFINE(ext_contexts, void *);
-
-	struct sieve_error_handler *ehandler;
+	ARRAY(void *) ext_contexts;
 
 	struct sieve_action_exec_env action_env;
 
@@ -91,12 +84,17 @@ struct sieve_result {
 
 	struct sieve_result_action *last_attempted_action;
 
-	struct hash_table *action_contexts;
+	HASH_TABLE(const struct sieve_action_def *,
+			   struct sieve_result_action_context *) action_contexts;
+
+	unsigned int executed:1;
+	unsigned int executed_delivery:1;
 };
 
 struct sieve_result *sieve_result_create
-(struct sieve_instance *svinst, const struct sieve_message_data *msgdata,
-	const struct sieve_script_env *senv, struct sieve_error_handler *ehandler)
+(struct sieve_instance *svinst,
+	const struct sieve_message_data *msgdata,
+	const struct sieve_script_env *senv)
 {
 	pool_t pool;
 	struct sieve_result *result;
@@ -109,10 +107,6 @@ struct sieve_result *sieve_result_create
 
 	p_array_init(&result->ext_contexts, pool, 4);
 
-	if ( ehandler != NULL )
-		sieve_error_handler_ref(ehandler);
-	result->ehandler = ehandler;
-
 	result->action_env.svinst = svinst;
 	result->action_env.result = result;
 	result->action_env.scriptenv = senv;
@@ -129,7 +123,6 @@ struct sieve_result *sieve_result_create
 	result->first_action = NULL;
 	result->last_action = NULL;
 
-	result->action_contexts = NULL;
 	return result;
 }
 
@@ -147,12 +140,9 @@ void sieve_result_unref(struct sieve_result **result)
 
 	sieve_message_context_unref(&(*result)->action_env.msgctx);
 
-	if ( (*result)->action_contexts != NULL )
+	if ( hash_table_is_created((*result)->action_contexts) )
         hash_table_destroy(&(*result)->action_contexts);
 
-	if ( (*result)->ehandler != NULL )
-		sieve_error_handler_unref(&(*result)->ehandler);
-
 	if ( (*result)->action_env.ehandler != NULL )
 		sieve_error_handler_unref(&(*result)->action_env.ehandler);
 
@@ -170,11 +160,6 @@ pool_t sieve_result_pool(struct sieve_result *result)
  * Getters/Setters
  */
 
-struct sieve_error_handler *sieve_result_get_error_handler
-(struct sieve_result *result)
-{
-	return result->ehandler;
-}
 const struct sieve_script_env *sieve_result_get_script_env
 (struct sieve_result *result)
 {
@@ -193,16 +178,6 @@ struct sieve_message_context *sieve_result_get_message_context
 	return result->action_env.msgctx;
 }
 
-void sieve_result_set_error_handler
-(struct sieve_result *result, struct sieve_error_handler *ehandler)
-{
-	if ( result->ehandler != ehandler ) {
-		sieve_error_handler_ref(ehandler);
-		sieve_error_handler_unref(&result->ehandler);
-		result->ehandler = ehandler;
-	}
-}
-
 /*
  * Extension support
  */
@@ -272,7 +247,6 @@ void sieve_result_global_warning
 	va_end(args);
 }
 
-
 void sieve_result_log
 (const struct sieve_action_exec_env *aenv, const char *fmt, ...)
 {
@@ -293,6 +267,60 @@ void sieve_result_global_log
 	va_end(args);
 }
 
+void sieve_result_global_log_error
+(const struct sieve_action_exec_env *aenv, const char *fmt, ...)
+{
+	va_list args;
+
+	va_start(args, fmt);
+	sieve_global_info_verror(aenv->svinst, aenv->ehandler, NULL, fmt, args);
+	va_end(args);
+}
+
+void sieve_result_global_log_warning
+(const struct sieve_action_exec_env *aenv, const char *fmt, ...)
+{
+	va_list args;
+
+	va_start(args, fmt);
+	sieve_global_info_vwarning(aenv->svinst, aenv->ehandler, NULL, fmt, args);
+	va_end(args);
+}
+
+void sieve_result_critical
+(const struct sieve_action_exec_env *aenv,
+	const char *user_prefix, const char *fmt, ...)
+{
+	va_list args;
+
+	va_start(args, fmt);
+
+	T_BEGIN {
+		sieve_vcritical
+			(aenv->svinst, aenv->ehandler, NULL, user_prefix, fmt, args);
+	} T_END;
+
+	va_end(args);
+}
+
+int sieve_result_mail_error
+(const struct sieve_action_exec_env *aenv, struct mail *mail,
+	const char *fmt, ...)
+{
+	const char *error_msg, *user_prefix;
+	va_list args;
+
+	error_msg = mailbox_get_last_error(mail->box, NULL);
+
+	va_start(args, fmt);
+	user_prefix = t_strdup_vprintf(fmt, args);
+	sieve_result_critical(aenv, user_prefix,
+		"%s: %s", user_prefix, error_msg);
+	va_end(args);
+
+	return 	SIEVE_EXEC_TEMP_FAILURE;
+}
+
 /*
  * Result composition
  */
@@ -307,12 +335,10 @@ void sieve_result_add_implicit_side_effect
 
 	to_action = to_keep ? &act_store : to_action;
 
-	if ( result->action_contexts == NULL ) {
-		result->action_contexts = hash_table_create
-			(default_pool, result->pool, 0, NULL, NULL);
+	if ( !hash_table_is_created(result->action_contexts) ) {
+		hash_table_create_direct(&result->action_contexts, result->pool, 0);
 	} else {
-		actctx = (struct sieve_result_action_context *)
-			hash_table_lookup(result->action_contexts, to_action);
+		actctx = hash_table_lookup(result->action_contexts, to_action);
 	}
 
 	if ( actctx == NULL ) {
@@ -321,8 +347,7 @@ void sieve_result_add_implicit_side_effect
 		actctx->action = to_action;
 		actctx->seffects = sieve_side_effects_list_create(result);
 
-		hash_table_insert(result->action_contexts, (void *) to_action,
-			(void *) actctx);
+		hash_table_insert(result->action_contexts, to_action, actctx);
 	}
 
 	seffect.object.def = &seff_def->obj_def;
@@ -604,12 +629,11 @@ static int _sieve_result_add_action
 		result->action_count++;
 
 		/* Apply any implicit side effects */
-		if ( result->action_contexts != NULL ) {
+		if ( hash_table_is_created(result->action_contexts) ) {
 			struct sieve_result_action_context *actctx;
 
 			/* Check for implicit side effects to this particular action */
-			actctx = (struct sieve_result_action_context *)
-				hash_table_lookup(result->action_contexts,
+			actctx = hash_table_lookup(result->action_contexts,
 					( keep ? &act_store : act_def ));
 
 			if ( actctx != NULL ) {
@@ -774,12 +798,11 @@ static void sieve_result_print_implicit_side_effects
 	bool dummy = TRUE;
 
 	/* Print any implicit side effects if applicable */
-	if ( result->action_contexts != NULL ) {
+	if ( hash_table_is_created(result->action_contexts) ) {
 		struct sieve_result_action_context *actctx;
 
 		/* Check for implicit side effects to keep action */
-		actctx = (struct sieve_result_action_context *)
-			hash_table_lookup(rpenv->result->action_contexts, &act_store);
+		actctx = hash_table_lookup(rpenv->result->action_contexts, &act_store);
 
 		if ( actctx != NULL && actctx->seffects != NULL )
 			sieve_result_print_side_effects
@@ -888,141 +911,154 @@ bool sieve_result_print
  * Result execution
  */
 
-static void _sieve_result_prepare_execution(struct sieve_result *result)
+static void _sieve_result_prepare_execution(struct sieve_result *result,
+	struct sieve_error_handler *ehandler,
+	enum sieve_execute_flags flags)
 {
-	const struct sieve_message_data *msgdata = result->action_env.msgdata;
 	const struct sieve_script_env *senv = result->action_env.scriptenv;
-	const struct var_expand_table *tab;
 
-	tab = mail_deliver_get_log_var_expand_table(msgdata->mail, NULL);
+	result->action_env.flags = flags;
+	result->action_env.ehandler = ehandler;
 	result->action_env.exec_status =
 		( senv->exec_status == NULL ?
 			t_new(struct sieve_exec_status, 1) : senv->exec_status );
-
-	if ( result->action_env.ehandler != NULL )
-		sieve_error_handler_unref(&result->action_env.ehandler);
-
-	if ( senv->action_log_format != NULL ) {
-		result->action_env.ehandler = sieve_varexpand_ehandler_create
-			(result->ehandler, senv->action_log_format, tab);
-	} else {
-		result->action_env.ehandler = sieve_varexpand_ehandler_create
-            (result->ehandler, DEFAULT_ACTION_LOG_FORMAT, tab);
-	}
 }
 
-static bool _sieve_result_implicit_keep
+static int _sieve_result_implicit_keep
 (struct sieve_result *result, bool rollback)
 {
+	const struct sieve_action_exec_env *aenv = &result->action_env;
 	struct sieve_result_action *rac;
-	bool success = TRUE;
+	int status = SIEVE_EXEC_OK;
 	struct sieve_result_side_effect *rsef, *rsef_first = NULL;
 	void *tr_context = NULL;
 	struct sieve_action act_keep;
 
+	if ( (aenv->flags & SIEVE_EXECUTE_FLAG_DEFER_KEEP) != 0 )
+		return SIEVE_EXEC_OK;
+
 	if ( rollback ) {
 		act_keep = result->failure_action;
 		act_keep.mail = NULL;
 	} else {
 		act_keep = result->keep_action;
-		act_keep.mail = sieve_message_get_mail(result->action_env.msgctx);
+		act_keep.mail = sieve_message_get_mail(aenv->msgctx);
 	}
 
 	/* If keep is a non-action, return right away */
-	if ( act_keep.def == NULL ) return TRUE;
+	if ( act_keep.def == NULL )
+		return SIEVE_EXEC_OK;
 
 	/* Scan for execution of keep-equal actions */
 	rac = result->first_action;
 	while ( rac != NULL ) {
 		if ( rac->action.def == act_keep.def && act_keep.def->equals != NULL &&
-			act_keep.def->equals
-				(result->action_env.scriptenv, NULL, &rac->action) &&
-					rac->action.executed )
-			return TRUE;
+			act_keep.def->equals(aenv->scriptenv, NULL, &rac->action) &&
+			rac->action.executed )
+			return SIEVE_EXEC_OK;
 
 		rac = rac->next;
 	}
 
 	/* Apply any implicit side effects if applicable */
-	if ( !rollback && result->action_contexts != NULL ) {
+	if ( !rollback && hash_table_is_created(result->action_contexts) ) {
 		struct sieve_result_action_context *actctx;
 
 		/* Check for implicit side effects to keep action */
-		actctx = (struct sieve_result_action_context *)
-				hash_table_lookup(result->action_contexts, act_keep.def);
+		actctx = hash_table_lookup(result->action_contexts, act_keep.def);
 
 		if ( actctx != NULL && actctx->seffects != NULL )
 			rsef_first = actctx->seffects->first_effect;
 	}
 
 	/* Start keep action */
-	if ( act_keep.def->start != NULL )
-		success = act_keep.def->start
-			(&act_keep, &result->action_env,  &tr_context);
+	if ( act_keep.def->start != NULL ) {
+		status = act_keep.def->start
+			(&act_keep, aenv, &tr_context);
+	}
 
 	/* Execute keep action */
-	if ( success ) {
+	if ( status == SIEVE_EXEC_OK ) {
 		rsef = rsef_first;
-		while ( success && rsef != NULL ) {
+		while ( status == SIEVE_EXEC_OK && rsef != NULL ) {
 			struct sieve_side_effect *sef = &rsef->seffect;
 
-			if ( sef->def->pre_execute != NULL )
-				success = success && sef->def->pre_execute
-					(sef, &act_keep, &result->action_env, &sef->context, tr_context);
+			if ( sef->def->pre_execute != NULL ) {
+				status = sef->def->pre_execute
+					(sef, &act_keep, aenv, &sef->context, tr_context);
+			}
 			rsef = rsef->next;
 		}
 
-		if ( act_keep.def->execute != NULL )
-			success = success && act_keep.def->execute
-				(&act_keep, &result->action_env, tr_context);
+		if ( status == SIEVE_EXEC_OK && act_keep.def->execute != NULL ) {
+			status = act_keep.def->execute
+				(&act_keep, aenv, tr_context);
+		}
 
 		rsef = rsef_first;
-		while ( success && rsef != NULL ) {
+		while ( status == SIEVE_EXEC_OK && rsef != NULL ) {
 			struct sieve_side_effect *sef = &rsef->seffect;
 
-			if ( sef->def->post_execute != NULL )
-				success = success && sef->def->post_execute
-					(sef, &act_keep, &result->action_env, tr_context);
+			if ( sef->def->post_execute != NULL ) {
+				status = sef->def->post_execute
+					(sef, &act_keep, aenv, tr_context);
+			}
 			rsef = rsef->next;
 		}
 	}
 
-	/* Finish keep action */
-	if ( success ) {
+	/* Commit keep action */
+	if ( status == SIEVE_EXEC_OK ) {
 		bool dummy = TRUE;
 
-		if ( act_keep.def->commit != NULL )
-			success = act_keep.def->commit
-				(&act_keep, &result->action_env, tr_context, &dummy);
+		if ( act_keep.def->commit != NULL ) {
+			status = act_keep.def->commit
+				(&act_keep, aenv, tr_context, &dummy);
+		}
 
 		rsef = rsef_first;
 		while ( rsef != NULL ) {
 			struct sieve_side_effect *sef = &rsef->seffect;
 			bool keep = TRUE;
 
-			if ( sef->def->post_commit != NULL )
+			if ( sef->def->post_commit != NULL ) {
 				sef->def->post_commit
-					(sef, &act_keep, &result->action_env, tr_context, &keep);
+					(sef, &act_keep, aenv, tr_context, &keep);
+			}
 			rsef = rsef->next;
 		}
-
-		return success;
+	} else {
+		/* Failed, rollback */
+		if ( act_keep.def->rollback != NULL ) {
+			act_keep.def->rollback(&act_keep,
+				aenv, tr_context, ( status == SIEVE_EXEC_OK ));
+		}
 	}
 
-	/* Failed, rollback */
-	if ( act_keep.def->rollback != NULL )
-		act_keep.def->rollback
-			(&act_keep, &result->action_env, tr_context, success);
+	/* Finish keep action */
+	if ( act_keep.def->finish != NULL ) {
+		act_keep.def->finish
+			(&act_keep, aenv, &rac->tr_context, status);
+	}
 
-	return FALSE;
+	return status;
 }
 
-bool sieve_result_implicit_keep
-(struct sieve_result *result)
+int sieve_result_implicit_keep
+(struct sieve_result *result,
+	struct sieve_error_handler *ehandler,
+	enum sieve_execute_flags flags,
+	bool success)
 {
-	_sieve_result_prepare_execution(result);
+	int ret;
 
-	return _sieve_result_implicit_keep(result, TRUE);
+	_sieve_result_prepare_execution(result, ehandler, flags);
+
+	ret = _sieve_result_implicit_keep(result, !success);
+
+	result->action_env.ehandler = NULL;
+
+	return ret;
 }
 
 void sieve_result_mark_executed(struct sieve_result *result)
@@ -1042,32 +1078,27 @@ void sieve_result_mark_executed(struct sieve_result *result)
 	}
 }
 
-int sieve_result_execute
-(struct sieve_result *result, bool *keep)
+
+bool sieve_result_executed(struct sieve_result *result)
 {
-	bool implicit_keep = TRUE;
-	bool success = TRUE, commit_ok;
-	struct sieve_result_action *rac, *first_action;
-	struct sieve_result_action *last_attempted;
+	return result->executed;
+}
 
-	if ( keep != NULL ) *keep = FALSE;
+bool sieve_result_executed_delivery(struct sieve_result *result)
+{
+	return result->executed_delivery;
+}
 
-	/* Prepare environment */
+static int sieve_result_transaction_start
+(struct sieve_result *result, struct sieve_result_action *first,
+	struct sieve_result_action **last_r)
+{
+	const struct sieve_script_env *senv = result->action_env.scriptenv;
+	struct sieve_result_action *rac = first;
+	int status = SIEVE_EXEC_OK;
+	bool dup_flushed = FALSE;
 
-	_sieve_result_prepare_execution(result);
-
-	/* Make notice of this attempt */
-
-	first_action = ( result->last_attempted_action == NULL ?
-		result->first_action : result->last_attempted_action->next );
-	result->last_attempted_action = result->last_action;
-
-	/*
-	 * Transaction start
-	 */
-
-	rac = first_action;
-	while ( success && rac != NULL ) {
+	while ( status == SIEVE_EXEC_OK && rac != NULL ) {
 		struct sieve_action *act = &rac->action;
 
 		/* Skip non-actions (inactive keep) and executed ones */
@@ -1076,22 +1107,32 @@ int sieve_result_execute
 			continue;
 		}
 
+		if ((act->def->flags & SIEVE_ACTFLAG_MAIL_STORAGE) != 0 &&
+			!dup_flushed) {
+			sieve_action_duplicate_flush(senv);
+			dup_flushed = TRUE;
+		}
+
 		if ( act->def->start != NULL ) {
-			rac->success = act->def->start
+			status = act->def->start
 				(act, &result->action_env, &rac->tr_context);
-			success = success && rac->success;
+			rac->success = ( status == SIEVE_EXEC_OK );
 		}
 
 		rac = rac->next;
 	}
 
-	/*
-	 * Transaction execute
-	 */
+	*last_r = rac;
+	return status;
+}
 
-	last_attempted = rac;
-	rac = first_action;
-	while ( success && rac != NULL ) {
+static int sieve_result_transaction_execute
+(struct sieve_result *result, struct sieve_result_action *first)
+{
+	struct sieve_result_action *rac = first;
+	int status = SIEVE_EXEC_OK;
+
+	while ( status == SIEVE_EXEC_OK && rac != NULL ) {
 		struct sieve_action *act = &rac->action;
 		struct sieve_result_side_effect *rsef;
 		struct sieve_side_effect *sef;
@@ -1104,120 +1145,292 @@ int sieve_result_execute
 
 		/* Execute pre-execute event of side effects */
 		rsef = rac->seffects != NULL ? rac->seffects->first_effect : NULL;
-		while ( success && rsef != NULL ) {
+		while ( status == SIEVE_EXEC_OK && rsef != NULL ) {
 			sef = &rsef->seffect;
 			if ( sef->def != NULL && sef->def->pre_execute != NULL )
-				success = success & sef->def->pre_execute
+				status = sef->def->pre_execute
 					(sef, act, &result->action_env, &sef->context, rac->tr_context);
 			rsef = rsef->next;
 		}
 
 		/* Execute the action itself */
-		if ( success && act->def != NULL && act->def->execute != NULL ) {
-			rac->success = act->def->execute
+		if ( status == SIEVE_EXEC_OK && act->def != NULL &&
+			act->def->execute != NULL ) {
+			status = act->def->execute
 				(act, &result->action_env, rac->tr_context);
-			success = success && rac->success;
 		}
 
 		/* Execute post-execute event of side effects */
 		rsef = rac->seffects != NULL ? rac->seffects->first_effect : NULL;
-		while ( success && rsef != NULL ) {
+		while ( status == SIEVE_EXEC_OK && rsef != NULL ) {
 			sef = &rsef->seffect;
 			if ( sef->def != NULL && sef->def->post_execute != NULL )
-				success = success && sef->def->post_execute
+				status = sef->def->post_execute
 					(sef, act, &result->action_env, rac->tr_context);
 			rsef = rsef->next;
 		}
 
+		rac->success = ( status == SIEVE_EXEC_OK );
 		rac = rac->next;
 	}
 
-	/*
-	 * Transaction commit/rollback
-	 */
+	return status;
+}
 
-	commit_ok = success;
-	rac = first_action;
-	while ( rac != NULL && rac != last_attempted ) {
+static int sieve_result_action_commit
+(struct sieve_result *result, struct sieve_result_action *rac,
+	bool *impl_keep)
+{
+	struct sieve_action *act = &rac->action;
+	struct sieve_result_side_effect *rsef;
+	int cstatus = SIEVE_EXEC_OK;
+
+	if ( act->def->commit != NULL ) {
+		cstatus = act->def->commit
+			(act, &result->action_env, rac->tr_context, impl_keep);
+		if ( cstatus == SIEVE_EXEC_OK ) {
+			act->executed = TRUE;
+			result->executed = TRUE;
+		}
+	}
+
+	if ( cstatus == SIEVE_EXEC_OK ) {
+		/* Execute post_commit event of side effects */
+		rsef = rac->seffects != NULL ? rac->seffects->first_effect : NULL;
+		while ( rsef != NULL ) {
+			struct sieve_side_effect *sef = &rsef->seffect;
+			if ( sef->def->post_commit != NULL )
+				sef->def->post_commit
+					(sef, act, &result->action_env, rac->tr_context, impl_keep);
+			rsef = rsef->next;
+		}
+	}
+
+	return cstatus;
+}
+
+static void sieve_result_action_rollback
+(struct sieve_result *result, struct sieve_result_action *rac)
+{
+	struct sieve_action *act = &rac->action;
+	struct sieve_result_side_effect *rsef;
+
+	if ( act->def->rollback != NULL ) {
+		act->def->rollback
+			(act, &result->action_env, rac->tr_context, rac->success);
+	}
+
+	/* Rollback side effects */
+	rsef = rac->seffects != NULL ? rac->seffects->first_effect : NULL;
+	while ( rsef != NULL ) {
+		struct sieve_side_effect *sef = &rsef->seffect;
+		if ( sef->def != NULL && sef->def->rollback != NULL )
+			sef->def->rollback
+				(sef, act, &result->action_env, rac->tr_context, rac->success);
+		rsef = rsef->next;
+	}
+}
+
+static int sieve_result_action_commit_or_rollback
+(struct sieve_result *result, struct sieve_result_action *rac,
+	int status, bool *implicit_keep, bool *keep, int *commit_status)
+{
+	struct sieve_action *act = &rac->action;
+
+	if ( status == SIEVE_EXEC_OK ) {
+		bool impl_keep = TRUE;
+		int cstatus = SIEVE_EXEC_OK;
+
+		if ( rac->keep && keep != NULL ) *keep = TRUE;
+
+		/* Skip non-actions (inactive keep) and executed ones */
+		if ( act->def == NULL || act->executed )
+			return status;
+
+		cstatus = sieve_result_action_commit(result, rac, &impl_keep);
+		if (cstatus != SIEVE_EXEC_OK) {
+			/* This is bad; try to salvage as much as possible */
+			if (*commit_status == SIEVE_EXEC_OK) {
+				*commit_status = cstatus;
+				if (!result->executed) {
+					/* We haven't executed anything yet; continue as rollback */
+					status = cstatus;
+				}
+			}
+			impl_keep = TRUE;
+		}
+
+		*implicit_keep = *implicit_keep && impl_keep;
+	} else {
+		/* Skip non-actions (inactive keep) and executed ones */
+		if ( act->def == NULL || act->executed )
+			return status;
+
+		sieve_result_action_rollback(result, rac);
+	}
+
+	return status;
+}
+
+static int sieve_result_transaction_commit_or_rollback
+(struct sieve_result *result, int status,
+	struct sieve_result_action *first,
+	struct sieve_result_action *last,
+	bool *implicit_keep, bool *keep)
+{
+	struct sieve_result_action *rac;
+	int commit_status = status;
+	bool seen_delivery = FALSE;
+
+	/* First commit/rollback all storage actions */
+	rac = first;
+	while ( rac != NULL && rac != last ) {
 		struct sieve_action *act = &rac->action;
-		struct sieve_result_side_effect *rsef;
-		struct sieve_side_effect *sef;
 
-		if ( success ) {
-			bool impl_keep = TRUE;
+		if (act->def == NULL ||
+			(act->def->flags & SIEVE_ACTFLAG_MAIL_STORAGE) == 0) {
+			rac = rac->next;
+			continue;
+		}
 
-			if ( rac->keep && keep != NULL ) *keep = TRUE;
+		status = sieve_result_action_commit_or_rollback
+			(result, rac, status, implicit_keep, keep, &commit_status);
+		
+		if ( act->def != NULL &&
+			(act->def->flags & SIEVE_ACTFLAG_TRIES_DELIVER) != 0 )
+			seen_delivery = TRUE;
 
-			/* Skip non-actions (inactive keep) and executed ones */
-			if ( act->def == NULL || act->executed ) {
-				rac = rac->next;
-				continue;
-			}
+		rac = rac->next;
+	}
 
-			if ( act->def->commit != NULL ) {
-				act->executed = act->def->commit
-					(act, &result->action_env, rac->tr_context, &impl_keep);
-				commit_ok = act->executed && commit_ok;
-			}
+	/* Then commit/rollback all other actions */
+	rac = first;
+	while ( rac != NULL && rac != last ) {
+		struct sieve_action *act = &rac->action;
 
-			/* Execute post_commit event of side effects */
-			rsef = rac->seffects != NULL ? rac->seffects->first_effect : NULL;
-			while ( rsef != NULL ) {
-				sef = &rsef->seffect;
-				if ( sef->def->post_commit != NULL )
-					sef->def->post_commit
-						(sef, act, &result->action_env, rac->tr_context, &impl_keep);
-				rsef = rsef->next;
-			}
+		if (act->def != NULL &&
+			(act->def->flags & SIEVE_ACTFLAG_MAIL_STORAGE) != 0) {
+			rac = rac->next;
+			continue;
+		}
 
-			implicit_keep = implicit_keep && impl_keep;
-		} else {
-			/* Skip non-actions (inactive keep) and executed ones */
-			if ( act->def == NULL || act->executed ) {
-				rac = rac->next;
-				continue;
-			}
+		status = sieve_result_action_commit_or_rollback
+			(result, rac, status, implicit_keep, keep, &commit_status);
 
-			if ( act->def->rollback != NULL )
-				act->def->rollback
-					(act, &result->action_env, rac->tr_context, rac->success);
+		if ( act->def != NULL &&
+			(act->def->flags & SIEVE_ACTFLAG_TRIES_DELIVER) != 0 )
+			seen_delivery = TRUE;
 
-			/* Rollback side effects */
-			rsef = rac->seffects != NULL ? rac->seffects->first_effect : NULL;
-			while ( rsef != NULL ) {
-				sef = &rsef->seffect;
-				if ( sef->def && sef->def->rollback != NULL )
-					sef->def->rollback
-						(sef, act, &result->action_env, rac->tr_context, rac->success);
-				rsef = rsef->next;
-			}
+		rac = rac->next;
+	}
+
+	if ( *implicit_keep && keep != NULL ) *keep = TRUE;
+
+	if ( commit_status == SIEVE_EXEC_OK ) {
+		result->executed_delivery =
+			result->executed_delivery || seen_delivery;
+	}
+
+	return commit_status;
+}
+
+static void sieve_result_transaction_finish
+(struct sieve_result *result, struct sieve_result_action *first,
+	int status)
+{
+	struct sieve_result_action *rac = first;
+
+	while ( rac != NULL ) {
+		struct sieve_action *act = &rac->action;
+		/* Skip non-actions (inactive keep) and executed ones */
+		if ( act->def == NULL || act->executed ) {
+			rac = rac->next;
+			continue;
+		}
+
+		if ( act->def->finish != NULL ) {
+			act->def->finish
+				(act, &result->action_env, &rac->tr_context, status);
 		}
 
 		rac = rac->next;
 	}
+}
 
-	if ( implicit_keep && keep != NULL ) *keep = TRUE;
+int sieve_result_execute
+(struct sieve_result *result, bool *keep,
+	struct sieve_error_handler *ehandler,
+	enum sieve_execute_flags flags)
+{
+	int status = SIEVE_EXEC_OK, result_status;
+	struct sieve_result_action *first_action, *last_action;
+	bool implicit_keep = TRUE;
+	int ret;
 
-	/* Return value indicates whether the caller should attempt an implicit keep
-	 * of its own. So, if the above transaction fails, but the implicit keep below
-	 * succeeds, the return value is still true. An error is/should be logged
-	 * though.
-	 */
+	if ( keep != NULL ) *keep = FALSE;
 
-	/* Execute implicit keep if the transaction failed or when the implicit keep
-	 * was not canceled during transaction.
-	 */
-	if ( !commit_ok || implicit_keep ) {
-		if ( !_sieve_result_implicit_keep(result, !commit_ok) )
-			return SIEVE_EXEC_KEEP_FAILED;
+	/* Prepare environment */
 
-		return ( commit_ok ?
-			SIEVE_EXEC_OK            /* Success */ :
-			SIEVE_EXEC_FAILURE       /* Implicit keep executed */ );
+	_sieve_result_prepare_execution(result, ehandler, flags);
+
+	/* Make notice of this attempt */
+
+	first_action = ( result->last_attempted_action == NULL ?
+		result->first_action : result->last_attempted_action->next );
+	result->last_attempted_action = result->last_action;
+
+	/* Transaction start */
+
+	status = sieve_result_transaction_start
+		(result, first_action, &last_action);
+
+	/* Transaction execute */
+
+	if (status == SIEVE_EXEC_OK) {
+		status = sieve_result_transaction_execute
+			(result, first_action);
 	}
 
-	/* Unconditional success */
-	return SIEVE_EXEC_OK;
+	/* Transaction commit/rollback */
+
+	status = sieve_result_transaction_commit_or_rollback
+		(result, status, first_action, last_action,
+			&implicit_keep, keep);
+
+	/* Perform implicit keep if necessary */
+
+	result_status = status;
+	if ( result->executed || status != SIEVE_EXEC_TEMP_FAILURE ) {
+		/* Execute implicit keep if the transaction failed or when the implicit
+		 * keep was not canceled during transaction.
+		 */
+		if ( status != SIEVE_EXEC_OK || implicit_keep ) {
+			switch ((ret=_sieve_result_implicit_keep
+				(result, ( status != SIEVE_EXEC_OK ))) ) {
+			case SIEVE_EXEC_OK:
+				if ( result_status == SIEVE_EXEC_TEMP_FAILURE )
+					result_status = SIEVE_EXEC_FAILURE;
+				break;
+			case SIEVE_EXEC_TEMP_FAILURE:
+				if (!result->executed) {
+					result_status = ret;
+					break;
+				}
+			default:
+				result_status = SIEVE_EXEC_KEEP_FAILED;
+			}
+		}
+		if (status == SIEVE_EXEC_OK)
+			status = result_status;
+	}
+
+	/* Finish execution */
+
+	sieve_result_transaction_finish
+		(result, first_action, status);
+
+	result->action_env.ehandler = NULL;
+	return result_status;
 }
 
 /*
