@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2013 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2016 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -48,9 +48,9 @@ struct sieve_command_registration {
 	const struct sieve_command_def *cmd_def;
 	const struct sieve_extension *ext;
 
-	ARRAY_DEFINE(normal_tags, struct sieve_tag_registration *);
-	ARRAY_DEFINE(instanced_tags, struct sieve_tag_registration *);
-	ARRAY_DEFINE(persistent_tags, struct sieve_tag_registration *);
+	ARRAY(struct sieve_tag_registration *) normal_tags;
+	ARRAY(struct sieve_tag_registration *) instanced_tags;
+	ARRAY(struct sieve_tag_registration *) persistent_tags;
 };
 
 /* Default (literal) arguments */
@@ -93,9 +93,9 @@ struct sieve_validator {
 
 	/* Registries */
 
-	struct hash_table *commands;
+	HASH_TABLE(const char *, struct sieve_command_registration *) commands;
 
-	ARRAY_DEFINE(extensions, struct sieve_validator_extension_reg);
+	ARRAY(struct sieve_validator_extension_reg) extensions;
 
 	/* This is currently a wee bit ugly and needs more thought */
 	struct sieve_default_argument default_arguments[SAT_COUNT];
@@ -179,8 +179,8 @@ struct sieve_validator *sieve_validator_create
 		sieve_extensions_get_count(valdtr->svinst));
 
 	/* Setup command registry */
-	valdtr->commands = hash_table_create
-		(default_pool, pool, 0, strcase_hash, (hash_cmp_callback_t *)strcasecmp);
+	hash_table_create
+		(&valdtr->commands, pool, 0, strcase_hash, strcasecmp);
 	sieve_validator_register_core_commands(valdtr);
 	sieve_validator_register_core_tests(valdtr);
 
@@ -274,8 +274,13 @@ static bool _cmd_unknown_validate
 }
 
 static const struct sieve_command_def unknown_command = {
-	"", SCT_NONE, 0, 0, FALSE, FALSE ,
-	NULL, NULL, _cmd_unknown_validate, NULL, NULL, NULL
+	.identifier = "",
+	.type = SCT_NONE,
+	.positional_args = 0,
+	.subtests = 0,
+	.block_allowed = FALSE,
+	.block_required = FALSE,
+	.validate = _cmd_unknown_validate
 };
 
 /* Registration of the core commands of the language */
@@ -306,8 +311,7 @@ static struct sieve_command_registration *
 sieve_validator_find_command_registration
 (struct sieve_validator *valdtr, const char *command)
 {
-	return (struct sieve_command_registration *)
-		hash_table_lookup(valdtr->commands, command);
+	return hash_table_lookup(valdtr->commands, command);
 }
 
 static struct sieve_command_registration *_sieve_validator_register_command
@@ -320,7 +324,7 @@ static struct sieve_command_registration *_sieve_validator_register_command
 	cmd_reg->cmd_def = cmd_def;
 	cmd_reg->ext = ext;
 
-	hash_table_insert(valdtr->commands, (void *) identifier, (void *) cmd_reg);
+	hash_table_insert(valdtr->commands, identifier, cmd_reg);
 
 	return cmd_reg;
 }
@@ -376,10 +380,8 @@ static bool _unknown_tag_validate
 }
 
 static const struct sieve_argument_def _unknown_tag = {
-	"",
-	NULL,
-	_unknown_tag_validate,
-	NULL, NULL, NULL
+	.identifier = "",
+	.validate = _unknown_tag_validate,
 };
 
 static inline bool _tag_registration_is_unknown
@@ -427,14 +429,23 @@ void sieve_validator_register_persistent_tag
 
 		struct sieve_tag_registration *reg;
 
+		if ( !array_is_created(&cmd_reg->persistent_tags) )
+			p_array_init(&cmd_reg->persistent_tags, valdtr->pool, 4);
+		else {
+			struct sieve_tag_registration *const *reg_idx;
+
+			/* Avoid dupplicate registration */
+			array_foreach (&cmd_reg->persistent_tags, reg_idx) {
+				if ((*reg_idx)->tag_def == tag_def)
+					return;
+			}
+		}
+
 		reg = p_new(valdtr->pool, struct sieve_tag_registration, 1);
 		reg->ext = ext;
 		reg->tag_def = tag_def;
 		reg->id_code = -1;
 
-		if ( !array_is_created(&cmd_reg->persistent_tags) )
-			p_array_init(&cmd_reg->persistent_tags, valdtr->pool, 4);
-
 		array_append(&cmd_reg->persistent_tags, &reg, 1);
 	}
 }
@@ -848,7 +859,7 @@ bool sieve_validate_tag_parameter
 			t_strdup_printf(" %d (%s)", arg_pos, arg_name) );
 
 		sieve_argument_validate_error(valdtr, tag,
-			"the :%s tag for the %s %s requires %s as parameter %s, "
+			"the :%s tag for the %s %s requires %s as parameter%s, "
 			"but no parameters were found", sieve_ast_argument_tag(tag),
 			sieve_command_identifier(cmd), sieve_command_type_name(cmd),
 			sieve_ast_argument_type_name(req_type), position);
@@ -886,26 +897,29 @@ bool sieve_validate_tag_parameter
 static bool sieve_validate_command_arguments
 (struct sieve_validator *valdtr, struct sieve_command *cmd)
 {
-	int arg_count = cmd->def->positional_arguments;
+	int arg_count = cmd->def->positional_args;
 	int real_count = 0;
 	struct sieve_ast_argument *arg;
 	struct sieve_command_registration *cmd_reg = cmd->reg;
 
-	/* Validate any tags that might be present */
+	/* Resolve tagged arguments */
 	arg = sieve_ast_argument_first(cmd->ast_node);
-
-	/* Visit tagged and optional arguments */
-	while ( arg != NULL && sieve_ast_argument_type(arg) == SAAT_TAG ) {
-		struct sieve_ast_argument *parg;
+	while ( arg != NULL ) {
 		void *arg_data = NULL;
-		struct sieve_tag_registration *tag_reg =
-			sieve_validator_command_tag_get(valdtr, cmd, arg, &arg_data);
+		struct sieve_tag_registration *tag_reg;
 		const struct sieve_argument_def *tag_def;
 
+		if (sieve_ast_argument_type(arg) != SAAT_TAG) {
+			arg = sieve_ast_argument_next(arg);
+			continue;
+		}
+
+		tag_reg = sieve_validator_command_tag_get(valdtr, cmd, arg, &arg_data);
+
 		if ( tag_reg == NULL ) {
 			sieve_argument_validate_error(valdtr, arg,
 				"unknown tagged argument ':%s' for the %s %s "
-				"(reported only once at first occurence)",
+				"(reported only once at first occurrence)",
 				sieve_ast_argument_tag(arg), sieve_command_identifier(cmd),
 				sieve_command_type_name(cmd));
 			sieve_validator_register_unknown_tag
@@ -924,29 +938,40 @@ static bool sieve_validate_command_arguments
 			(arg->ast, tag_def, tag_reg->ext, tag_reg->id_code);
 		arg->argument->data = arg_data;
 
+		arg = sieve_ast_argument_next(arg);
+	}
+
+	/* Validate tagged arguments */
+	arg = sieve_ast_argument_first(cmd->ast_node);
+	while ( arg != NULL && sieve_ast_argument_type(arg) == SAAT_TAG) {
+		const struct sieve_argument_def *tag_def = arg->argument->def;
+		struct sieve_ast_argument *parg;
+
 		/* Scan backwards for any duplicates */
-		parg = sieve_ast_argument_prev(arg);
-		while ( parg != NULL ) {
-			if ( (sieve_ast_argument_type(parg) == SAAT_TAG &&
-					parg->argument->def == tag_reg->tag_def)
-				|| (tag_reg->id_code > 0 && parg->argument != NULL &&
-					parg->argument->id_code == tag_reg->id_code) )
-			{
-				const char *tag_id = sieve_ast_argument_tag(arg);
-				const char *tag_desc =
-					strcmp(tag_def->identifier, tag_id) != 0 ?
-					t_strdup_printf("%s argument (:%s)", tag_def->identifier, tag_id) :
-					t_strdup_printf(":%s argument", tag_def->identifier);
+		if ( (tag_def->flags & SIEVE_ARGUMENT_FLAG_MULTIPLE) == 0 ) {
+			parg = sieve_ast_argument_prev(arg);
+			while ( parg != NULL ) {
+				if ( (sieve_ast_argument_type(parg) == SAAT_TAG &&
+						parg->argument->def == tag_def)
+					|| (arg->argument->id_code > 0 && parg->argument != NULL &&
+						parg->argument->id_code == arg->argument->id_code) )
+				{
+					const char *tag_id = sieve_ast_argument_tag(arg);
+					const char *tag_desc =
+						strcmp(tag_def->identifier, tag_id) != 0 ?
+						t_strdup_printf("%s argument (:%s)", tag_def->identifier, tag_id) :
+						t_strdup_printf(":%s argument", tag_def->identifier);
 
-				sieve_argument_validate_error(valdtr, arg,
-					"encountered duplicate %s for the %s %s",
-					tag_desc, sieve_command_identifier(cmd),
-					sieve_command_type_name(cmd));
+					sieve_argument_validate_error(valdtr, arg,
+						"encountered duplicate %s for the %s %s",
+						tag_desc, sieve_command_identifier(cmd),
+						sieve_command_type_name(cmd));
 
-				return FALSE;
+					return FALSE;
+				}
+
+				parg = sieve_ast_argument_prev(parg);
 			}
-
-			parg = sieve_ast_argument_prev(parg);
 		}
 
 		/* Call the validation function for the tag (if present)
@@ -1203,7 +1228,7 @@ static bool sieve_validate_command_context
 	}	else {
 		sieve_validator_error(
 			valdtr, cmd_node->source_line,
-			"unknown %s '%s' (only reported once at first occurence)",
+			"unknown %s '%s' (only reported once at first occurrence)",
 			sieve_ast_type_name(ast_type), cmd_node->identifier);
 
 		sieve_validator_register_unknown_command(valdtr, cmd_node->identifier);
@@ -1280,9 +1305,7 @@ static bool sieve_validate_command
 			}
 		}
 
-		/* Skip block if result of test is const FALSEconst struct sieve_extension *sieve_validator_extension_load
-(struct sieve_validator *valdtr, struct sieve_command *cmd,
-	struct sieve_ast_argument *ext_arg, string_t *ext_name)  */
+		/* Skip block if result of test is const FALSE */
 		if ( result && *const_r == 0 )
 			return TRUE;
 
@@ -1352,18 +1375,26 @@ static bool sieve_validate_block
 			int const_value = -2;
 
 			next = sieve_ast_command_next(cmd_node);
-			command_success = sieve_validate_command_context(valdtr, cmd_node);
-			result = command_success && result;
 
 	 		/* Check if this is the first non-require command */
-			if ( command_success && sieve_ast_node_type(block) == SAT_ROOT
-				&& !valdtr->finished_require && cmd_node->command != NULL
-				&& !sieve_command_is(cmd_node->command, cmd_require) ) {
+			if ( sieve_ast_node_type(block) == SAT_ROOT
+				&& !valdtr->finished_require &&
+				strcasecmp(cmd_node->identifier, cmd_require.identifier) != 0 ) {
 				const struct sieve_validator_extension_reg *extrs;
+				const struct sieve_extension *const *exts;
 				unsigned int ext_count, i;
 
 				valdtr->finished_require = TRUE;
 
+				/* Load implicit extensions */
+				exts = sieve_extensions_get_all(valdtr->svinst, &ext_count);
+				for (i = 0; i < ext_count; i++) {
+					if ( exts[i]->implicit ) {
+						(void)sieve_validator_extension_load
+							(valdtr, NULL, NULL, exts[i]);
+					}
+				}
+
 				/* Validate all 'require'd extensions */
 				extrs = array_get(&valdtr->extensions, &ext_count);
 				for ( i = 0; i < ext_count; i++ ) {
@@ -1371,13 +1402,17 @@ static bool sieve_validate_block
 						&& extrs[i].valext->validate != NULL ) {
 
 						if ( !extrs[i].valext->validate
-							(extrs[i].ext, valdtr, extrs[i].context, extrs[i].arg) )
-						fatal = TRUE;
-						break;
+							(extrs[i].ext, valdtr, extrs[i].context, extrs[i].arg) ) {
+							fatal = TRUE;
+							break;
+						}
 					}
 				}
 			}
 
+			command_success = sieve_validate_command_context(valdtr, cmd_node);
+			result = command_success && result;
+
 			result = !fatal && sieve_validate_command(valdtr, cmd_node, &const_value)
 				&& result;
 
@@ -1404,7 +1439,7 @@ struct sieve_validator_object_reg {
 
 struct sieve_validator_object_registry {
 	struct sieve_validator *valdtr;
-	ARRAY_DEFINE(registrations, struct sieve_validator_object_reg);
+	ARRAY(struct sieve_validator_object_reg) registrations;
 };
 
 struct sieve_validator_object_registry *sieve_validator_object_registry_get
