@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2015 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -8,13 +8,12 @@
 #include "eacces-error.h"
 #include "home-expand.h"
 
-#include "sieve-limits.h"
 #include "sieve-settings.h"
 #include "sieve-extensions.h"
 #include "sieve-plugins.h"
 
 #include "sieve-script.h"
-#include "sieve-script-file.h"
+#include "sieve-storage-private.h"
 #include "sieve-ast.h"
 #include "sieve-binary.h"
 #include "sieve-actions.h"
@@ -47,8 +46,7 @@ struct sieve_instance *sieve_init
 	const struct sieve_callbacks *callbacks, void *context, bool debug)
 {
 	struct sieve_instance *svinst;
-	unsigned long long int uint_setting;
-	size_t size_setting;
+	const char *domain;
 	pool_t pool;
 
 	/* Create Sieve engine instance */
@@ -58,34 +56,48 @@ struct sieve_instance *sieve_init
 	svinst->callbacks = callbacks;
 	svinst->context = context;
 	svinst->debug = debug;
-	svinst->hostname = p_strdup_empty(pool, env->hostname);
 	svinst->base_dir = p_strdup_empty(pool, env->base_dir);
 	svinst->username = p_strdup_empty(pool, env->username);
 	svinst->home_dir = p_strdup_empty(pool, env->home_dir);
+	svinst->temp_dir = p_strdup_empty(pool, env->temp_dir);
 	svinst->flags = env->flags;
+	svinst->env_location = env->location;
+	svinst->delivery_phase = env->delivery_phase;
+
+	/* Determine domain */
+	if ( env->domainname != NULL && *(env->domainname) != '\0' ) {
+		domain = env->domainname;
+	} else {
+		/* Fall back to parsing username localpart@domain */
+		domain = svinst->username == NULL ? NULL :
+			strchr(svinst->username, '@');
+		if ( domain == NULL || *(domain+1) == '\0' ) {
+			/* Fall back to parsing hostname host.domain */
+			domain = ( env->hostname != NULL ? strchr(env->hostname, '.') : NULL );
+			if ( domain == NULL || *(domain+1) == '\0'
+				|| strchr(domain+1, '.') == NULL ) {
+				/* Fall back to bare hostname */
+				domain = env->hostname;
+			} else {
+				domain++;
+			}
+		} else {
+			domain++;
+		}
+	}
+	svinst->hostname = p_strdup_empty(pool, env->hostname);
+	svinst->domainname = p_strdup(pool, domain);
 
 	sieve_errors_init(svinst);
 
-	/* Read limits from configuration */
-
-	svinst->max_script_size = SIEVE_DEFAULT_MAX_SCRIPT_SIZE;
-	svinst->max_actions = SIEVE_DEFAULT_MAX_ACTIONS;
-	svinst->max_redirects = SIEVE_DEFAULT_MAX_REDIRECTS;
-
-	if ( sieve_setting_get_size_value
-		(svinst, "sieve_max_script_size", &size_setting) ) {
-		svinst->max_script_size = size_setting;
+	if ( debug ) {
+		sieve_sys_debug(svinst, "%s version %s initializing",
+			PIGEONHOLE_NAME, PIGEONHOLE_VERSION_FULL);
 	}
 
-	if ( sieve_setting_get_uint_value
-		(svinst, "sieve_max_actions", &uint_setting) ) {
-		svinst->max_actions = (unsigned int) uint_setting;
-	}
+	/* Read configuration */
 
-	if ( sieve_setting_get_uint_value
-		(svinst, "sieve_max_redirects", &uint_setting) ) {
-		svinst->max_redirects = (unsigned int) uint_setting;
-	}
+	sieve_settings_load(svinst);
 
 	/* Initialize extensions */
 	if ( !sieve_extensions_init(svinst) ) {
@@ -93,24 +105,29 @@ struct sieve_instance *sieve_init
 		return NULL;
 	}
 
+	/* Initialize storage classes */
+	sieve_storages_init(svinst);
+
+	/* Initialize plugins */
 	sieve_plugins_load(svinst, NULL, NULL);
 
+	/* Configure extensions */
 	sieve_extensions_configure(svinst);
 
 	return svinst;
 }
 
-void sieve_deinit(struct sieve_instance **svinst)
+void sieve_deinit(struct sieve_instance **_svinst)
 {
-	sieve_plugins_unload(*svinst);
+	struct sieve_instance *svinst = *_svinst;
 
-	sieve_extensions_deinit(*svinst);
+	sieve_plugins_unload(svinst);
+	sieve_storages_deinit(svinst);
+	sieve_extensions_deinit(svinst);
+	sieve_errors_deinit(svinst);
 
-	sieve_errors_deinit(*svinst);
-
-	pool_unref(&(*svinst)->pool);
-
-	*svinst = NULL;
+	pool_unref(&(svinst)->pool);
+	*_svinst = NULL;
 }
 
 void sieve_set_extensions
@@ -215,10 +232,22 @@ struct sieve_binary *sieve_compile_script
 {
 	struct sieve_ast *ast;
 	struct sieve_binary *sbin;
+	enum sieve_error error;
+
+	if ( error_r != NULL )
+		*error_r = SIEVE_ERROR_NONE;
+	else
+		error_r = &error;
 
 	/* Parse */
 	if ( (ast = sieve_parse(script, ehandler, error_r)) == NULL ) {
- 		sieve_error(ehandler, sieve_script_name(script), "parse failed");
+		switch ( *error_r ) {
+		case SIEVE_ERROR_NOT_FOUND:
+			sieve_error(ehandler, sieve_script_name(script), "script not found");
+			break;
+		default:
+	 		sieve_error(ehandler, sieve_script_name(script), "parse failed");
+		}
 		return NULL;
 	}
 
@@ -242,7 +271,7 @@ struct sieve_binary *sieve_compile_script
 	sieve_ast_unref(&ast);
 
 	if ( error_r != NULL )
-		error_r = SIEVE_ERROR_NONE;
+		*error_r = SIEVE_ERROR_NONE;
 
 	return sbin;
 }
@@ -254,15 +283,26 @@ struct sieve_binary *sieve_compile
 {
 	struct sieve_script *script;
 	struct sieve_binary *sbin;
+	enum sieve_error error;
 
-	if ( (script = sieve_script_create
-		(svinst, script_location, script_name, ehandler, error_r)) == NULL )
+	if ( (script = sieve_script_create_open
+		(svinst, script_location, script_name, &error)) == NULL ) {
+		if (error_r != NULL)
+			*error_r = error;
+		switch ( error ) {
+		case SIEVE_ERROR_NOT_FOUND:
+			sieve_error(ehandler, script_name, "script not found");
+			break;
+		default:
+			sieve_internal_error(ehandler, script_name, "failed to open script");
+		}
 		return NULL;
+	}
 
 	sbin = sieve_compile_script(script, ehandler, flags, error_r);
 
 	if ( svinst->debug && sbin != NULL ) {
-		sieve_sys_debug(svinst, "script `%s' from %s successfully compiled",
+		sieve_sys_debug(svinst, "Script `%s' from %s successfully compiled",
 			sieve_script_name(script), sieve_script_location(script));
 	}
 
@@ -334,9 +374,10 @@ struct sieve_binary *sieve_open_script
 			/* Ok, it exists; now let's see if it is up to date */
 			if ( !sieve_binary_up_to_date(sbin, flags) ) {
 				/* Not up to date */
-				if ( svinst->debug )
-					sieve_sys_debug(svinst, "script binary %s is not up-to-date",
+				if ( svinst->debug ) {
+					sieve_sys_debug(svinst, "Script binary %s is not up-to-date",
 						sieve_binary_path(sbin));
+				}
 
 				sieve_binary_unref(&sbin);
 				sbin = NULL;
@@ -347,18 +388,21 @@ struct sieve_binary *sieve_open_script
 		 * to (re-)compile.
 		 */
 		if ( sbin != NULL ) {
-			if ( svinst->debug )
-				sieve_sys_debug(svinst, "script binary %s successfully loaded",
+			if ( svinst->debug ) {
+				sieve_sys_debug(svinst,
+					"Script binary %s successfully loaded",
 					sieve_binary_path(sbin));
+			}
 
 		} else {
 			sbin = sieve_compile_script(script, ehandler, flags, error_r);
 
-			/* Save the binary if compile was successful */
 			if ( sbin != NULL ) {
-				if ( svinst->debug )
-					sieve_sys_debug(svinst, "script `%s' from %s successfully compiled",
+				if ( svinst->debug ) {
+					sieve_sys_debug(svinst,
+						"Script `%s' from %s successfully compiled",
 						sieve_script_name(script), sieve_script_location(script));
+				}
 			}
 		}
 	} T_END;
@@ -373,13 +417,21 @@ struct sieve_binary *sieve_open
 {
 	struct sieve_script *script;
 	struct sieve_binary *sbin;
+	enum sieve_error error;
 
 	/* First open the scriptfile itself */
-	script = sieve_script_create
-		(svinst, script_location, script_name, ehandler, error_r);
-
-	if ( script == NULL ) {
+	if ( (script=sieve_script_create_open
+		(svinst, script_location, script_name, &error)) == NULL ) {
 		/* Failed */
+		if (error_r != NULL)
+			*error_r = error;
+		switch ( error ) {
+		case SIEVE_ERROR_NOT_FOUND:
+			sieve_error(ehandler, script_name, "script not found");
+			break;
+		default:
+			sieve_internal_error(ehandler, script_name, "failed to open script");
+		}
 		return NULL;
 	}
 
@@ -506,12 +558,17 @@ int sieve_execute
 	if ( ret > 0 ) {
 		/* Execute result */
 		ret = sieve_result_execute(result, keep);
-	} else if ( ret == 0 ) {
+	} else if ( ret == SIEVE_EXEC_FAILURE ) {
 		/* Perform implicit keep if script failed with a normal runtime error */
-		if ( !sieve_result_implicit_keep(result) ) {
-			ret = SIEVE_EXEC_KEEP_FAILED;
-		} else {
+		switch ( sieve_result_implicit_keep(result) ) {
+		case SIEVE_EXEC_OK:
 			if ( keep != NULL ) *keep = TRUE;
+			break;
+		case SIEVE_EXEC_TEMP_FAILURE:
+			ret = SIEVE_EXEC_TEMP_FAILURE;
+			break;
+		default:
+			ret = SIEVE_EXEC_KEEP_FAILED;
 		}
 	}
 
@@ -645,6 +702,39 @@ int sieve_multiscript_status(struct sieve_multiscript *mscript)
 	return mscript->status;
 }
 
+int sieve_multiscript_tempfail(struct sieve_multiscript **mscript,
+	struct sieve_error_handler *ehandler)
+{
+	struct sieve_result *result = (*mscript)->result;
+	int ret = (*mscript)->status;
+
+	if ( ehandler != NULL )
+		sieve_result_set_error_handler(result, ehandler);
+
+	if ( (*mscript)->active ) {
+		ret = SIEVE_EXEC_TEMP_FAILURE;
+
+		if ( !(*mscript)->teststream && sieve_result_executed(result) ) {
+			/* Part of the result is already executed, need to fall back to
+			 * to implicit keep (FIXME)
+			 */
+			switch ( sieve_result_implicit_keep(result) ) {
+			case SIEVE_EXEC_OK:
+				ret = SIEVE_EXEC_FAILURE;
+				break;
+			default:
+				ret = SIEVE_EXEC_KEEP_FAILED;
+			}
+		}
+	}
+
+	/* Cleanup */
+	sieve_result_unref(&result);
+	*mscript = NULL;
+
+	return ret;
+}
+
 int sieve_multiscript_finish(struct sieve_multiscript **mscript,
 	struct sieve_error_handler *ehandler, bool *keep)
 {
@@ -652,7 +742,7 @@ int sieve_multiscript_finish(struct sieve_multiscript **mscript,
 	int ret = (*mscript)->status;
 
 	if ( ehandler != NULL )
-		sieve_result_set_error_handler((*mscript)->result, ehandler);
+		sieve_result_set_error_handler(result, ehandler);
 
 	if ( (*mscript)->active ) {
 		ret = SIEVE_EXEC_FAILURE;
@@ -660,10 +750,19 @@ int sieve_multiscript_finish(struct sieve_multiscript **mscript,
 		if ( (*mscript)->teststream ) {
 			(*mscript)->keep = TRUE;
 		} else {
-			if ( !sieve_result_implicit_keep((*mscript)->result) )
-				ret = SIEVE_EXEC_KEEP_FAILED;
-			else
+			switch ( sieve_result_implicit_keep(result) ) {
+			case SIEVE_EXEC_OK:
 				(*mscript)->keep = TRUE;
+				break;
+			case SIEVE_EXEC_TEMP_FAILURE:
+				if (!sieve_result_executed(result)) {
+					ret = SIEVE_EXEC_TEMP_FAILURE;
+					break;
+				}
+				/* fall through */
+			default:
+				ret = SIEVE_EXEC_KEEP_FAILED;
+			}
 		}
 	}
 
@@ -695,160 +794,5 @@ size_t sieve_max_script_size(struct sieve_instance *svinst)
 	return svinst->max_script_size;
 }
 
-/*
- * Script directory
- */
-
-struct sieve_directory {
-		struct sieve_instance *svinst;
-		DIR *dirp;
-
-		const char *path;
-};
-
-struct sieve_directory *sieve_directory_open
-(struct sieve_instance *svinst, const char *path, enum sieve_error *error_r)
-{
-	struct sieve_directory *sdir = NULL;
-	DIR *dirp;
-	struct stat st;
-
-	if ( error_r != NULL )
-		*error_r = SIEVE_ERROR_NONE;
-
-	if ( (path[0] == '~' && (path[1] == '/' || path[1] == '\0')) ||
-		(((svinst->flags & SIEVE_FLAG_HOME_RELATIVE) != 0 ) && path[0] != '/') ) {
-		/* Home-relative path. change to absolute. */
-		const char *home = sieve_environment_get_homedir(svinst);
-
-		if ( home != NULL ) {
-			if ( path[0] == '~' && (path[1] == '/' || path[1] == '\0') )
-				path = home_expand_tilde(path, home);
-			else
-				path = t_strconcat(home, "/", path, NULL);
-		} else {
-			sieve_sys_error(svinst,
-				"sieve dir path %s is relative to home directory, "
-				"but home directory is not available.", path);
-			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_TEMP_FAIL;
-			return NULL;
-		}
-	}
-
-	/* Specified path can either be a regular file or a directory */
-	if ( stat(path, &st) != 0 ) {
-		switch ( errno ) {
-		case ENOENT:
-			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_NOT_FOUND;
-			break;
-		case EACCES:
-			sieve_sys_error(svinst, "failed to open sieve dir: %s",
-				eacces_error_get("stat", path));
-			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_NO_PERM;
-			break;
-		default:
-			sieve_sys_error(svinst, "failed to open sieve dir: stat(%s) failed: %m",
-				path);
-			if ( error_r != NULL )
-				*error_r = SIEVE_ERROR_TEMP_FAIL;
-			break;
-		}
-		return NULL;
-	}
-
-	if ( S_ISDIR(st.st_mode) ) {
-
-		/* Open the directory */
-		if ( (dirp = opendir(path)) == NULL ) {
-			switch ( errno ) {
-			case ENOENT:
-				if ( error_r != NULL )
-					*error_r = SIEVE_ERROR_NOT_FOUND;
-				break;
-			case EACCES:
-				sieve_sys_error(svinst, "failed to open sieve dir: %s",
-					eacces_error_get("opendir", path));
-				if ( error_r != NULL )
-					*error_r = SIEVE_ERROR_NO_PERM;
-				break;
-			default:
-				sieve_sys_error(svinst, "failed to open sieve dir: opendir(%s) failed: "
-					"%m", path);
-				if ( error_r != NULL )
-					*error_r = SIEVE_ERROR_TEMP_FAIL;
-				break;
-			}
-			return NULL;
-		}
-
-		/* Create object */
-		sdir = t_new(struct sieve_directory, 1);
-		sdir->path = path;
-		sdir->dirp = dirp;
-	} else {
-		sdir = t_new(struct sieve_directory, 1);
-		sdir->path = path;
-		sdir->dirp = NULL;
-	}
-
-	sdir->svinst = svinst;
-
-	return sdir;
-}
-
-const char *sieve_directory_get_scriptfile(struct sieve_directory *sdir)
-{
-	const char *script = NULL;
-	struct dirent *dp;
-
-	if ( sdir->dirp != NULL ) {
-		while ( script == NULL ) {
-			const char *file;
-			struct stat st;
-
-			errno = 0;
-			if ( (dp = readdir(sdir->dirp)) == NULL ) {
-				if ( errno != 0 ) {
-					sieve_sys_error(sdir->svinst, "failed to read sieve dir: "
-						"readdir(%s) failed: %m", sdir->path);
-				}
-
-				return NULL;
-			}
-
-			if ( !sieve_scriptfile_has_extension(dp->d_name) )
-				continue;
-
-			if ( sdir->path[strlen(sdir->path)-1] == '/' )
-				file = t_strconcat(sdir->path, dp->d_name, NULL);
-			else
-				file = t_strconcat(sdir->path, "/", dp->d_name, NULL);
-
-			if ( stat(file, &st) != 0 || !S_ISREG(st.st_mode) )
-				continue;
-
-			script = file;
-		}
-	} else {
-		script = sdir->path;
-		sdir->path = NULL;
-	}
-
-	return script;
-}
-
-void sieve_directory_close(struct sieve_directory **sdir)
-{
-	/* Close the directory */
-	if ( (*sdir)->dirp != NULL && closedir((*sdir)->dirp) < 0 )
-		sieve_sys_error((*sdir)->svinst, "failed to close sieve dir: "
-			"closedir(%s) failed: %m", (*sdir)->path);
-
-	*sdir = NULL;
-}
-
 
 
