@@ -1,7 +1,8 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2014 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
+#include "abspath.h"
 #include "mempool.h"
 #include "hostpid.h"
 #include "ioloop.h"
@@ -30,6 +31,53 @@ struct sieve_storage_script {
 	struct sieve_storage *storage;
 };
 
+static int _file_path_cmp(const char *path1, const char *path2)
+{
+	const char *p1, *p2;
+	int ret;
+	
+	p1 = path1; p2 = path2;
+	if (*p2 == '\0' && p1 != '\0')
+		return 1;
+	if (*p1 == '\0' && p2 != '\0')
+		return -1;
+	if (*p1 == '/' && *p2 != '/')
+		return 1;
+	if (*p2 == '/' && *p1 != '/')
+		return -1;	
+	for (;;) {
+		const char *s1, *s2;
+		size_t size1, size2;
+
+		/* skip repeated slashes */
+		for (; *p1 == '/'; p1++);
+		for (; *p2 == '/'; p2++);	
+		/* check for end of comparison */
+		if (*p1 == '\0' || *p2 == '\0')
+			break;
+		/* mark start of path element */
+		s1 = p1;
+		s2 = p2;
+		/* scan to end of path elements */
+		for (; *p1 != '\0' && *p1 != '/'; p1++);
+		for (; *p2 != '\0' && *p2 != '/'; p2++);
+		/* compare sizes */
+		size1 = p1 - s1;
+		size2 = p2 - s2;
+		if (size1 != size2)
+			return size1 - size2;
+		/* compare */
+		if (size1 > 0 && (ret=memcmp(s1, s2, size1)) != 0)
+			return ret;
+	}
+	if (*p1 == '\0') {
+		if (*p2 == '\0')
+			return 0;
+		return -1;
+	}
+	return 1;
+}
+
 struct sieve_script *sieve_storage_script_init_from_path
 (struct sieve_storage *storage, const char *path,
 	const char *scriptname)
@@ -41,7 +89,7 @@ struct sieve_script *sieve_storage_script_init_from_path
 	/* Prevent initializing the active script link as a script when it
 	 * resides in the sieve storage directory.
 	 */
-	if ( *(storage->link_path) == '\0' ) {
+	if ( scriptname != NULL && *(storage->link_path) == '\0' ) {
 		const char *fname;
 
 		fname = strrchr(path, '/');
@@ -63,18 +111,18 @@ struct sieve_script *sieve_storage_script_init_from_path
 	st_script->file.script.pool = pool;
 	st_script->storage = storage;
 
-	if ( sieve_script_init
+	sieve_script_init
 		(&st_script->file.script, storage->svinst, &sieve_file_script, path,
-			scriptname,	sieve_storage_get_error_handler(storage), &error) != NULL ) {
-		return &st_script->file.script;
+			scriptname,	sieve_storage_get_error_handler(storage));
+
+	if ( sieve_script_open(&st_script->file.script, &error) < 0 ) {
+		if ( error == SIEVE_ERROR_NOT_FOUND )
+			sieve_storage_set_error(storage, error, "Script does not exist.");
+		pool_unref(&pool);
+		return NULL;
 	}
 
-	pool_unref(&pool);
-
-	if ( error == SIEVE_ERROR_NOT_FOUND )
-		sieve_storage_set_error(storage, error, "Script does not exist.");
-
-	return NULL;
+	return &st_script->file.script;
 }
 
 struct sieve_script *sieve_storage_script_init
@@ -118,26 +166,27 @@ static struct sieve_script *sieve_storage_script_init_from_file
 static int sieve_storage_read_active_link
 (struct sieve_storage *storage, const char **link_r)
 {
-  char linkbuf[PATH_MAX];
-  int ret;
+	int ret;
 
-	*link_r = NULL;
-
-	ret = readlink(storage->active_path, linkbuf, sizeof(linkbuf));
+	ret = t_readlink(storage->active_path, link_r);
 
 	if ( ret < 0 ) {
-		if (errno == EINVAL) {
+		*link_r = NULL;
+
+		if ( errno == EINVAL ) {
 			/* Our symlink is no symlink. Report 'no active script'.
 			 * Activating a script will automatically resolve this, so
 			 * there is no need to panic on this one.
 			 */
-			i_warning
-			  ("sieve-storage: Active sieve script symlink %s is no symlink.",
-			   storage->active_path);
+			if ( (storage->flags & SIEVE_STORAGE_FLAG_SYNCHRONIZING) == 0 ) {
+				i_warning
+				  ("sieve-storage: Active sieve script symlink %s is no symlink.",
+				   storage->active_path);
+			}
 			return 0;
 		}
 
-		if (errno == ENOENT ) {
+		if ( errno == ENOENT ) {
 			/* Symlink not found */
 			return 0;
 		}
@@ -150,11 +199,10 @@ static int sieve_storage_read_active_link
 	}
 
 	/* ret is now assured to be valid, i.e. > 0 */
-	*link_r = t_strndup(linkbuf, ret);
 	return 1;
 }
 
-static const char *sieve_storage_parse_link
+static const char *sieve_storage_parse_active_link
 (struct sieve_storage *storage, const char *link, const char **scriptname_r)
 {
 	const char *fname, *scriptname, *scriptpath;
@@ -182,8 +230,8 @@ static const char *sieve_storage_parse_link
 	}
 
 	/* Check whether the path is any good */
-	if ( strcmp(scriptpath, storage->link_path) != 0 &&
-		strcmp(scriptpath, storage->dir) != 0 ) {
+	if ( _file_path_cmp(scriptpath, storage->link_path) != 0 &&
+		_file_path_cmp(scriptpath, storage->dir) != 0 ) {
 		i_warning
 			("sieve-storage: Active sieve script symlink %s is broken: "
 				"invalid/unknown path to storage (points to %s).",
@@ -197,7 +245,7 @@ static const char *sieve_storage_parse_link
 	return fname;
 }
 
-int sieve_storage_get_active_scriptfile
+int sieve_storage_active_script_get_file
 (struct sieve_storage *storage, const char **file_r)
 {
 	const char *link, *scriptfile;
@@ -210,7 +258,7 @@ int sieve_storage_get_active_scriptfile
 		return ret;
 
 	/* Parse the link */
-	scriptfile = sieve_storage_parse_link(storage, link, NULL);
+	scriptfile = sieve_storage_parse_active_link(storage, link, NULL);
 
 	if (scriptfile == NULL) {
 		/* Obviously someone has been playing with our symlink,
@@ -224,7 +272,7 @@ int sieve_storage_get_active_scriptfile
 	return 1;
 }
 
-int sieve_storage_get_active_scriptname
+int sieve_storage_active_script_get_name
 (struct sieve_storage *storage, const char **name_r)
 {
 	const char *link;
@@ -236,7 +284,7 @@ int sieve_storage_get_active_scriptname
 	if ( (ret=sieve_storage_read_active_link(storage, &link)) <= 0 )
 		return ret;
 
-	if ( sieve_storage_parse_link(storage, link, name_r) == NULL ) {
+	if ( sieve_storage_parse_active_link(storage, link, name_r) == NULL ) {
 		/* Obviously someone has been playing with our symlink,
 		 * ignore this situation and report 'no active script'.
 		 * Activation should fix this situation.
@@ -247,7 +295,13 @@ int sieve_storage_get_active_scriptname
 	return 1;
 }
 
-struct sieve_script *sieve_storage_get_active_script
+const char *sieve_storage_active_script_get_path
+	(struct sieve_storage *storage)
+{
+	return storage->active_path;
+}
+
+struct sieve_script *sieve_storage_active_script_get
 (struct sieve_storage *storage)
 {
 	struct sieve_script *script;
@@ -268,7 +322,7 @@ struct sieve_script *sieve_storage_get_active_script
 	}
 
 	/* Parse the link */
-	scriptfile = sieve_storage_parse_link(storage, link, NULL);
+	scriptfile = sieve_storage_parse_active_link(storage, link, NULL);
 
 	if (scriptfile == NULL) {
 		/* Obviously someone has been playing with our symlink,
@@ -289,6 +343,55 @@ struct sieve_script *sieve_storage_get_active_script
 	return script;
 }
 
+int sieve_storage_active_script_get_last_change
+(struct sieve_storage *storage, time_t *last_change_r)
+{
+	struct stat st;
+
+	/* Try direct lstat first */
+	if (lstat(storage->active_path, &st) == 0) {
+		if (!S_ISLNK(st.st_mode)) {
+			*last_change_r = st.st_mtime;
+			return 0;
+		}
+	}
+	/* Check error */
+	else if (errno != ENOENT) {
+		sieve_storage_set_critical(storage, "lstat(%s) failed: %m",
+			   storage->active_path);
+	}
+
+	/* Fall back to statting storage directory */
+	return sieve_storage_get_last_change(storage, last_change_r);
+}
+
+int
+sieve_storage_active_script_is_no_link(struct sieve_storage *storage)
+{
+	struct stat st;
+
+	/* Stat the file */
+	if ( lstat(storage->active_path, &st) != 0 ) {
+		if ( errno != ENOENT ) {
+			sieve_storage_set_critical(storage,
+				"Failed to stat active sieve script symlink (%s): %m.",
+				storage->active_path);
+			return -1;
+		}
+		return 0;
+	}
+
+	if ( S_ISLNK( st.st_mode ) )
+		return 0;
+	if ( !S_ISREG( st.st_mode ) ) {
+		sieve_storage_set_critical( storage,
+			"Active sieve script file '%s' is no symlink nor a regular file.",
+			storage->active_path );
+		return -1;
+	}
+	return 1;
+}
+
 int sieve_storage_script_is_active(struct sieve_script *script)
 {
 	struct sieve_storage_script *st_script =
@@ -297,7 +400,7 @@ int sieve_storage_script_is_active(struct sieve_script *script)
 	int ret = 0;
 
 	T_BEGIN {
-		ret = sieve_storage_get_active_scriptfile(st_script->storage, &afile);
+		ret = sieve_storage_active_script_get_file(st_script->storage, &afile);
 
 		if ( ret > 0 ) {
 		 	/* Is the requested script active? */
@@ -334,6 +437,10 @@ int sieve_storage_script_delete(struct sieve_script **script)
 		}
 	}
 
+	/* unset INBOX mailbox attribute */
+	if ( ret >= 0 )
+		sieve_storage_inbox_script_attribute_unset(storage, (*script)->name);
+
 	/* Always deinitialize the script object */
 	sieve_script_unref(script);
 	return ret;
@@ -341,6 +448,7 @@ int sieve_storage_script_delete(struct sieve_script **script)
 
 static bool sieve_storage_rescue_regular_file(struct sieve_storage *storage)
 {
+	bool debug = ( (storage->flags & SIEVE_STORAGE_FLAG_DEBUG) != 0 );
 	struct stat st;
 
 	/* Stat the file */
@@ -355,8 +463,8 @@ static bool sieve_storage_rescue_regular_file(struct sieve_storage *storage)
 	}
 
 	if ( S_ISLNK( st.st_mode ) ) {
-		if ( storage->debug )
-			i_info( "sieve-storage: nothing to rescue %s.", storage->active_path);
+		if ( debug )
+			i_debug( "sieve-storage: nothing to rescue %s.", storage->active_path);
 		return TRUE; /* Nothing to rescue */
 	}
 
@@ -391,7 +499,7 @@ static bool sieve_storage_rescue_regular_file(struct sieve_storage *storage)
 	return FALSE;
 }
 
-int sieve_storage_deactivate(struct sieve_storage *storage)
+int sieve_storage_deactivate(struct sieve_storage *storage, time_t mtime)
 {
 	int ret;
 
@@ -406,10 +514,12 @@ int sieve_storage_deactivate(struct sieve_storage *storage)
 			sieve_storage_set_critical(storage, "sieve_storage_deactivate(): "
 				"error on unlink(%s): %m", storage->active_path);
 			return -1;
-		} else
+		} else {
 			return 0;
+		}
 	}
 
+	sieve_storage_set_modified(storage, mtime);
 	return 1;
 }
 
@@ -469,7 +579,7 @@ static int sieve_storage_replace_active_link
 	return 1;
 }
 
-static int _sieve_storage_script_activate(struct sieve_script *script)
+static int _sieve_storage_script_activate(struct sieve_script *script, time_t mtime)
 {
 	struct sieve_storage_script *st_script =
 		(struct sieve_storage_script *) script;
@@ -484,7 +594,7 @@ static int _sieve_storage_script_activate(struct sieve_script *script)
 	 * resolves automatically. This step is only necessary to provide a
 	 * proper return value indicating whether the script was already active.
 	 */
-	ret = sieve_storage_get_active_scriptfile(storage, &afile);
+	ret = sieve_storage_active_script_get_file(storage, &afile);
 
 	/* Is the requested script already active? */
 	if ( ret <= 0 || strcmp(st_script->file.filename, afile) != 0 )
@@ -528,15 +638,16 @@ static int _sieve_storage_script_activate(struct sieve_script *script)
 		}
 	}
 
+	sieve_storage_set_modified(storage, mtime);
 	return activated;
 }
 
-int sieve_storage_script_activate(struct sieve_script *script)
+int sieve_storage_script_activate(struct sieve_script *script, time_t mtime)
 {
 	int ret;
 
 	T_BEGIN {
-		ret = _sieve_storage_script_activate(script);
+		ret = _sieve_storage_script_activate(script, mtime);
 	} T_END;
 
 	return ret;
@@ -548,7 +659,7 @@ int sieve_storage_script_rename
 	struct sieve_storage_script *st_script =
 		(struct sieve_storage_script *) script;
 	struct sieve_storage *storage = st_script->storage;
-	const char *newpath, *newfile, *link_path;
+	const char *oldname = script->name, *newpath, *newfile, *link_path;
 	int ret = 0;
 
 	/* Check script name */
@@ -619,6 +730,10 @@ int sieve_storage_script_rename
 		}
 	} T_END;
 
+	/* rename INBOX mailbox attribute */
+	if ( ret >= 0 && oldname != NULL )
+		sieve_storage_inbox_script_attribute_rename(storage, oldname, newname);
+
 	return ret;
 }
 
