@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2013 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -7,6 +7,7 @@
 #include "mkdir-parents.h"
 #include "eacces-error.h"
 #include "unlink-old-files.h"
+#include "mail-storage-private.h"
 
 #include "sieve.h"
 #include "sieve-common.h"
@@ -22,6 +23,7 @@
 #include <sys/stat.h>
 #include <ctype.h>
 #include <time.h>
+#include <utime.h>
 
 #define SIEVE_DEFAULT_PATH "~/.dovecot."SIEVE_SCRIPT_FILEEXT
 
@@ -201,13 +203,36 @@ static int check_tmp(const char *path)
 	return 1;
 }
 
+static int _sieve_storage_open_inbox
+(struct mail_user *user, struct mailbox **box_r)
+{
+	struct mail_namespace *ns;
+	struct mailbox *box;
+	enum mailbox_flags flags = MAILBOX_FLAG_IGNORE_ACLS;
+	enum mail_error error;
+
+	ns = mail_namespace_find_inbox(user->namespaces);
+	*box_r = box = mailbox_alloc(ns->list, "INBOX", flags);
+	if (mailbox_open(box) == 0)
+		return 0;
+
+	i_warning("sieve-storage: "
+		"Failed to open user INBOX for attribute modifications: %s",
+		mailbox_get_last_error(box, &error));
+	return -1;
+}
+
 static struct sieve_storage *_sieve_storage_create
-(struct sieve_instance *svinst, const char *user, const char *home, bool debug)
+(struct sieve_instance *svinst, struct mail_user *user, const char *home,
+	enum sieve_storage_flags flags)
 {
 	pool_t pool;
 	struct sieve_storage *storage;
+	struct mailbox *inbox = NULL;
+	bool debug = ( (flags & SIEVE_STORAGE_FLAG_DEBUG) != 0 );
 	const char *tmp_dir, *link_path, *path;
 	const char *sieve_data, *active_path, *active_fname, *storage_dir;
+	const char *username = user->username;
 	mode_t dir_create_mode, file_create_mode;
 	gid_t file_create_gid;
 	const char *file_create_gid_origin;
@@ -262,7 +287,8 @@ static struct sieve_storage *_sieve_storage_create
 	path = home_expand_tilde(active_path, home);
 	if ( path == NULL ) {
 		i_error("sieve-storage: userdb(%s) didn't return a home directory "
-			"for substitition in active script path (sieve=%s)", user, active_path);
+			"for substitition in active script path (sieve=%s)",
+			username, active_path);
 		return NULL;
 	}
 
@@ -365,7 +391,7 @@ static struct sieve_storage *_sieve_storage_create
 	if ( path == NULL ) {
 		i_error("sieve-storage: userdb(%s) didn't return a home directory "
 			"for substitition in storage root directory (sieve_dir=%s)",
-			user, storage_dir);
+			username, storage_dir);
 		return NULL;
 	}
 
@@ -378,6 +404,17 @@ static struct sieve_storage *_sieve_storage_create
 			"using sieve script storage directory: %s", storage_dir);
 	}
 
+	/* get the storage mtime before we modify it ourself. FIXME: do this
+	   later, only just before modifying the sieve dir */
+	struct stat st;
+	if (stat(storage_dir, &st) < 0) {
+		if (errno != ENOENT) {
+			i_error("stat(%s) failed: %m", storage_dir);
+			return NULL;
+		}
+		st.st_mtime = 0;
+	}
+
 	/* Get permissions */
 
 	sieve_storage_get_permissions
@@ -400,6 +437,10 @@ static struct sieve_storage *_sieve_storage_create
 		file_create_gid_origin, debug) < 0 )
 		return NULL;
 
+	/* Open user's INBOX for attribute updates if necessary */
+	if ( (flags & SIEVE_STORAGE_FLAG_SYNCHRONIZING) == 0 )
+		(void)_sieve_storage_open_inbox(user, &inbox);
+
 	/*
 	 * Create storage object
 	 */
@@ -407,17 +448,20 @@ static struct sieve_storage *_sieve_storage_create
 	pool = pool_alloconly_create("sieve-storage", 512+256);
 	storage = p_new(pool, struct sieve_storage, 1);
 	storage->svinst = svinst;
-	storage->debug = debug;
+	storage->flags = flags;
 	storage->pool = pool;
 	storage->dir = p_strdup(pool, storage_dir);
-	storage->user = p_strdup(pool, user);
+	storage->username = p_strdup(pool, username);
 	storage->active_path = p_strdup(pool, active_path);
 	storage->active_fname = p_strdup(pool, active_fname);
+	storage->prev_mtime = st.st_mtime;
 
 	storage->dir_create_mode = dir_create_mode;
 	storage->file_create_mode = file_create_mode;
 	storage->file_create_gid = file_create_gid;
 
+	storage->inbox = inbox;
+
 	/* Get the path to be prefixed to the script name in the symlink pointing
 	 * to the active script.
 	 */
@@ -447,38 +491,77 @@ static struct sieve_storage *_sieve_storage_create
 
 	if ( debug ) {
 		if ( storage->max_storage > 0 ) {
-			i_info("sieve-storage: quota: storage limit: %llu bytes",
+			i_debug("sieve-storage: quota: storage limit: %llu bytes",
 				(unsigned long long int) storage->max_storage);
 		}
 		if ( storage->max_scripts > 0 ) {
-			i_info("sieve-storage: quota: script count limit: %llu scripts",
+			i_debug("sieve-storage: quota: script count limit: %llu scripts",
 				(unsigned long long int) storage->max_scripts);
 		}
 	}
-
 	return storage;
 }
 
 struct sieve_storage *sieve_storage_create
-(struct sieve_instance *svinst, const char *user, const char *home, bool debug)
+(struct sieve_instance *svinst, struct mail_user *user, const char *home,
+	enum sieve_storage_flags flags)
 {
 	struct sieve_storage *storage;
 
 	T_BEGIN {
-		storage = _sieve_storage_create(svinst, user, home, debug);
+		storage = _sieve_storage_create(svinst, user, home, flags);
 	} T_END;
 
 	return storage;
 }
 
-
 void sieve_storage_free(struct sieve_storage *storage)
 {
+	if (storage->inbox != NULL)
+		mailbox_free(&storage->inbox);
 	sieve_error_handler_unref(&storage->ehandler);
 
 	pool_unref(&storage->pool);
 }
 
+int sieve_storage_get_last_change
+(struct sieve_storage *storage, time_t *last_change_r)
+{
+	*last_change_r = storage->prev_mtime;
+	return 0;
+}
+
+void sieve_storage_set_modified
+(struct sieve_storage *storage, time_t mtime)
+{
+	struct utimbuf times;
+	time_t cur_mtime;
+
+	if ( mtime != (time_t)-1 ) {
+		if ( sieve_storage_get_last_change(storage, &cur_mtime) >= 0 &&
+			cur_mtime > mtime )
+			return;
+	} else {
+		mtime = ioloop_time;
+	}
+
+	times.actime = mtime;
+	times.modtime = mtime;
+	if ( utime(storage->dir, &times) < 0 ) {
+		switch ( errno ) {
+		case ENOENT:
+			break;
+		case EACCES:
+			i_error("sieve-storage: %s", eacces_error_get("utime", storage->dir));
+			break;
+		default:
+			i_error("sieve-storage: utime(%s) failed: %m", storage->dir);
+		}
+	} else {
+		storage->prev_mtime = mtime;
+	}
+}
+
 /* Error handling */
 
 struct sieve_error_handler *sieve_storage_get_error_handler
@@ -540,20 +623,6 @@ void sieve_storage_set_error
 	storage->error_code = error;
 }
 
-void sieve_storage_set_internal_error(struct sieve_storage *storage)
-{
-	struct tm *tm;
-	char str[256];
-
-	tm = localtime(&ioloop_time);
-
-	i_free(storage->error);
-	storage->error_code = SIEVE_ERROR_TEMP_FAIL;
-	storage->error =
-	  strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
-	  i_strdup(str) : i_strdup(CRITICAL_MSG);
-}
-
 void sieve_storage_set_critical
 (struct sieve_storage *storage, const char *fmt, ...)
 {
@@ -561,14 +630,31 @@ void sieve_storage_set_critical
 
 	sieve_storage_clear_error(storage);
 	if (fmt != NULL) {
-		va_start(va, fmt);
-		i_error("sieve-storage: %s", t_strdup_vprintf(fmt, va));
-		va_end(va);
+		i_free(storage->error);
+		storage->error_code = SIEVE_ERROR_TEMP_FAIL;
 
-		/* critical errors may contain sensitive data, so let user
-		   see only "Internal error" with a timestamp to make it
-		   easier to look from log files the actual error message. */
-		sieve_storage_set_internal_error(storage);
+		if ( (storage->flags & SIEVE_STORAGE_FLAG_SYNCHRONIZING) == 0 ) {
+			struct tm *tm;
+			char str[256];
+
+			va_start(va, fmt);
+			i_error("sieve-storage: %s", t_strdup_vprintf(fmt, va));
+			va_end(va);
+
+			/* critical errors may contain sensitive data, so let user
+			   see only "Internal error" with a timestamp to make it
+			   easier to look from log files the actual error message. */
+			tm = localtime(&ioloop_time);
+			storage->error =
+				strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
+				i_strdup(str) : i_strdup(CRITICAL_MSG);
+		} else {
+			/* no user is involved while synchronizing, so do it the
+			   normal way */
+			va_start(va, fmt);
+			storage->error = i_strdup_vprintf(fmt, va);
+			va_end(va);
+		}
 	}
 }
 
@@ -586,4 +672,73 @@ const char *sieve_storage_get_last_error
 	return storage->error != NULL ? storage->error : "Unknown error";
 }
 
+/*
+ * INBOX attributes
+ */
+
+static void sieve_storage_inbox_transaction_finish
+(struct sieve_storage *storage, struct mailbox_transaction_context **t)
+{
+	struct mailbox *inbox = storage->inbox;
+
+	if (mailbox_transaction_commit(t) < 0) {
+		enum mail_error error;
+		
+		i_warning("sieve-storage: Failed to update INBOX attributes: %s",
+			mail_storage_get_last_error(mailbox_get_storage(inbox), &error));
+	}
+}
+
+void sieve_storage_inbox_script_attribute_set
+(struct sieve_storage *storage, const char *name)
+{
+	struct mailbox_transaction_context *t;
+	const char *key;
+
+	if (storage->inbox == NULL)
+		return;
+
+	key = t_strconcat
+		(MAILBOX_ATTRIBUTE_PREFIX_SIEVE_FILES, name, NULL);
+	t = mailbox_transaction_begin(storage->inbox, 0);	
+	mail_index_attribute_set(t->itrans, TRUE, key, ioloop_time, 0);
+	sieve_storage_inbox_transaction_finish(storage, &t);
+}
+
+void sieve_storage_inbox_script_attribute_rename
+(struct sieve_storage *storage, const char *oldname, const char *newname)
+{
+	struct mailbox_transaction_context *t;
+	const char *oldkey, *newkey;
+
+	if (storage->inbox == NULL)
+		return;
+
+	oldkey = t_strconcat
+		(MAILBOX_ATTRIBUTE_PREFIX_SIEVE_FILES, oldname, NULL);
+	newkey = t_strconcat
+		(MAILBOX_ATTRIBUTE_PREFIX_SIEVE_FILES, newname, NULL);
+
+	t = mailbox_transaction_begin(storage->inbox, 0);	
+	mail_index_attribute_unset(t->itrans, TRUE, oldkey, ioloop_time);
+	mail_index_attribute_set(t->itrans, TRUE, newkey, ioloop_time, 0);
+	sieve_storage_inbox_transaction_finish(storage, &t);
+}
+
+void sieve_storage_inbox_script_attribute_unset
+(struct sieve_storage *storage, const char *name)
+{
+	struct mailbox_transaction_context *t;
+	const char *key;
+
+	if (storage->inbox == NULL)
+		return;
+
+	key = t_strconcat
+		(MAILBOX_ATTRIBUTE_PREFIX_SIEVE_FILES, name, NULL);
+	
+	t = mailbox_transaction_begin(storage->inbox, 0);	
+	mail_index_attribute_unset(t->itrans, TRUE, key, ioloop_time);
+	sieve_storage_inbox_transaction_finish(storage, &t);
+}
 
