@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2015 Pigeonhole authors, see the included COPYING file
  */
 
 #ifndef __CLIENT_AUTHENTICATE_H
@@ -9,12 +9,13 @@ struct managesieve_arg;
 const char *client_authenticate_get_capabilities
 	(struct client *client);
 
-bool managesieve_client_auth_handle_reply(struct client *client,
-				   const struct client_auth_reply *reply);
+void managesieve_client_auth_result
+	(struct client *client, enum client_auth_result result,
+		const struct client_auth_reply *reply, const char *text);
 
 void managesieve_client_auth_send_challenge
 	(struct client *client, const char *data);
-int managesieve_client_auth_parse_response
+void managesieve_client_auth_parse_response
 	(struct client *client);
 
 int cmd_authenticate
