@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2015 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -8,8 +8,7 @@
 
 #include "sieve.h"
 #include "sieve-script.h"
-
-#include "sieve-storage-quota.h"
+#include "sieve-storage.h"
 
 #include "managesieve-client.h"
 #include "managesieve-quota.h"
