@@ -1,11 +1,11 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2014 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
 #include "str.h"
 
+#include "sieve.h"
 #include "sieve-storage.h"
-#include "sieve-storage-list.h"
 
 #include "managesieve-quote.h"
 
@@ -15,7 +15,7 @@
 bool cmd_listscripts(struct client_command_context *cmd)
 {
   struct client *client = cmd->client;
-	struct sieve_list_context *ctx;
+	struct sieve_storage_list_context *ctx;
 	const char *scriptname;
 	bool active;
 	string_t *str;
