@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2012 Pigeonhole authors, see the included COPYING file
+/* Copyright (c) 2002-2015 Pigeonhole authors, see the included COPYING file
  */
 
 #include "lib.h"
@@ -6,14 +6,16 @@
 #include "llist.h"
 #include "str.h"
 #include "hostpid.h"
-#include "network.h"
+#include "net.h"
 #include "istream.h"
 #include "ostream.h"
+#include "iostream.h"
 #include "var-expand.h"
 #include "master-service.h"
 #include "mail-storage-service.h"
 #include "mail-namespace.h"
 
+#include "sieve.h"
 #include "sieve-storage.h"
 
 #include "managesieve-quote.h"
@@ -21,7 +23,6 @@
 #include "managesieve-commands.h"
 #include "managesieve-client.h"
 
-#include <stdlib.h>
 #include <unistd.h>
 
 extern struct mail_storage_callbacks mail_storage_callbacks;
@@ -58,30 +59,35 @@ static void client_idle_timeout(struct client *client)
 }
 
 static struct sieve_storage *client_get_storage
-(struct sieve_instance *svinst, struct mail_user *user,
-	const struct managesieve_settings *set)
+(struct sieve_instance *svinst, struct mail_user *user)
 {
 	struct sieve_storage *storage;
-	const char *home;
-
-	if ( mail_user_get_home(user, &home) <= 0 )
-		home = NULL;
-
-	storage = sieve_storage_create
-		(svinst, user->username, home, set->mail_debug);
+	enum sieve_error error;
 
+	storage = sieve_storage_create_main
+		(svinst, user, SIEVE_STORAGE_FLAG_READWRITE, &error);
 	if (storage == NULL) {
-		struct tm *tm;
-		char str[256];
+		switch (error) {
+		case SIEVE_ERROR_NOT_FOUND:
+			printf("BYE \"Sieve filtering is disabled for this user.\"\n");
 
-		tm = localtime(&ioloop_time);
+			i_fatal("Failed to open Sieve storage: Sieve disabled for user");
+			break;
+		default:
+			{ 
+				struct tm *tm;
+				char str[256];
 
-		printf("BYE \"%s\"\n",
-			strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
-				i_strdup(str) : i_strdup(CRITICAL_MSG));
+				tm = localtime(&ioloop_time);
 
-		i_fatal("Failed to open Sieve storage.");
-    }
+				printf("BYE \"%s\"\n",
+					strftime(str, sizeof(str), CRITICAL_MSG_STAMP, tm) > 0 ?
+						i_strdup(str) : i_strdup(CRITICAL_MSG));
+
+				i_fatal("Failed to open Sieve storage.");
+			}
+		}
+	}
 
 	return storage;
 }
@@ -98,12 +104,7 @@ struct client *client_create
 	struct sieve_storage *storage;
 	pool_t pool;
 
-	/* Always use nonblocking I/O */
-
-	net_set_nonblock(fd_in, TRUE);
-	net_set_nonblock(fd_out, TRUE);
-
-	/* Initialize Sieve instance */
+	/* Initialize Sieve */
 
 	memset((void*)&svenv, 0, sizeof(svenv));
 	svenv.username = user->username;
@@ -116,7 +117,7 @@ struct client *client_create
 
 	/* Get Sieve storage */
 
-	storage = client_get_storage(svinst, user, set);
+	storage = client_get_storage(svinst, user);
 
 	/* always use nonblocking I/O */
 	net_set_nonblock(fd_in, TRUE);
@@ -134,9 +135,13 @@ struct client *client_create
 		(fd_in, set->managesieve_max_line_length, FALSE);
 	client->output = o_stream_create_fd(fd_out, (size_t)-1, FALSE);
 
+	o_stream_set_no_error_handling(client->output, TRUE);
+	i_stream_set_name(client->input, "<managesieve client>");
+	o_stream_set_name(client->output, "<managesieve client>");
+
 	o_stream_set_flush_callback(client->output, client_output, client);
 
-	client->io = io_add(fd_in, IO_READ, client_input, client);
+	client->io = io_add_istream(client->input, client_input, client);
 	client->last_input = ioloop_time;
 	client->parser = managesieve_parser_create
 		(client->input, set->managesieve_max_line_length);
@@ -170,6 +175,14 @@ struct client *client_create
 static const char *client_stats(struct client *client)
 {
 	static struct var_expand_table static_tab[] = {
+		{ 't', NULL, "put_bytes" },
+		{ 'p', NULL, "put_count" },
+		{ 'b', NULL, "get_bytes" },
+		{ 'g', NULL, "get_count" },
+		{ 'v', NULL, "check_bytes" },
+		{ 'c', NULL, "check_count" },
+		{ 'd', NULL, "deleted_count" },
+		{ 'r', NULL, "renamed_count" },
 		{ 'i', NULL, "input" },
 		{ 'o', NULL, "output" },
 		{ '\0', NULL, "session" },
@@ -181,24 +194,23 @@ static const char *client_stats(struct client *client)
 	tab = t_malloc(sizeof(static_tab));
 	memcpy(tab, static_tab, sizeof(static_tab));
 
-	tab[0].value = dec2str(client->input->v_offset);
-	tab[1].value = dec2str(client->output->offset);
-	tab[2].value = client->session_id;
+	tab[0].value = dec2str(client->put_bytes);
+	tab[1].value = dec2str(client->put_count);
+	tab[2].value = dec2str(client->get_bytes);
+	tab[3].value = dec2str(client->get_count);
+	tab[4].value = dec2str(client->check_bytes);
+	tab[5].value = dec2str(client->check_count);
+	tab[6].value = dec2str(client->deleted_count);
+	tab[7].value = dec2str(client->renamed_count);
+	tab[8].value = dec2str(i_stream_get_absolute_offset(client->input));
+	tab[9].value = dec2str(client->output->offset);
+	tab[10].value = client->session_id;
 
 	str = t_str_new(128);
 	var_expand(str, client->set->managesieve_logout_format, tab);
 	return str_c(str);
 }
 
-static const char *client_get_disconnect_reason(struct client *client)
-{
-	errno = client->input->stream_errno != 0 ?
-		client->input->stream_errno :
-		client->output->stream_errno;
-	return errno == 0 || errno == EPIPE ? "Connection closed" :
-		t_strdup_printf("Connection closed: %m");
-}
-
 void client_destroy(struct client *client, const char *reason)
 {
 	int ret;
@@ -209,14 +221,13 @@ void client_destroy(struct client *client, const char *reason)
 
 	if (!client->disconnected) {
 		client->disconnected = TRUE;
-		if (reason == NULL)
-			reason = client_get_disconnect_reason(client);
+		if (reason == NULL) {
+			reason = io_stream_get_disconnect_reason(client->input,
+								 client->output);
+		}
 		i_info("%s %s", reason, client_stats(client));
 	}
 
-	managesieve_client_count--;
-	DLLIST_REMOVE(&managesieve_clients, client);
-
 	if (client->command_pending) {
 		/* try to deinitialize the command */
 		i_assert(client->cmd.func != NULL);
@@ -248,23 +259,29 @@ void client_destroy(struct client *client, const char *reason)
 	i_stream_destroy(&client->input);
 	o_stream_destroy(&client->output);
 
-	if (close(client->fd_in) < 0)
-		i_error("close(client in) failed: %m");
-	if (client->fd_in != client->fd_out) {
-		if (close(client->fd_out) < 0)
-			i_error("close(client out) failed: %m");
-	}
+	net_disconnect(client->fd_in);
+	if (client->fd_in != client->fd_out)
+		net_disconnect(client->fd_out);
 
-	sieve_storage_free(client->storage);
+	sieve_storage_unref(&client->storage);
 	sieve_deinit(&client->svinst);
 
 	pool_unref(&client->cmd.pool);
+	mail_storage_service_user_free(&client->service_user);
+
+	managesieve_client_count--;
+	DLLIST_REMOVE(&managesieve_clients, client);
 	pool_unref(&client->pool);
 
 	master_service_client_connection_destroyed(master_service);
 	managesieve_refresh_proctitle();
 }
 
+static void client_destroy_timeout(struct client *client)
+{
+	client_destroy(client, NULL);
+}
+
 void client_disconnect(struct client *client, const char *reason)
 {
 	i_assert(reason != NULL);
@@ -274,10 +291,15 @@ void client_disconnect(struct client *client, const char *reason)
 
 	i_info("Disconnected: %s %s", reason, client_stats(client));
 	client->disconnected = TRUE;
-	(void)o_stream_flush(client->output);
+	o_stream_flush(client->output);
+	o_stream_uncork(client->output);
 
 	i_stream_close(client->input);
 	o_stream_close(client->output);
+
+	if (client->to_idle != NULL)
+		timeout_remove(&client->to_idle);
+	client->to_idle = timeout_add(0, client_destroy_timeout, client);
 }
 
 void client_disconnect_with_error(struct client *client, const char *msg)
@@ -378,12 +400,11 @@ void client_send_storage_error
 	error = sieve_storage_get_last_error(storage, &error_code);
 
 	switch ( error_code ) {
-	case SIEVE_ERROR_TEMP_FAIL:
+	case SIEVE_ERROR_TEMP_FAILURE:
 		client_send_noresp(client, "TRYLATER", error);
 		break;
 
 	case SIEVE_ERROR_NO_QUOTA:
-	case SIEVE_ERROR_NO_SPACE: /* Not sure if this is appropriate */
 		client_send_noresp(client, "QUOTA", error);
 		break;
 
